Package: ROCR
Authors@R: c(person("Tobias","Sing",
                    email = "tobias.sing@gmail.com",role="aut"),
             person("Oliver","Sander",
                    email = "osander@gmail.com",role="aut"),
             person("Niko","Beerenwinkel",
                    role="aut"),
             person("Thomas","Lengauer",
                    role="aut"),
             person("Thomas","Unterthiner",
                    role="ctb"),
             person("Felix G.M.","Ernst",
                    email = "felix.gm.ernst@outlook.com",role="cre",
                    comment = c(ORCID = "0000-0001-5064-0928")))
Version: 1.0-12
Date: 2026-01-22
Title: Visualizing the Performance of Scoring Classifiers
Description: ROC graphs, sensitivity/specificity curves, lift charts,
  and precision/recall plots are popular examples of trade-off
  visualizations for specific pairs of performance measures. ROCR is a
  flexible tool for creating cutoff-parameterized 2D performance curves
  by freely combining two from over 25 performance measures (new
  performance measures can be added using a standard interface).
  Curves from different cross-validation or bootstrapping runs can be
  averaged by different methods, and standard deviations, standard
  errors or box plots can be used to visualize the variability across
  the runs. The parameterization can be visualized by printing cutoff
  values at the corresponding curve positions, or by coloring the
  curve according to cutoff. All components of a performance plot can
  be quickly adjusted using a flexible parameter dispatching
  mechanism. Despite its flexibility, ROCR is easy to use, with only
  three commands and reasonable default values for all optional
  parameters.
Encoding: UTF-8
License: GPL (>= 2)
NeedsCompilation: no
Depends: R (>= 3.6)
Imports: methods, graphics, grDevices, gplots, stats
Suggests: testthat, knitr, rmarkdown
URL: https://ipa-tys.github.io/ROCR/
BugReports: https://github.com/ipa-tys/ROCR/issues
RoxygenNote: 7.3.3
VignetteBuilder: knitr
Packaged: 2026-01-22 15:23:44 UTC; flixr
Author: Tobias Sing [aut],
  Oliver Sander [aut],
  Niko Beerenwinkel [aut],
  Thomas Lengauer [aut],
  Thomas Unterthiner [ctb],
  Felix G.M. Ernst [cre] (ORCID: <https://orcid.org/0000-0001-5064-0928>)
Maintainer: Felix G.M. Ernst <felix.gm.ernst@outlook.com>
Repository: CRAN
Date/Publication: 2026-01-23 06:41:23 UTC
Built: R 4.6.0; ; 2026-04-27 06:17:10 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(plot,performance)
export(performance)
export(prediction)
exportClasses(performance)
exportClasses(prediction)
exportMethods(plot)
import(graphics)
import(methods)
import(stats)
importFrom(grDevices,chull)
importFrom(grDevices,rainbow)
importFrom(grDevices,xy.coords)
importFrom(graphics,par)
importFrom(graphics,plot.default)
importFrom(graphics,plot.xy)
importFrom(stats,approxfun)
importFrom(stats,chisq.test)
importFrom(stats,median)
importFrom(stats,sd)
importFrom(stats,uniroot)

# ROCR <img src="https://raw.githubusercontent.com/ipa-tys/ROCR/rocr-images/ROCR_small.png" align="right">

<!-- badges: start -->
<!-- [![R-CMD-check](https://github.com/ipa-tys/ROCR/workflows/R-CMD-check/badge.svg)](https://github.com/ipa-tys/ROCR/actions?query=workflow:R-CMD-check) -->
[![CRAN Status](https://www.r-pkg.org/badges/version/ROCR)](https://CRAN.r-project.org/package=ROCR)
[![codecov](https://app.codecov.io/gh/ipa-tys/ROCR/branch/master/graph/badge.svg)](https://app.codecov.io/gh/ipa-tys/ROCR)
<!-- badges: end -->

*visualizing classifier performance in R, with only 3 commands*

![](https://raw.githubusercontent.com/ipa-tys/ROCR/rocr-images/ourplot_website.gif)

### Please support our work by citing the ROCR article in your publications:

***Sing T, Sander O, Beerenwinkel N, Lengauer T. [2005]
ROCR: visualizing classifier performance in R.
Bioinformatics 21(20):3940-1.***

Free full text:
http://bioinformatics.oxfordjournals.org/content/21/20/3940.full

[<img src="https://raw.githubusercontent.com/ipa-tys/ROCR/rocr-images/logo_mpi_430.png" align="right" width="300">](https://www.mpi-inf.mpg.de/home/)  
`ROCR` was originally developed at the [Max Planck Institute for Informatics](https://www.mpi-inf.mpg.de/home/)   


## Introduction

`ROCR` (with obvious pronounciation) is an R package for evaluating and visualizing classifier performance. It is...

 - ...easy to use: adds only three new commands to R.
 - ...flexible: integrates tightly with R's built-in graphics facilities.
 - ...powerful: Currently, 28 performance measures are implemented, which can be freely combined to form parametric curves such as ROC curves, precision/recall curves, or lift curves. Many options such as curve averaging (for cross-validation or bootstrap), augmenting the averaged curves by standard error bar or boxplots, labeling cutoffs to the curve, or coloring curves according to cutoff.

### Performance measures that `ROCR` knows:

Accuracy, error rate, true positive rate, false positive rate, true negative rate, false negative rate, sensitivity, specificity, recall, positive predictive value, negative predictive value, precision, fallout, miss, phi correlation coefficient, Matthews correlation coefficient, mutual information, chi square statistic, odds ratio, lift value, precision/recall F measure, ROC convex hull, area under the ROC curve, precision/recall break-even point, calibration error, mean cross-entropy, root mean squared error, SAR measure, expected cost, explicit cost.

### `ROCR` features:

ROC curves, precision/recall plots, lift charts, cost curves, custom curves by freely selecting one performance measure for the x axis and one for the y axis, handling of data from cross-validation or bootstrapping, curve averaging (vertically, horizontally, or by threshold), standard error bars, box plots, curves that are color-coded by cutoff, printing threshold values on the curve, tight integration with Rs plotting facilities (making it easy to adjust plots or to combine multiple plots), fully customizable, easy to use (only 3 commands).

## Installation of `ROCR`

The most straightforward way to install and use `ROCR` is to install it from 
`CRAN` by starting `R` and using the `install.packages` function:

```
install.packages("ROCR")
```

Alternatively you can install it from command line using the tar ball like this:

```
R CMD INSTALL ROCR_*.tar.gz
```

## Getting started

from within R ...

```
library(ROCR)
demo(ROCR)
help(package=ROCR)
```

## Examples

Using ROCR's 3 commands to produce a simple ROC plot:

```
pred <- prediction(predictions, labels)
perf <- performance(pred, measure = "tpr", x.measure = "fpr")
plot(perf, col=rainbow(10))
```

## Documentation

 - The Reference Manual found [here](https://CRAN.r-project.org/package=ROCR)
 - Slide deck for a tutorial talk (feel free to re-use for teaching, but please give appropriate credits and write us an email) [[PPT](https://raw.githubusercontent.com/ipa-tys/ROCR/rocr-images/ROCR_Talk_Tobias_Sing.ppt)]
 - A few pointers to the literature on classifier evaluation


## Contact
Questions, comments, and suggestions are very welcome. Open an issue on GitHub and we can discuss. We are also interested in seeing how ROCR is used in publications. Thus, if you have prepared a paper using ROCR we'd be happy to know.

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup--------------------------------------------------------------------
library(ROCR)

## ----echo = FALSE, results = 'asis'-------------------------------------------
table <- data.frame(group = c("Contingency ratios",
                              "Discrete covariation measures",
                              "Information retrieval measures",
                              "Performance in ROC space",
                              "Absolute scoring performance",
                              "Cost measures"),
                    measure = c("error rate, accuracy, sensitivity, specificity, true/false positive rate, fallout, miss, precision, recall, negative predictive value, prediction-conditioned fallout/miss.",
                                "Phi/Matthews correlation coefficient, mutual information, Chi-squared test statistic, odds ratio",
                                "F-measure, lift, precision-recall break-even point",
                                "ROC convex hull, area under the ROC curve",
                                "calibration error, mean cross-entropy, root mean-squared error",
                                "expected cost, explicit cost"))
knitr::kable(table,
             caption = "***Table 1:**Performance measures in the ROCR package*",
             col.names = c("",""),
             align = "l")

## -----------------------------------------------------------------------------
data(ROCR.hiv)
predictions <- ROCR.hiv$hiv.svm$predictions
labels <- ROCR.hiv$hiv.svm$labels
pred <- prediction(predictions, labels)
pred

## ----fig.asp=1, fig.width=5, fig.align='center'-------------------------------
perf <- performance(pred, "tpr", "fpr")
perf
plot(perf,
     avg="threshold",
     spread.estimate="boxplot")

## ----echo=FALSE, results='asis', fig.asp=0.35, fig.width=7, fig.align='center',fig.cap="***Fig 1:** Visualizations of classifier performance (HIV coreceptor usage data): (a) receiver operating characteristic (ROC) curve; (b) peak accuracy across a range of cutoffs; (c) absolute difference between empirical and predicted rate of positives for windowed cutoff ranges, in order to evaluate how well the scores are calibrated as probability estimates. Owing to the probabilistic interpretation, cutoffs need to be in the interval [0,1], in contrast to other performance plots. (d) Score density estimates for the negative (solid) and positive (dotted) class.*"----
data(ROCR.hiv)
pp.unnorm <- ROCR.hiv$hiv.svm$predictions
ll <- ROCR.hiv$hiv.svm$labels

# normalize predictions to 0..1
v <- unlist(pp.unnorm)
pp <- lapply(pp.unnorm, function(run) {approxfun(c(min(v), max(v)), c(0,1))(run)})

par(mfrow=c(1,4))
pred<- prediction(pp, ll)
perf <- performance(pred, "tpr", "fpr")

plot(perf, avg= "threshold", colorize=TRUE, lwd= 3,
     coloraxis.at=seq(0,1,by=0.2),)
plot(perf, col="gray78", add=TRUE)
plot(perf, avg= "threshold", colorize=TRUE, colorkey=FALSE,lwd= 3,,add=TRUE)
mtext(paste0("(a)"), side = 3, adj = 0.01,line = 1)

perf <- performance(pred, "acc")
plot(perf, avg= "vertical", spread.estimate="boxplot", lwd=3,col='blue',
     show.spread.at= seq(0.1, 0.9, by=0.1),)
mtext(paste0("(b)"), side = 3, adj = 0.01,line = 1)


plot(performance(pred, "cal", window.size= 10),
     avg="vertical",)
mtext(paste0("(c)"), side = 3, adj = 0.01,line = 1)

plot(0,0,type="n", xlim= c(0,1), ylim=c(0,7),
     xlab="Cutoff", ylab="Density",)
mtext(paste0("(d)"), side = 3, adj = 0.01,line = 1)
for (runi in 1:length(pred@predictions)) {
  lines(density(pred@predictions[[runi]][pred@labels[[runi]]=="-1"]), col= "red")
  lines(density(pred@predictions[[runi]][pred@labels[[runi]]=="1"]), col="green")
}

## ----fig.asp=1, fig.width=5, fig.align='center'-------------------------------
perf <- performance(pred, "tpr", "fpr")
plot(perf,
     avg= "threshold",
     colorize=TRUE,
     lwd= 3,
     main= "With ROCR you can produce standard plots\nlike ROC curves ...")
plot(perf,
     lty=3,
     col="grey78",
     add=TRUE)

## ----fig.asp=1, fig.width=5, fig.align='center'-------------------------------
perf <- performance(pred, "prec", "rec")
plot(perf,
     avg= "threshold",
     colorize=TRUE,
     lwd= 3,
     main= "... Precision/Recall graphs ...")
plot(perf,
     lty=3,
     col="grey78",
     add=TRUE)

## ----fig.asp=1, fig.width=5, fig.align='center'-------------------------------
perf <- performance(pred, "sens", "spec")
plot(perf,
     avg= "threshold",
     colorize=TRUE,
     lwd= 3,
     main="... Sensitivity/Specificity plots ...")
plot(perf,
     lty=3,
     col="grey78",
     add=TRUE)

## ----fig.asp=1, fig.width=5, fig.align='center'-------------------------------
perf <- performance(pred, "lift", "rpp")
plot(perf,
     avg= "threshold",
     colorize=TRUE,
     lwd= 3,
     main= "... and Lift charts.")
plot(perf,
     lty=3,
     col="grey78",
     add=TRUE)

## -----------------------------------------------------------------------------
data(ROCR.xval)
predictions <- ROCR.xval$predictions
labels <- ROCR.xval$labels
length(predictions)

## -----------------------------------------------------------------------------
pred <- prediction(predictions, labels)
perf <- performance(pred,'tpr','fpr')

## ----fig.asp=1, fig.width=5, fig.align='center'-------------------------------
plot(perf,
     colorize=TRUE,
     lwd=2,
     main='ROC curves from 10-fold cross-validation')

## ----fig.asp=1, fig.width=5, fig.align='center'-------------------------------
plot(perf,
     avg='vertical',
     spread.estimate='stderror',
     lwd=3,main='Vertical averaging + 1 standard error',
     col='blue')

## ----fig.asp=1, fig.width=5, fig.align='center'-------------------------------
plot(perf,
     avg='horizontal',
     spread.estimate='boxplot',
     lwd=3,
     main='Horizontal averaging + boxplots',
     col='blue')

## ----fig.asp=1, fig.width=5, fig.align='center'-------------------------------
plot(perf,
     avg='threshold',
     spread.estimate='stddev',
     lwd=2,
     main='Threshold averaging + 1 standard deviation',
     colorize=TRUE)

## ----fig.asp=1, fig.width=6, fig.align='center'-------------------------------
plot(perf,
     print.cutoffs.at=seq(0,1,by=0.2),
     text.cex=0.8,
     text.y=lapply(as.list(seq(0,0.5,by=0.05)), function(x) { rep(x,length(perf@x.values[[1]])) } ),
     col= as.list(terrain.colors(10)),
     text.col= as.list(terrain.colors(10)), 
     points.col= as.list(terrain.colors(10)), 
     main= "Cutoff stability")

## -----------------------------------------------------------------------------
perf <- performance(pred,"pcmiss","lift")

## ----fig.asp=1, fig.width=5, fig.align='center'-------------------------------
plot(perf,
     colorize=TRUE,
     print.cutoffs.at=seq(0,1,by=0.1),
     text.adj=c(1.2,1.2),
     avg="threshold",
     lwd=3,
     main= "You can freely combine performance measures ...")

## -----------------------------------------------------------------------------------
## Demo file for ROCR; start with 'demo(ROCR)'
## -----------------------------------------------------------------------------------

# if(dev.cur() <= 1) get(getOption("device"))()
if(dev.cur() <= 1) dev.new()

opar <- par(ask = interactive() &&
            (.Device %in% c("X11", "GTK", "gnome", "windows","quartz")))

data(ROCR.hiv)
pp <- ROCR.hiv$hiv.svm$predictions
ll <- ROCR.hiv$hiv.svm$labels

par(mfrow=c(2,2))
pred<- prediction(pp, ll)

perf <- performance(pred, "tpr", "fpr")
plot(perf, avg= "threshold", colorize=TRUE, lwd= 3,
     main= "With ROCR you can produce standard plots like ROC curves ...")
plot(perf, lty=3, col="grey78", add=TRUE)

perf <- performance(pred, "prec", "rec")
plot(perf, avg= "threshold", colorize=TRUE, lwd= 3,
     main= "... Precision/Recall graphs ...")
plot(perf, lty=3, col="grey78", add=TRUE)

perf <- performance(pred, "sens", "spec")
plot(perf, avg= "threshold", colorize=TRUE, lwd= 3,
     main="... Sensitivity/Specificity plots ...")
plot(perf, lty=3, col="grey78", add=TRUE)

perf <- performance(pred, "lift", "rpp")
plot(perf, avg= "threshold", colorize=TRUE, lwd= 3,
     main= "... and Lift charts.")
plot(perf, lty=3, col="grey78", add=TRUE)

# ------------------------------------------------------------------------------------

data(ROCR.xval)
pp <- ROCR.xval$predictions
ll <- ROCR.xval$labels
pred <- prediction(pp,ll)
perf <- performance(pred,'tpr','fpr')
par(mfrow=c(2,2))
plot(perf, colorize=TRUE, lwd=2,main='ROC curves from 10-fold cross-validation')
plot(perf, avg='vertical', spread.estimate='stderror',lwd=3,main='Vertical averaging + 1 standard error',col='blue')
plot(perf, avg='horizontal', spread.estimate='boxplot',lwd=3,main='Horizontal averaging + boxplots',col='blue')
plot(perf, avg='threshold', spread.estimate='stddev',lwd=2, main='Threshold averaging + 1 standard deviation',colorize=TRUE)


# ------------------------------------------------------------------------------------
data(ROCR.hiv)
pp.unnorm <- ROCR.hiv$hiv.svm$predictions
ll <- ROCR.hiv$hiv.svm$labels

# normalize predictions to 0..1
v <- unlist(pp.unnorm)
pp <- lapply(pp.unnorm, function(run) {approxfun(c(min(v), max(v)), c(0,1))(run)})

par(mfrow=c(2,2))
pred<- prediction(pp, ll)
perf <- performance(pred, "tpr", "fpr")

plot(perf, avg= "threshold", colorize=TRUE, lwd= 3,
     coloraxis.at=seq(0,1,by=0.2),
     main= "ROC curve")
plot(perf, col="gray78", add=TRUE)
plot(perf, avg= "threshold", colorize=TRUE, colorkey=FALSE,lwd= 3,
     main= "ROC curve",add=TRUE)

perf <- performance(pred, "acc")
plot(perf, avg= "vertical", spread.estimate="boxplot", lwd=3,col='blue',
     show.spread.at= seq(0.1, 0.9, by=0.1),
     main= "Accuracy across the range of possible cutoffs")


plot(performance(pred, "cal", window.size= 10),
     avg="vertical",
     main= "How well are the probability predictions calibrated?")

plot(0,0,type="n", xlim= c(0,1), ylim=c(0,7),
     xlab="Cutoff", ylab="Density",
     main="How well do the predictions separate the classes?")
for (runi in 1:length(pred@predictions)) {
  lines(density(pred@predictions[[runi]][pred@labels[[runi]]=="-1"]), col= "red")
  lines(density(pred@predictions[[runi]][pred@labels[[runi]]=="1"]), col="green")
}

#---------------------------------------------------------------------

par(mfrow= c(2,2))

# ...you can freely combine performance measures (pcmiss,lift)
data(ROCR.xval)
pred <- prediction(ROCR.xval$predictions, ROCR.xval$labels)
perf <- performance(pred,"pcmiss","lift")
# plot(perf, colorize=TRUE)
plot(perf, colorize=TRUE, print.cutoffs.at=seq(0,1,by=0.1), text.adj=c(1.2,1.2), avg="threshold", lwd=3,
     main= "You can freely combine performance measures ...")

data(ROCR.simple)
pred <- prediction(ROCR.simple$predictions, ROCR.simple$labels)
perf <- performance(pred,"tpr","fpr")
plot(perf, colorize=TRUE, colorkey.pos="top", print.cutoffs.at=seq(0,1,by=0.1), text.cex=1,
     text.adj=c(1.2, 1.2), lwd=2)

# ... cutoff stacking
data(ROCR.xval)
pred <- prediction(ROCR.xval$predictions, ROCR.xval$labels)
perf <- performance(pred,"tpr","fpr")
plot(perf,
     print.cutoffs.at=seq(0,1,by=0.2),
     text.cex=0.8,
     text.y=lapply(as.list(seq(0,0.5,by=0.05)), function(x) { rep(x,length(perf@x.values[[1]])) } ),
     col= as.list(terrain.colors(10)),
     text.col= as.list(terrain.colors(10)), 
     points.col= as.list(terrain.colors(10)), 
     main= "Cutoff stability")

# .... no functional dependencies needed, truly parametrized curve
data(ROCR.xval)
pred <- prediction(ROCR.xval$predictions, ROCR.xval$labels)
perf <- performance(pred,"acc","lift")
plot(perf, colorize=TRUE, main="Truly parametrized curves")
plot(perf, colorize=TRUE, print.cutoffs.at=seq(0,1,by=0.1), add=TRUE, text.adj=c(1.2, 1.2), avg="threshold", lwd=3)

# --------------------------------------------------------------------
# (Expected cost) curve + ROC convex hull

par(mfrow=c(1,2))
data(ROCR.hiv)
plot(0,0,xlim=c(0,1),ylim=c(0,1),xlab='Probability cost function',
     ylab="Normalized expected cost",
     main='HIV data: Expected cost curve (Drummond & Holte)')

pred<-prediction(ROCR.hiv$hiv.nn$predictions,ROCR.hiv$hiv.nn$labels)
lines(c(0,1),c(0,1))
lines(c(0,1),c(1,0))
perf1 <- performance(pred,'fpr','fnr')
for (i in 1:length(perf1@x.values)) {
    for (j in 1:length(perf1@x.values[[i]])) {
        lines(c(0,1),c(perf1@y.values[[i]][j], perf1@x.values[[i]][j]),col=rev(terrain.colors(10))[i],lty=3)
    }
}
perf<-performance(pred,'ecost')
plot(perf,lwd=1.5,xlim=c(0,1),ylim=c(0,1),add=TRUE)

# RCH
data(ROCR.simple)
ROCR.simple$labels[ROCR.simple$predictions >= 0.7 & ROCR.simple$predictions < 0.85] <- 0
  #as.numeric(!labels[predictions >= 0.6 & predictions < 0.85])
pred <- prediction(ROCR.simple$predictions, ROCR.simple$labels)
perf <- performance(pred,'tpr','fpr')
plot(perf, main="ROC curve with concavities (suboptimal) and ROC convex hull (Fawcett)")
perf1 <- performance(pred,'rch')
plot(perf1,add=TRUE,col='red',lwd=2)


#---------------------------------------------------------------------
# (plotting cutoff vs. measure)

data(ROCR.hiv)
pp <- ROCR.hiv$hiv.svm$predictions
ll <- ROCR.hiv$hiv.svm$labels

measures <- c('tpr','fpr','acc','err','rec','sens','fnr','tnr','spec',
                  'ppv','prec','npv','fall','miss','pcfall','pcmiss',
                  'phi','mat','mi','chisq','odds','lift','f')

## Don't be surprised by the decreased cutoff regions produced by 'odds ratio'.
## Cf. ?performance for details.

pred <- prediction(pp, ll)
par(mfrow=c(5,5))
for (measure in measures) {
    perf <- performance(pred, measure)
    plot(perf,avg="vertical",spread.estimate="boxplot")
}

#---------------------------------------------------------------------
measures <- c('tpr','err','prec','phi','mi','chisq','odds','lift','f')
par(mfrow=c(6,6))
for (i in 1:(length(measures)-1)) {
    for (j in (i+1):length(measures)) {
        perf <- performance(pred, measures[i], measures[j])
        plot(perf, avg="threshold", colorize=TRUE)
    }
}

#---------------------------------------------------------------------

data(ROCR.hiv)
pp <- ROCR.hiv$hiv.svm$predictions
ll <- ROCR.hiv$hiv.svm$labels

data(ROCR.xval)
pp <- ROCR.xval$predictions
ll <- ROCR.xval$labels
pred <- prediction(pp, ll)

par(mfrow=c(3,3))
perf <- performance(pred, "odds", "fpr")
plot(perf, colorize=TRUE)
plot(perf, avg="threshold", lwd=2, add=TRUE)

perf <- performance(pred, "phi", "err")
plot(perf, colorize=TRUE)
plot(perf, avg="threshold", lwd=2, add=TRUE)
perf <- performance(pred, "f", "err")
plot(perf, colorize=TRUE)
plot(perf, avg="threshold", lwd=2, add=TRUE)
perf <- performance(pred, "f", "ppv")
plot(perf, colorize=TRUE)
plot(perf, avg="threshold", lwd=2, add=TRUE)
perf <- performance(pred, "mat", "ppv")
plot(perf, colorize=TRUE)
plot(perf, avg="threshold", lwd=2, add=TRUE)
perf <- performance(pred, "npv", "ppv")
plot(perf, colorize=TRUE)
plot(perf, avg="threshold", lwd=2, add=TRUE)
perf <- performance(pred, "acc", "phi")
plot(perf, colorize=TRUE)
plot(perf, avg="threshold", lwd=2, add=TRUE)
perf <- performance(pred, "lift", "phi")
plot(perf, colorize=TRUE)
plot(perf, avg="threshold", lwd=2, add=TRUE)
perf <- performance(pred, "f", "phi")
plot(perf, colorize=TRUE)
plot(perf, avg="threshold", lwd=2, add=TRUE)

perf <- performance(pred, "mi", "phi")
plot(perf, colorize=TRUE)
plot(perf, avg="threshold", lwd=2, add=TRUE)
perf <- performance(pred, "chisq", "phi")
plot(perf, colorize=TRUE)
plot(perf, avg="threshold", lwd=2, add=TRUE)
perf <- performance(pred, "acc", "mi")
plot(perf, colorize=TRUE)
plot(perf, avg="threshold", lwd=2, add=TRUE)
perf <- performance(pred, "fall", "odds")
plot(perf, colorize=TRUE)
plot(perf, avg="threshold", lwd=2, add=TRUE)
perf <- performance(pred, "tpr", "lift")
plot(perf, colorize=TRUE)
plot(perf, avg="threshold", lwd=2, add=TRUE)
perf <- performance(pred, "fall", "lift")
plot(perf, colorize=TRUE)
plot(perf, avg="threshold", lwd=2, add=TRUE)
perf <- performance(pred, "npv", "f")
plot(perf, colorize=TRUE)
plot(perf, avg="threshold", lwd=2, add=TRUE)
perf <- performance(pred, "prec", "f")
plot(perf, colorize=TRUE)
plot(perf, avg="threshold", lwd=2, add=TRUE)
perf <- performance(pred, "tpr", "f")
plot(perf, colorize=TRUE)
plot(perf, avg="threshold", lwd=2, add=TRUE)

par(opar)

## ----------------------------------------------------------------------------
## plot method for objects of class 'performance'
## ----------------------------------------------------------------------------

#' @importFrom graphics plot.default plot.xy par
.get.arglist <- function( fname, arglist ) {
  if (fname=='plot')
    return(.select.args(arglist,
                        union(names(formals(graphics::plot.default)),
                              names(graphics::par()))))
  else if (fname=='plot.xy')
    return(.select.args(arglist,
                        union( names(formals(graphics::plot.xy)),
                               names(graphics::par()))))
  else return( .select.prefix( arglist, fname) )
}

.downsample <- function( perf, downsampling ) {
  for (i in 1:length(perf@alpha.values)) {
    if (downsampling < 1 && downsampling > 0)
      ind <- round(seq(1, length(perf@alpha.values[[i]]),
                       length=(length(perf@alpha.values[[i]]) *
                                 downsampling)))
    else if (downsampling > 1)
      ind <- round(seq(1, length(perf@alpha.values[[i]]),
                       length=downsampling))
    else ind <- 1:length(perf@alpha.values[[i]])
    perf@alpha.values[[i]] <- perf@alpha.values[[i]][ind]
    perf@x.values[[i]] <- perf@x.values[[i]][ind]
    perf@y.values[[i]] <- perf@y.values[[i]][ind]
  }
  return(perf)
}

.check_performance_for_plotting <- function(perf, colorize, print.cutoffs.at,
                                            avg){
  if (length(perf@y.values) != length(perf@x.values)) {
    stop("Performance object cannot be plotted. Length of x and y values ",
         "does not match.",
         call. = FALSE)
  }
  if ((is.null(perf@alpha.values) || length(perf@alpha.values) == 0L) && 
      (colorize==TRUE || length(print.cutoffs.at) > 0L)) {
    stop("Threshold coloring or labeling cannot be performed: ",
         "performance object has no threshold information.",
         call. = FALSE)
  }
  if ((avg=="vertical" || avg=="horizontal") &&
      (colorize==TRUE || length(print.cutoffs.at) > 0L)) {
    stop("Threshold coloring or labeling is only well-defined for",
         "'no' or 'threshold' averaging.",
         call. = FALSE)
  }
}

#' @importFrom grDevices rainbow
.plot.performance <-
  function(perf,
           ...,
           avg = "none",
           spread.estimate = "none",
           spread.scale = 1,
           show.spread.at = c(),
           colorize = FALSE,
           colorize.palette = rev(grDevices::rainbow(256, start = 0, end = 4 / 6)),
           colorkey = colorize,
           colorkey.relwidth = 0.25,
           colorkey.pos = "right",
           print.cutoffs.at = c(),
           cutoff.label.function = function(x) {
             round(x, 2)
           },
           downsampling = 0,
           add = FALSE) {
    # Input checks
    .check_performance_for_plotting(perf, colorize, print.cutoffs.at, avg)
    # getting the arguments
    arglist <- c(lapply(as.list(environment()), eval ), list(...) )

    if (downsampling >0 ) perf <- .downsample( perf, downsampling)

    ## for infinite cutoff, assign maximal finite cutoff + mean difference
    ## between adjacent cutoff pairs
    if (length(perf@alpha.values) != 0) {
      FUN <- function(x) {
        isfin <- is.finite(x)
        # if only one finite is available the mean cannot be calculated without
        # the first/last value, since the leaves no value
        if(sum(isfin) > 1L){ 
          inf_replace <- max(x[isfin]) + 
            mean(abs(x[isfin][-1] - x[isfin][-length(x[isfin])]))
        } else {
          inf_replace <- 0
        }
        x[is.infinite(x)] <- inf_replace
        x
      }
      perf@alpha.values <- lapply(perf@alpha.values,FUN)
    }
    ## remove samples with x or y not finite
    for (i in 1:length(perf@x.values)) {
      ind.bool <- (is.finite(perf@x.values[[i]]) &
                     is.finite(perf@y.values[[i]]))

      if (length(perf@alpha.values)>0)
        perf@alpha.values[[i]] <- perf@alpha.values[[i]][ind.bool]

      perf@x.values[[i]] <- perf@x.values[[i]][ind.bool]
      perf@y.values[[i]] <- perf@y.values[[i]][ind.bool]
    }
    arglist <- .sarg( arglist, perf=perf)

    if (add==FALSE) do.call( ".performance.plot.canvas", arglist )

    if (avg=="none") do.call(".performance.plot.no.avg", arglist)
    else if (avg=="vertical")
      do.call(".performance.plot.vertical.avg", arglist)
    else if (avg=="horizontal")
      do.call(".performance.plot.horizontal.avg", arglist)
    else if (avg=="threshold")
      do.call(".performance.plot.threshold.avg", arglist)
  }

## ---------------------------------------------------------------------------
## initializing plots and plotting a canvas
## (can be skipped using 'plot( ..., add=TRUE)'
## ---------------------------------------------------------------------------

#' @import stats
#' @import graphics
.performance.plot.canvas <- function(perf, avg, ...) {

  # requireNamespace("stats")
  # requireNamespace("graphics")

  arglist <- list(...)

  axis.names <- list(x=perf@x.name, y=perf@y.name)
  if (avg=="horizontal" || avg=="threshold")
    axis.names$x <- paste("Average", tolower(axis.names$x))
  if (avg=="vertical" || avg=="threshold")
    axis.names$y <- paste("Average", tolower(axis.names$y))
  arglist <- .farg(arglist, xlab=axis.names$x, ylab=axis.names$y)

  arglist <-
    .farg(arglist,
          xlim=c(min(unlist(perf@x.values)), max(unlist(perf@x.values))),
          ylim=c(min(unlist(perf@y.values)), max(unlist(perf@y.values))))

  do.call("plot", .sarg(.slice.run(.get.arglist('plot', arglist)),
                        x=0.5, y=0.5, type='n', axes=FALSE))
  do.call( "axis", .sarg(.slice.run(.get.arglist('xaxis', arglist)),
                         side=1))
  do.call( "axis", .sarg(.slice.run(.get.arglist('yaxis', arglist)),
                         side=2))

  if (.garg(arglist,'colorkey')==TRUE) {
    colors <- rev( .garg(arglist,'colorize.palette') )
    max.alpha <- max(unlist(perf@alpha.values))
    min.alpha <- min(unlist(perf@alpha.values))
    col.cutoffs <- rev(seq(min.alpha,max.alpha, length=length( colors )))

    if ( .garg(arglist,'colorkey.pos')=="right") {

      ## axis drawing (ticks + labels)
      ## The interval [min.alpha,max.alpha] needs to be mapped onto
      ## the interval [min.y,max.y], rather than onto the interval
      ## [ylim[1],ylim[2]] ! In the latter case, NAs could occur in
      ## approxfun below, because axTicks can be out of the ylim-range
      ## ('yxaxs': 4%region)
      max.y <- max(axTicks(4))
      min.y <- min(axTicks(4))
      alpha.ticks <- .garg( arglist, c("coloraxis.at"))
      if (length(alpha.ticks)==0)
        alpha.ticks <- approxfun(c(min.y, max.y),
                                 c(min.alpha, max.alpha)) (axTicks(4))
      alpha2y <- approxfun(c(min(alpha.ticks), max(alpha.ticks)),
                           c(min.y,max.y))
      arglist <-
        .sarg(arglist,
              coloraxis.labels=.garg(arglist,
                                     'cutoff.label.function')(alpha.ticks),
              coloraxis.at=alpha2y(alpha.ticks))

      do.call("axis",
              .sarg(.slice.run(.get.arglist('coloraxis', arglist)),
                    side=4))

      ## draw colorkey
      ## each entry in display.bool corresponds to one rectangle of
      ## the colorkey.
      ## Only rectangles within the alpha.ticks range are plotted.
      ## y.lower, y.upper, and colors, are the attributes of the visible
      ## rectangles (those for which display.bool=TRUE)
      display.bool <- (col.cutoffs >= min(alpha.ticks) &
                         col.cutoffs < max(alpha.ticks))
      y.lower <- alpha2y( col.cutoffs )[display.bool]
      colors <- colors[display.bool]
      if (length(y.lower>=2)) {
        y.width <- y.lower[2] - y.lower[1]
        y.upper <- y.lower + y.width
        x.left <- .garg(arglist,'xlim')[2] +
          ((.garg(arglist,'xlim')[2] - .garg(arglist,'xlim')[1]) *
             (1-.garg(arglist,'colorkey.relwidth'))*0.04)
        x.right <- .garg(arglist,'xlim')[2] +
          (.garg(arglist,'xlim')[2] -.garg(arglist,'xlim')[1]) * 0.04
        rect(x.left, y.lower, x.right, y.upper,
             col=colors, border=colors,xpd=NA)
      }
    } else if (.garg(arglist, 'colorkey.pos') == "top") {
      ## axis drawing (ticks + labels)
      max.x <- max(axTicks(3))
      min.x <- min(axTicks(3))
      alpha.ticks <- .garg( arglist, c("coloraxis.at"))
      if (length(alpha.ticks)==0) {
        alpha.ticks <- approxfun(c(min.x, max.x),
                                 c(min.alpha, max.alpha))(axTicks(3))
      }
      alpha2x <- approxfun(c( min(alpha.ticks), max(alpha.ticks)),
                           c( min.x, max.x))
      arglist <- .sarg(arglist,
                       coloraxis.labels=.garg(arglist,
                                              'cutoff.label.function')(alpha.ticks),
                       coloraxis.at= alpha2x(alpha.ticks))
      do.call("axis",
              .sarg(.slice.run( .get.arglist('coloraxis', arglist)),
                    side=3))

      ## draw colorkey
      display.bool <- (col.cutoffs >= min(alpha.ticks) &
                         col.cutoffs < max(alpha.ticks))
      x.left <- alpha2x( col.cutoffs )[display.bool]
      colors <- colors[display.bool]
      if (length(x.left)>=2) {
        x.width <- x.left[2] - x.left[1]
        x.right <- x.left + x.width
        y.lower <- .garg(arglist,'ylim')[2] +
          (.garg(arglist,'ylim')[2] - .garg(arglist,'ylim')[1]) *
          (1-.garg(arglist,'colorkey.relwidth'))*0.04
        y.upper <- .garg(arglist,'ylim')[2] +
          (.garg(arglist,'ylim')[2] - .garg(arglist,'ylim')[1]) * 0.04
        rect(x.left, y.lower, x.right, y.upper,
             col=colors, border=colors, xpd=NA)
      }
    }
  }

  do.call( "box", .slice.run( .get.arglist( 'box', arglist)))
}

## ----------------------------------------------------------------------------
## plotting performance objects when no curve averaging is wanted
## ----------------------------------------------------------------------------

#' @importFrom grDevices xy.coords
#' @importFrom stats approxfun
.performance.plot.no.avg <- function( perf, ... ) {

  arglist <- list(...)
  arglist <- .farg(arglist, type= 'l')

  if (.garg(arglist, 'colorize') == TRUE) {
    colors <- rev( .garg( arglist, 'colorize.palette') )
    max.alpha <- max(unlist(perf@alpha.values))
    min.alpha <- min(unlist(perf@alpha.values))
    col.cutoffs <- rev(seq(min.alpha,max.alpha, length=length(colors)+1))
    col.cutoffs <- col.cutoffs[2:length(col.cutoffs)]
  }

  for (i in 1:length(perf@x.values)) {
    if (.garg(arglist, 'colorize') == FALSE) {
      do.call("plot.xy",
              .sarg(.slice.run(.get.arglist('plot.xy', arglist), i),
                    xy=(grDevices::xy.coords(perf@x.values[[i]],
                                             perf@y.values[[i]]))))
    } else {
      for (j in 1:(length(perf@x.values[[i]])-1)) {
        segment.coloring <-
          colors[min(which(col.cutoffs <= perf@alpha.values[[i]][j]))]
        do.call("plot.xy",
                .sarg(.slice.run(.get.arglist('plot.xy', arglist), i),
                      xy=(grDevices::xy.coords(perf@x.values[[i]][j:(j+1)],
                                               perf@y.values[[i]][j:(j+1)])),
                      col= segment.coloring))
      }
    }

    print.cutoffs.at <- .garg(arglist, 'print.cutoffs.at',i)
    if (! is.null(print.cutoffs.at)) {
      text.x <- stats::approxfun(perf@alpha.values[[i]], perf@x.values[[i]],
                                 rule=2, ties=mean)(print.cutoffs.at)
      text.y <- stats::approxfun(perf@alpha.values[[i]], perf@y.values[[i]],
                                 rule=2, ties=mean)(print.cutoffs.at)
      do.call("points",
              .sarg(.slice.run(.get.arglist('points', arglist),i),
                    x= text.x,
                    y= text.y))
      do.call("text",
              .farg(.slice.run( .get.arglist('text', arglist),i),
                    x= text.x,
                    y= text.y,
                    labels=(.garg(arglist,
                                  'cutoff.label.function',
                                  i)(print.cutoffs.at))))
    }
  }
}

## ----------------------------------------------------------------------------
## plotting performance objects when vertical curve averaging is wanted
## ----------------------------------------------------------------------------

#' @importFrom stats approxfun sd
.performance.plot.vertical.avg <- function( perf, ...) {
  arglist <- list(...)
  arglist <- .farg(arglist,
                   show.spread.at= (seq(min(unlist(perf@x.values)),
                                        max(unlist(perf@x.values)),
                                        length=11)))
  perf.avg <- perf
  x.values <- seq(min(unlist(perf@x.values)), max(unlist(perf@x.values)),
                  length=max( sapply(perf@x.values, length)))
  for (i in 1:length(perf@y.values)) {
    perf.avg@y.values[[i]] <-
      stats::approxfun(perf@x.values[[i]], perf@y.values[[i]],
                       ties=mean, rule=2)(x.values)
  }
  perf.avg@y.values <- list(rowMeans( data.frame( perf.avg@y.values )))
  perf.avg@x.values <- list(x.values)
  perf.avg@alpha.values <- list()

  ## y.values at show.spread.at (midpoint of error bars )
  show.spread.at.y.values <-
    lapply(as.list(1:length(perf@x.values)),
           function(i) {
             stats::approxfun(perf@x.values[[i]], perf@y.values[[i]],
                              rule=2,
                              ties=mean)( .garg(arglist, 'show.spread.at'))
           })

  show.spread.at.y.values <- as.matrix(data.frame(show.spread.at.y.values ))
  colnames(show.spread.at.y.values) <- c()
  ## now, show.spread.at.y.values[i,] contains the curve y values at the
  ## sampling x value .garg(arglist,'show.spread.at')[i]

  if (.garg(arglist, 'spread.estimate') == "stddev" ||
      .garg(arglist, 'spread.estimate') == "stderror") {
    bar.width <- apply(show.spread.at.y.values, 1, stats::sd)
    if (.garg(arglist, 'spread.estimate') == "stderror") {
      bar.width <- bar.width / sqrt( ncol(show.spread.at.y.values) )
    }
    bar.width <- .garg(arglist, 'spread.scale') * bar.width

    suppressWarnings(do.call(gplots::plotCI,
                             .farg(.sarg(.get.arglist('plotCI', arglist),
                                         x=.garg(arglist,
                                                 'show.spread.at'),
                                         y=rowMeans(
                                           show.spread.at.y.values),
                                         uiw= bar.width,
                                         liw= bar.width,
                                         err= 'y',
                                         add= TRUE),
                                   gap= 0,
                                   type= 'n')))
  }

  if (.garg(arglist, 'spread.estimate') == "boxplot") {
    do.call("boxplot",
            .farg(.sarg(.get.arglist( 'boxplot', arglist),
                        x= data.frame(t(show.spread.at.y.values)),
                        at= .garg(arglist, 'show.spread.at'),
                        add= TRUE,
                        axes= FALSE),
                  boxwex= (1/(2*(length(.garg(arglist,
                                              'show.spread.at')))))))
    do.call("points",
            .sarg(.get.arglist( 'points', arglist),
                  x= .garg(arglist, 'show.spread.at'),
                  y= rowMeans(show.spread.at.y.values)))
  }

  do.call( ".plot.performance", .sarg(arglist,
                                      perf= perf.avg,
                                      avg= 'none',
                                      add= TRUE))
}

## ----------------------------------------------------------------------------
## plotting performance objects when horizontal curve averaging is wanted
## ----------------------------------------------------------------------------

#' @importFrom stats approxfun sd
.performance.plot.horizontal.avg <- function( perf, ...) {
  arglist <- list(...)
  arglist <- .farg(arglist,
                   show.spread.at= seq(min(unlist(perf@y.values)),
                                       max(unlist(perf@y.values)),
                                       length=11))
  perf.avg <- perf
  y.values <- seq(min(unlist(perf@y.values)), max(unlist(perf@y.values)),
                  length=max( sapply(perf@y.values, length)))
  for (i in 1:length(perf@x.values)) {
    perf.avg@x.values[[i]] <- stats::approxfun(perf@y.values[[i]],
                                               perf@x.values[[i]],
                                               ties=mean, rule=2)(y.values)
  }
  perf.avg@x.values <- list(rowMeans( data.frame( perf.avg@x.values )))
  perf.avg@y.values <- list(y.values)
  perf.avg@alpha.values <- list()

  ## x.values at show.spread.at (midpoint of error bars )
  show.spread.at.x.values <-
    lapply(as.list(1:length(perf@y.values)),
           function(i) {
             stats::approxfun(perf@y.values[[i]],
                              perf@x.values[[i]],
                              rule=2, ties=mean)(.garg(arglist,'show.spread.at'))
           } )
  show.spread.at.x.values <- as.matrix(data.frame(show.spread.at.x.values))
  colnames(show.spread.at.x.values) <- c()
  ## now, show.spread.at.x.values[i,] contains the curve x values at the
  ## sampling y value .garg(arglist,'show.spread.at')[i]

  if (.garg(arglist,'spread.estimate') == 'stddev' ||
      .garg(arglist,'spread.estimate') == 'stderror') {
    bar.width <- apply(show.spread.at.x.values, 1, stats::sd)
    if (.garg(arglist,'spread.estimate')== 'stderror') {
      bar.width <- bar.width / sqrt( ncol(show.spread.at.x.values) )
    }
    bar.width <- .garg(arglist,'spread.scale') * bar.width

    suppressWarnings(do.call(gplots::plotCI,
                             .farg(.sarg(.get.arglist('plotCI', arglist),
                                         x= rowMeans(
                                           show.spread.at.x.values),
                                         y= .garg(arglist,
                                                  'show.spread.at'),
                                         uiw= bar.width,
                                         liw= bar.width,
                                         err= 'x',
                                         add= TRUE),
                                   gap= 0,
                                   type= 'n')))
  }

  if (.garg(arglist,'spread.estimate') == "boxplot") {
    do.call("boxplot",
            .farg(.sarg(.get.arglist( 'boxplot', arglist),
                        x= data.frame(t(show.spread.at.x.values)),
                        at= .garg(arglist,'show.spread.at'),
                        add= TRUE,
                        axes= FALSE,
                        horizontal= TRUE),
                  boxwex= 1/(2*(length(.garg(arglist,'show.spread.at'))))))
    do.call("points", .sarg(.get.arglist( 'points', arglist),
                            x= rowMeans(show.spread.at.x.values),
                            y= .garg(arglist,'show.spread.at')))
  }

  do.call( ".plot.performance", .sarg(arglist,
                                      perf= perf.avg,
                                      avg= 'none',
                                      add= TRUE))
}

## ----------------------------------------------------------------------------
## plotting performance objects when threshold curve averaging is wanted
## ----------------------------------------------------------------------------

#' @importFrom stats approxfun sd
.performance.plot.threshold.avg <- function( perf, ...) {
  arglist <- list(...)
  arglist <- .farg(arglist,
                   show.spread.at= seq(min(unlist(perf@x.values)),
                                       max(unlist(perf@x.values)),
                                       length=11))

  perf.sampled <- perf
  alpha.values <- rev(seq(min(unlist(perf@alpha.values)),
                          max(unlist(perf@alpha.values)),
                          length=max( sapply(perf@alpha.values, length))))
  for (i in 1:length(perf.sampled@y.values)) {
    perf.sampled@x.values[[i]] <-
      stats::approxfun(perf@alpha.values[[i]],perf@x.values[[i]],
                       rule=2, ties=mean)(alpha.values)
    perf.sampled@y.values[[i]] <-
      stats::approxfun(perf@alpha.values[[i]], perf@y.values[[i]],
                       rule=2, ties=mean)(alpha.values)
  }

  ## compute average curve
  perf.avg <- perf.sampled
  perf.avg@x.values <- list( rowMeans( data.frame( perf.avg@x.values)))
  perf.avg@y.values <- list(rowMeans( data.frame( perf.avg@y.values)))
  perf.avg@alpha.values <- list( alpha.values )

  x.values.spread <-
    lapply(as.list(1:length(perf@x.values)),
           function(i) {
             stats::approxfun(perf@alpha.values[[i]], perf@x.values[[i]],
                              rule=2, ties=mean)(.garg(arglist,'show.spread.at'))
           } )
  x.values.spread <- as.matrix(data.frame( x.values.spread ))
  y.values.spread <-
    lapply(as.list(1:length(perf@y.values)),
           function(i) {
             stats::approxfun(perf@alpha.values[[i]], perf@y.values[[i]],
                              rule=2, ties=mean)(.garg(arglist,'show.spread.at'))
           } )
  y.values.spread <- as.matrix(data.frame( y.values.spread ))

  if (.garg(arglist,'spread.estimate')=="stddev" ||
      .garg(arglist,'spread.estimate')=="stderror") {

    x.bar.width <- apply(x.values.spread, 1, stats::sd)
    y.bar.width <- apply(y.values.spread, 1, stats::sd)
    if (.garg(arglist,'spread.estimate')=="stderror") {
      x.bar.width <- x.bar.width / sqrt( ncol(x.values.spread) )
      y.bar.width <- y.bar.width / sqrt( ncol(x.values.spread) )
    }
    x.bar.width <- .garg(arglist,'spread.scale') * x.bar.width
    y.bar.width <- .garg(arglist,'spread.scale') * y.bar.width

    suppressWarnings( do.call(gplots::plotCI,
                              .farg(.sarg(.get.arglist('plotCI', arglist),
                                          x= rowMeans(x.values.spread),
                                          y= rowMeans(y.values.spread),
                                          uiw= x.bar.width,
                                          liw= x.bar.width,
                                          err= 'x',
                                          add= TRUE),
                                    gap= 0,
                                    type= 'n')))

    suppressWarnings( do.call(gplots::plotCI,
                              .farg(.sarg(.get.arglist('plotCI', arglist),
                                          x= rowMeans(x.values.spread),
                                          y= rowMeans(y.values.spread),
                                          uiw= y.bar.width,
                                          liw= y.bar.width,
                                          err= 'y',
                                          add= TRUE),
                                    gap= 0,
                                    type= 'n')))
  }

  if (.garg(arglist,'spread.estimate')=="boxplot") {
    do.call("boxplot",
            .farg(.sarg(.get.arglist('boxplot', arglist),
                        x= data.frame(t(x.values.spread)),
                        at= rowMeans(y.values.spread),
                        add= TRUE,
                        axes= FALSE,
                        horizontal= TRUE),
                  boxwex= 1/(2*(length(.garg(arglist,'show.spread.at'))))))
    do.call("boxplot",
            .farg(.sarg(.get.arglist('boxplot', arglist),
                        x= data.frame(t(y.values.spread)),
                        at= rowMeans(x.values.spread),
                        add= TRUE,
                        axes= FALSE),
                  boxwex= 1/(2*(length(.garg(arglist,'show.spread.at'))))))
    do.call("points", .sarg(.get.arglist('points', arglist),
                            x= rowMeans(x.values.spread),
                            y= rowMeans(y.values.spread)))
  }

  do.call( ".plot.performance", .sarg(arglist,
                                      perf= perf.avg,
                                      avg= 'none',
                                      add= TRUE))
}

#' @import methods
NULL

#' @name prediction-class
#' @aliases prediction-class
#'
#' @title Class \code{prediction}
#'
#' @description
#' Object to encapsulate numerical predictions together with the
#' corresponding true class labels, optionally collecting predictions and
#' labels for several cross-validation or bootstrapping runs.
#'
#' @section Objects from the Class:
#' Objects can be created by using the \code{prediction} function.
#'
#' @note
#' Every \code{prediction} object contains information about the 2x2
#' contingency table consisting of tp,tn,fp, and fn, along with the
#' marginal sums n.pos,n.neg,n.pos.pred,n.neg.pred, because these form
#' the basis for many derived performance measures.
#'
#' @slot predictions A list, in which each element is a vector of predictions
#'   (the list has length > 1 for x-validation data.
#' @slot labels Analogously, a list in which each element is a vector of true
#'   class labels.
#' @slot cutoffs A list in which each element is a vector of all necessary
#'   cutoffs. Each cutoff vector consists of the predicted scores (duplicates
#'   removed), in descending order.
#' @slot fp A list in which each element is a vector of the number (not the
#'   rate!) of false positives induced by the cutoffs given in the corresponding
#'   'cutoffs' list entry.
#' @slot tp As fp, but for true positives.
#' @slot tn As fp, but for true negatives.
#' @slot fn As fp, but for false negatives.
#' @slot n.pos A list in which each element contains the number of positive
#'   samples in the given x-validation run.
#' @slot n.neg As n.pos, but for negative samples.
#' @slot n.pos.pred A list in which each element is a vector of the number of
#'   samples predicted as positive at the cutoffs given in the corresponding
#'   'cutoffs' entry.
#' @slot n.neg.pred As n.pos.pred, but for negatively predicted samples.
#'
#' @references
#' A detailed list of references can be found on the ROCR homepage at
#' \url{https://ipa-tys.github.io/ROCR/}.
#'
#' @author
#' Tobias Sing \email{tobias.sing@gmail.com}, Oliver Sander
#' \email{osander@gmail.com}
#'
#' @seealso
#' \code{\link{prediction}},
#' \code{\link{performance}},
#' \code{\link{performance-class}},
#' \code{\link{plot.performance}}
#'
#' @export
setClass("prediction",
         representation(predictions = "list",
                        labels      = "list",
                        cutoffs     = "list",
                        fp          = "list",
                        tp          = "list",
                        tn          = "list",
                        fn          = "list",
                        n.pos       = "list",
                        n.neg       = "list",
                        n.pos.pred  = "list",
                        n.neg.pred  = "list"))

setMethod("show","prediction",
          function(object){
              cat("A ", class(object), " instance\n", sep = "")
              if(length(object@predictions) > 1L){
                  cat("  with ", length(object@predictions)," cross ",
                      "validation runs ", sep = "")
                  if(length(unique(vapply(object@predictions,length,integer(1))))){
                      cat("(equal lengths)", sep = "")
                  } else {
                      cat("(different lengths)", sep = "")
                  }
              } else {
                  cat("  with ", length(object@predictions[[1L]]),
                      " data points", sep = "")
              }
          })

#' @name performance-class
#' @aliases performance-class
#'
#' @title Class \code{performance}
#'
#' @description
#' Object to capture the result of a performance evaluation, optionally
#' collecting evaluations from several cross-validation or bootstrapping runs.
#'
#' @section Objects from the Class:
#' Objects can be created by using the \code{performance} function.
#'
#' @details
#' A \code{performance} object can capture information from four
#' different evaluation scenarios:
#'   \itemize{
#'     \item The behaviour of a cutoff-dependent performance measure across
#'     the range of all cutoffs (e.g. \code{performance( predObj, 'acc' )} ). Here,
#'     \code{x.values} contains the cutoffs, \code{y.values} the
#'     corresponding values of the performance measure, and
#'     \code{alpha.values} is empty.\cr
#'     \item The trade-off between two performance measures across the
#'     range of all cutoffs (e.g. \code{performance( predObj,
#'                                                   'tpr', 'fpr' )} ). In this case, the cutoffs are stored in
#'     \code{alpha.values}, while \code{x.values} and \code{y.values}
#'     contain the corresponding values of the two performance measures.\cr
#'     \item A performance measure that comes along with an obligatory
#'     second axis (e.g. \code{performance( predObj, 'ecost' )} ). Here, the measure values are
#'     stored in \code{y.values}, while the corresponding values of the
#'     obligatory axis are stored in \code{x.values}, and \code{alpha.values}
#'     is empty.\cr
#'     \item A performance measure whose value is just a scalar
#'     (e.g. \code{performance( predObj, 'auc' )} ). The value is then stored in
#'     \code{y.values}, while \code{x.values} and \code{alpha.values} are
#'     empty.
#'   }
#'
#' @slot x.name Performance measure used for the x axis.
#' @slot y.name Performance measure used for the y axis.
#' @slot alpha.name Name of the unit that is used to create the parametrized
#'   curve. Currently, curves can only be parametrized by cutoff, so
#'   \code{alpha.name} is either \code{none} or \code{cutoff}.
#' @slot x.values A list in which each entry contains the x values of the curve
#'   of this particular cross-validation run. \code{x.values[[i]]},
#'   \code{y.values[[i]]}, and \code{alpha.values[[i]]} correspond to each
#'   other.
#' @slot y.values A list in which each entry contains the y values of the curve
#'   of this particular cross-validation run.
#' @slot alpha.values A list in which each entry contains the cutoff values of
#'   the curve of this particular cross-validation run.
#'
#' @references
#' A detailed list of references can be found on the ROCR homepage at
#' \url{https://ipa-tys.github.io/ROCR/}.
#'
#' @author
#' Tobias Sing \email{tobias.sing@gmail.com}, Oliver Sander
#' \email{osander@gmail.com}
#'
#' @seealso
#' \code{\link{prediction}}
#' \code{\link{performance}},
#' \code{\link{prediction-class}},
#' \code{\link{plot.performance}}
#'
#' @export
setClass("performance",
         representation(x.name       = "character",
                        y.name       = "character",
                        alpha.name   = "character",
                        x.values     = "list",
                        y.values     = "list",
                        alpha.values = "list" ))

setMethod("show","performance",
          function(object){
              cat("A ", class(object), " instance\n", sep = "")
              if(length(object@y.values[[1L]]) > 1L){
                  cat("  '", object@x.name, "' vs. '", object@y.name,
                      "' (alpha: '",object@alpha.name,"')\n", sep = "")
              } else {
                  cat("  '", object@y.name, "'\n", sep = "")
              }
              if(length(object@y.values) > 1L){
                  cat("  for ", length(object@y.values)," cross ",
                      "validation runs ", sep = "")
              } else {
                  if(length(object@y.values[[1L]]) > 1L){
                      cat("  with ", length(object@y.values[[1L]])," data points",
                          sep = "")
                  }
              }
          })

#' @name plot-methods
#' @aliases plot,performance,missing-method plot.performance
#'
#' @title Plot method for performance objects
#'
#' @description
#' This is the method to plot all objects of class performance.
#'
#' @param x an object of class \code{performance}
#' @param y not used
#' @param ... Optional graphical parameters to adjust different components of
#'   the performance plot. Parameters are directed to their target component by
#'   prefixing them with the name of the component (\code{component.parameter},
#'   e.g. \code{text.cex}). The following components are available:
#'   \code{xaxis}, \code{yaxis}, \code{coloraxis}, \code{box} (around the
#'   plotting region), \code{points}, \code{text}, \code{plotCI} (error bars),
#'   \code{boxplot}. The names of these components are influenced by the R
#'   functions that are used to create them. Thus, \code{par(component)} can be
#'   used to see which parameters are available for a given component (with the
#'   expection of the three axes; use \code{par(axis)} here). To adjust the
#'   canvas or the performance curve(s), the standard \code{plot} parameters can
#'   be used without any prefix.
#' @param avg If the performance object describes several curves (from
#'   cross-validation runs or bootstrap evaluations of one particular method),
#'   the curves from each of the runs can be averaged. Allowed values are
#'   \code{none} (plot all curves separately), \code{horizontal} (horizontal
#'   averaging), \code{vertical} (vertical averaging), and \code{threshold}
#'   (threshold (=cutoff) averaging). Note that while threshold averaging is
#'   always feasible, vertical and horizontal averaging are not well-defined if
#'   the graph cannot be represented as a function x->y and y->x, respectively.
#' @param spread.estimate When curve averaging is enabled, the variation around
#'   the average curve can be visualized as standard error bars
#'   (\code{stderror}), standard deviation bars (\code{stddev}), or by using box
#'   plots (\code{boxplot}). Note that the function \code{plotCI}, which is used
#'   internally by ROCR to draw error bars, might raise a warning if the spread
#'   of the curves at certain positions is 0.
#' @param spread.scale For \code{stderror} or \code{stddev}, this is a scalar
#'   factor to be multiplied with the length of the standard error/deviation
#'   bar. For example, under normal assumptions, \code{spread.scale=2} can be
#'   used to get approximate 95\% confidence intervals.
#' @param show.spread.at For vertical averaging, this vector determines the x
#'   positions for which the spread estimates should be visualized. In contrast,
#'   for horizontal and threshold averaging, the y positions and cutoffs are
#'   determined, respectively. By default, spread estimates are shown at 11
#'   equally spaced positions.
#' @param colorize This logical determines whether the curve(s) should be
#'   colorized according to cutoff.
#' @param colorize.palette If curve colorizing is enabled, this determines the
#'   color palette onto which the cutoff range is mapped.
#' @param colorkey If true, a color key is drawn into the 4\% border
#'   region (default of \code{par(xaxs)} and \code{par(yaxs)}) of the
#'   plot. The color key visualizes the mapping from cutoffs to colors.
#' @param colorkey.relwidth Scalar between 0 and 1 that determines the
#'   fraction of the 4\% border region that is occupied by the colorkey.
#' @param colorkey.pos Determines if the colorkey is drawn vertically at
#'   the \code{right} side, or horizontally at the \code{top} of the
#'   plot.
#' @param print.cutoffs.at This vector specifies the cutoffs which should
#'   be printed as text along the curve at the corresponding curve positions.
#' @param cutoff.label.function By default, cutoff annotations along the curve
#'   or at the color key are rounded to two decimal places before printing.
#'   Using a custom \code{cutoff.label.function}, any other transformation can
#'   be performed on the cutoffs instead (e.g. rounding with different precision
#'   or taking the logarithm).
#' @param downsampling ROCR can efficiently compute most performance measures
#'   even for data sets with millions of elements. However, plotting of large
#'   data sets can be slow and lead to PS/PDF documents of considerable size. In
#'   that case, performance curves that are indistinguishable from the original
#'   can be obtained by using only a fraction of the computed performance
#'   values. Values for downsampling between 0 and 1 indicate the fraction of
#'   the original data set size to which the performance object should be
#'   downsampled, integers above 1 are interpreted as the actual number of
#'   performance values to which the curve(s) should be downsampled.
#' @param add If \code{TRUE}, the curve(s) is/are added to an already existing
#'   plot; otherwise a new plot is drawn.
#'
#' @references
#' A detailed list of references can be found on the ROCR homepage at
#' \url{https://ipa-tys.github.io/ROCR/}.
#'
#' @author
#' Tobias Sing \email{tobias.sing@gmail.com}, Oliver Sander
#' \email{osander@gmail.com}
#'
#' @seealso
#' \code{\link{prediction}},
#' \code{\link{performance}},
#' \code{\link{prediction-class}},
#' \code{\link{performance-class}}
#'
#' @export
#'
#' @examples
#' # plotting a ROC curve:
#' library(ROCR)
#' data(ROCR.simple)
#' pred <- prediction( ROCR.simple$predictions, ROCR.simple$labels )
#' pred
#' perf <- performance( pred, "tpr", "fpr" )
#' perf
#' plot( perf )
#'
#' # To entertain your children, make your plots nicer
#' # using ROCR's flexible parameter passing mechanisms
#' # (much cheaper than a finger painting set)
#' par(bg="lightblue", mai=c(1.2,1.5,1,1))
#' plot(perf, main="ROCR fingerpainting toolkit", colorize=TRUE,
#'      xlab="Mary's axis", ylab="", box.lty=7, box.lwd=5,
#'      box.col="gold", lwd=17, colorkey.relwidth=0.5, xaxis.cex.axis=2,
#'      xaxis.col='blue', xaxis.col.axis="blue", yaxis.col='green', yaxis.cex.axis=2,
#'      yaxis.at=c(0,0.5,0.8,0.85,0.9,1), yaxis.las=1, xaxis.lwd=2, yaxis.lwd=3,
#'      yaxis.col.axis="orange", cex.lab=2, cex.main=2)
setMethod("plot",
          signature(x="performance",y="missing"),
          function(x,
                   y,
                   ...,
                   avg = "none",
                   spread.estimate = "none",
                   spread.scale = 1,
                   show.spread.at = c(),
                   colorize = FALSE,
                   colorize.palette = rev(rainbow(256, start = 0, end = 4 / 6)),
                   colorkey = colorize,
                   colorkey.relwidth = 0.25,
                   colorkey.pos = "right",
                   print.cutoffs.at = c(),
                   cutoff.label.function = function(x) {
                     round(x, 2)
                   },
                   downsampling = 0,
                   add = FALSE ) {

            .plot.performance(x,...,
                              avg = avg,
                              spread.estimate = spread.estimate,
                              spread.scale = spread.scale,
                              show.spread.at = show.spread.at,
                              colorize = colorize,
                              colorize.palette = colorize.palette,
                              colorkey = colorkey,
                              colorkey.relwidth = colorkey.relwidth,
                              colorkey.pos = colorkey.pos,
                              print.cutoffs.at = print.cutoffs.at,
                              cutoff.label.function = cutoff.label.function,
                              downsampling = downsampling,
                              add = add)
          }
)

#' @rdname plot-methods
#' @method plot performance
#' @export
"plot.performance" <- function(...) plot(...)

#' @name ROCR.hiv
#'
#' @docType data
#' @keywords datasets
#'
#' @title Data set: Support vector machines and neural networks applied to the
#'   prediction of HIV-1 coreceptor usage
#'
#' @description
#' Linear support vector machines (libsvm) and neural networks (R package
#' nnet) were applied to predict usage of the coreceptors CCR5 and CXCR4
#' based on sequence data of the third variable loop of the HIV envelope
#' protein.
#'
#' @format
#' A list consisting of the SVM (\code{ROCR.hiv$hiv.svm}) and NN
#' (\code{ROCR.hiv$hiv.nn}) classification data. Each of those is in turn a list
#' consisting of the two elements \code{$predictions} and \code{$labels} (10
#' element list representing cross-validation data).
#'
#' @references
#' Sing, T. & Beerenwinkel, N. & Lengauer, T.  "Learning mixtures
#' of localized rules by maximizing the area under the ROC curve".  1st
#'   International Workshop on ROC Analysis in AI, 89-96, 2004.
#'
#' @usage data(ROCR.hiv)
#'
#' @examples
#' library(ROCR)
#' data(ROCR.hiv)
#' attach(ROCR.hiv)
#' pred.svm <- prediction(hiv.svm$predictions, hiv.svm$labels)
#' pred.svm
#' perf.svm <- performance(pred.svm, 'tpr', 'fpr')
#' perf.svm
#' pred.nn <- prediction(hiv.nn$predictions, hiv.svm$labels)
#' pred.nn
#' perf.nn <- performance(pred.nn, 'tpr', 'fpr')
#' perf.nn
#' plot(perf.svm, lty=3, col="red",main="SVMs and NNs for prediction of
#' HIV-1 coreceptor usage")
#' plot(perf.nn, lty=3, col="blue",add=TRUE)
#' plot(perf.svm, avg="vertical", lwd=3, col="red",
#'      spread.estimate="stderror",plotCI.lwd=2,add=TRUE)
#' plot(perf.nn, avg="vertical", lwd=3, col="blue",
#'      spread.estimate="stderror",plotCI.lwd=2,add=TRUE)
#' legend(0.6,0.6,c('SVM','NN'),col=c('red','blue'),lwd=3)
"ROCR.hiv"


#' @name ROCR.simple
#'
#' @docType data
#' @keywords datasets
#'
#' @title Data set: Simple artificial prediction data for use with ROCR
#'
#' @description
#' A mock data set containing a simple set of predictions and corresponding
#' class labels.
#'
#' @format
#' A two element list. The first element, \code{ROCR.simple$predictions}, is a
#' vector of numerical predictions. The second element,
#' \code{ROCR.simple$labels}, is a vector of corresponding class labels.
#'
#' @usage data(ROCR.simple)
#'
#' @examples
#' # plot a ROC curve for a single prediction run
#' # and color the curve according to cutoff.
#' library(ROCR)
#' data(ROCR.simple)
#' pred <- prediction(ROCR.simple$predictions, ROCR.simple$labels)
#' pred
#' perf <- performance(pred,"tpr","fpr")
#' perf
#' plot(perf,colorize=TRUE)
"ROCR.simple"


#' @name ROCR.xval
#'
#' @docType data
#' @keywords datasets
#'
#' @title Data set: Artificial cross-validation data for use with ROCR
#'
#' @description
#' A mock data set containing 10 sets of predictions and corresponding labels as
#' would be obtained from 10-fold cross-validation.
#'
#' @format
#' A two element list. The first element, \code{ROCR.xval$predictions}, is
#' itself a 10 element list. Each of these 10 elements is a vector of numerical
#' predictions for each cross-validation run. Likewise, the second list entry,
#' \code{ROCR.xval$labels} is a 10 element list in which each element is a
#' vector of true class labels corresponding to the predictions.
#'
#' @usage data(ROCR.xval)
#'
#' @examples
#' # plot ROC curves for several cross-validation runs (dotted
#' # in grey), overlaid by the vertical average curve and boxplots
#' # showing the vertical spread around the average.
#' library(ROCR)
#' data(ROCR.xval)
#' pred <- prediction(ROCR.xval$predictions, ROCR.xval$labels)
#' pred
#' perf <- performance(pred,"tpr","fpr")
#' perf
#' plot(perf,col="grey82",lty=3)
#' plot(perf,lwd=3,avg="vertical",spread.estimate="boxplot",add=TRUE)
"ROCR.xval"

#' @name prediction
#'
#' @title Function to create prediction objects
#'
#' @description
#' Every classifier evaluation using ROCR starts with creating a
#' \code{prediction} object. This function is used to transform the input data
#' (which can be in vector, matrix, data frame, or list form) into a
#' standardized format.
#'
#' @details
#' \code{predictions} and \code{labels} can simply be vectors of the same
#' length. However, in the case of cross-validation data, different
#' cross-validation runs can be provided as the *columns* of a matrix or
#' data frame, or as the entries of a list. In the case of a matrix or
#' data frame, all cross-validation runs must have the same length, whereas
#' in the case of a list, the lengths can vary across the cross-validation
#' runs. Internally, as described in section 'Value', all of these input
#' formats are converted to list representation.
#'
#' Since scoring classifiers give relative tendencies towards a negative
#' (low scores) or positive (high scores) class, it has to be declared
#' which class label denotes the negative, and which the positive class.
#' Ideally, labels should be supplied as ordered factor(s), the lower
#' level corresponding to the negative class, the upper level to the
#' positive class. If the labels are factors (unordered), numeric,
#' logical or characters, ordering of the labels is inferred from
#' R's built-in \code{<} relation (e.g. 0 < 1, -1 < 1, 'a' < 'b',
#' FALSE < TRUE). Use \code{label.ordering} to override this default
#' ordering. Please note that the ordering can be locale-dependent
#' e.g. for character labels '-1' and '1'.
#'
#' Currently, ROCR supports only binary classification (extensions toward
#' multiclass classification are scheduled for the next release,
#' however). If there are more than two distinct label symbols, execution
#' stops with an error message. If all predictions use the same two
#' symbols that are used for the labels, categorical predictions are
#' assumed. If there are more than two predicted values, but all numeric,
#' continuous predictions are assumed (i.e. a scoring
#' classifier). Otherwise, if more than two symbols occur in the
#' predictions, and not all of them are numeric, execution stops with an
#' error message.
#'
#' @param predictions A vector, matrix, list, or data frame containing the
#'   predictions.
#' @param labels A vector, matrix, list, or data frame containing the true class
#'   labels. Must have the same dimensions as \code{predictions}.
#' @param label.ordering The default ordering (cf.details)  of the classes can
#'   be changed by supplying a vector containing the negative and the positive
#'   class label.
#'
#' @return An S4 object of class \code{prediction}.
#'
#' @references
#' A detailed list of references can be found on the ROCR homepage at
#' \url{https://ipa-tys.github.io/ROCR/}.
#'
#' @author
#' Tobias Sing \email{tobias.sing@gmail.com}, Oliver Sander
#' \email{osander@gmail.com}
#'
#' @seealso
#' \code{\link{prediction-class}},
#' \code{\link{performance}},
#' \code{\link{performance-class}},
#' \code{\link{plot.performance}}
#'
#' @export
#'
#' @examples
#' # create a simple prediction object
#' library(ROCR)
#' data(ROCR.simple)
#' pred <- prediction(ROCR.simple$predictions,ROCR.simple$labels)
#' pred
prediction <- function(predictions, labels, label.ordering=NULL) {

  ## bring 'predictions' and 'labels' into list format,
  ## each list entry representing one x-validation run

  ## convert predictions into canonical list format
  if (is.data.frame(predictions)) {
    names(predictions) <- c()
    predictions <- as.list(predictions)
  } else if (is.matrix(predictions)) {
    predictions <- as.list(data.frame(predictions))
    names(predictions) <- c()
  } else if (is.vector(predictions) && !is.list(predictions)) {
    predictions <- list(predictions)
  } else if (!is.list(predictions)) {
    stop("Format of predictions is invalid. It couldn't be coerced to a list.",
         call. = FALSE)
  }
  ## if predictions is a list -> keep unaltered
  if(any(vapply(predictions,anyNA,logical(1)))){
    stop("'predictions' contains NA.", call. = FALSE)
  }

  ## convert labels into canonical list format
  if (is.data.frame(labels)) {
    names(labels) <- c()
    labels <- as.list( labels)
  } else if (is.matrix(labels)) {
    labels <- as.list( data.frame( labels))
    names(labels) <- c()
  } else if ((is.vector(labels) ||
              is.ordered(labels) ||
              is.factor(labels)) &&
             !is.list(labels)) {
    labels <- list( labels)
  } else if (!is.list(labels)) {
    stop("Format of labels is invalid. It couldn't be coerced to a list.",
         call. = FALSE)
  }
  ## if labels is a list -> keep unaltered

  ## Length consistency checks
  if (length(predictions) != length(labels))
    stop(paste("Number of cross-validation runs must be equal",
               "for predictions and labels."))
  if (! all(sapply(predictions, length) == sapply(labels, length)))
    stop(paste("Number of predictions in each run must be equal",
               "to the number of labels for each run."))

  ## only keep prediction/label pairs that are finite numbers
  for (i in 1:length(predictions)) {
    finite.bool <- is.finite( predictions[[i]] )
    predictions[[i]] <- predictions[[i]][ finite.bool ]
    labels[[i]] <- labels[[i]][ finite.bool ]
  }

  ## abort if 'labels' format is inconsistent across
  ## different cross-validation runs
  label.format=""  ## one of 'normal','factor','ordered'
  if (all(sapply( labels, is.factor)) &&
      !any(sapply(labels, is.ordered))) {
    label.format <- "factor"
  } else if (all(sapply( labels, is.ordered))) {
    label.format <- "ordered"
  } else if (all(sapply( labels, is.character)) ||
             all(sapply( labels, is.numeric)) ||
             all(sapply( labels, is.logical))) {
    label.format <- "normal"
  } else {
    stop(paste("Inconsistent label data type across different",
               "cross-validation runs."))
  }

  ## abort if levels are not consistent across different
  ## cross-validation runs
  if (! all(sapply(labels, levels)==levels(labels[[1]])) ) {
    stop(paste("Inconsistent factor levels across different",
               "cross-validation runs."))
  }

  ## convert 'labels' into ordered factors, aborting if the number
  ## of classes is not equal to 2.
  levels <- c()
  if ( label.format == "ordered" ) {
    if (!is.null(label.ordering)) {
      stop(paste("'labels' is already ordered. No additional",
                 "'label.ordering' must be supplied."))
    } else {
      levels <- levels(labels[[1]])
    }
  } else {
    if ( is.null( label.ordering )) {
      if ( label.format == "factor" ) levels <- sort(levels(labels[[1]]))
      else levels <- sort( unique( unlist( labels)))
    } else {
      ## if (!setequal( levels, label.ordering)) {
      if (!setequal( unique(unlist(labels)), label.ordering )) {
        stop("Label ordering does not match class labels.")
      }
      levels <- label.ordering
    }
    for (i in 1:length(labels)) {
      if (is.factor(labels))
        labels[[i]] <- ordered(as.character(labels[[i]]),
                               levels=levels)
      else labels[[i]] <- ordered( labels[[i]], levels=levels)
    }

  }

  if (length(levels) != 2) {
    message <- paste("Number of classes is not equal to 2.\n",
                     "ROCR currently supports only evaluation of ",
                     "binary classification tasks.",sep="")
    stop(message)
  }

  ## determine whether predictions are continuous or categorical
  ## (in the latter case stop; scheduled for the next ROCR version)
  if (!is.numeric( unlist( predictions ))) {
    stop("Currently, only continuous predictions are supported by ROCR.")
  }

  ## compute cutoff/fp/tp data

  cutoffs <- list()
  fp <- list()
  tp <- list()
  fn <- list()
  tn <- list()
  n.pos <- list()
  n.neg <- list()
  n.pos.pred <- list()
  n.neg.pred <- list()
  for (i in 1:length(predictions)) {
    n.pos <- c( n.pos, sum( labels[[i]] == levels[2] ))
    n.neg <- c( n.neg, sum( labels[[i]] == levels[1] ))
    ans <- .compute.unnormalized.roc.curve( predictions[[i]], labels[[i]] )
    cutoffs <- c( cutoffs, list( ans$cutoffs ))
    fp <- c( fp, list( ans$fp ))
    tp <- c( tp, list( ans$tp ))
    fn <- c( fn, list( n.pos[[i]] - tp[[i]] ))
    tn <- c( tn, list( n.neg[[i]] - fp[[i]] ))
    n.pos.pred <- c(n.pos.pred, list(tp[[i]] + fp[[i]]) )
    n.neg.pred <- c(n.neg.pred, list(tn[[i]] + fn[[i]]) )
  }


  return( new("prediction", predictions=predictions,
              labels=labels,
              cutoffs=cutoffs,
              fp=fp,
              tp=tp,
              fn=fn,
              tn=tn,
              n.pos=n.pos,
              n.neg=n.neg,
              n.pos.pred=n.pos.pred,
              n.neg.pred=n.neg.pred))
}


## fast fp/tp computation based on cumulative summing
.compute.unnormalized.roc.curve <- function( predictions, labels ) {
  ## determine the labels that are used for the pos. resp. neg. class :
  pos.label <- levels(labels)[2]
  neg.label <- levels(labels)[1]

  pred.order <- order(predictions, decreasing=TRUE)
  predictions.sorted <- predictions[pred.order]
  tp <- cumsum(labels[pred.order]==pos.label)
  fp <- cumsum(labels[pred.order]==neg.label)

  ## remove fp & tp for duplicated predictions
  ## as duplicated keeps the first occurrence, but we want the last, two
  ## rev are used.
  ## Highest cutoff (Infinity) corresponds to tp=0, fp=0
  dups <- rev(duplicated(rev(predictions.sorted)))
  tp <- c(0, tp[!dups])
  fp <- c(0, fp[!dups])
  cutoffs <- c(Inf, predictions.sorted[!dups])

  return(list( cutoffs=cutoffs, fp=fp, tp=tp ))
}

context("consistency")
test_that("consistency:",{
  .get.performance.measures <- function(pred, measures) {
    
    ans <- list()
    for (measure in measures) {
      ## need to enclose y.values into a list to avoid flattening
      perf <- performance(pred, measure)
      .check.performance.object( perf )
      ans <- c(ans, list(perf@y.values))
    }
    names(ans) <- measures
    ans
  }
  
  .check.consistency <- function(measures) {
    ## check entries of contingency table for consistency
    
    for (measure in c("acc", "err", "fnr", "tpr", "fpr", "tnr", "pcfall", "prec", "npv", "pcmiss",'rpp','rnp')) {
      if (!measure %in% names(measures)) {
        stop(paste("Performance measure", measure, "not in argument list."))
      }
    }
    for (i in 1:length(measures$acc)) {
      finite.bool <- is.finite(measures$acc[[i]]) & is.finite(measures$err[[i]])
      expect_equal(measures$acc[[i]][finite.bool] + measures$err[[i]][finite.bool],
                   rep(1,length(measures$acc[[i]]))[finite.bool])
      
      finite.bool <- is.finite(measures$fnr[[i]]) & is.finite(measures$tpr[[i]])
      expect_equal(measures$fnr[[i]][finite.bool] + measures$tpr[[i]][finite.bool],
                   rep(1,length(measures$fnr[[i]]))[finite.bool])
      
      finite.bool <- is.finite(measures$fpr[[i]]) & is.finite(measures$tnr[[i]])
      expect_equal(measures$fpr[[i]][finite.bool] + measures$tnr[[i]][finite.bool],
                   rep(1,length(measures$fpr[[i]]))[finite.bool])
      
      finite.bool <- is.finite(measures$prec[[i]]) & is.finite(measures$pcfall[[i]])
      expect_equal(measures$prec[[i]][finite.bool] + measures$pcfall[[i]][finite.bool],
                   rep(1,length(measures$acc[[i]]))[finite.bool])
      
      finite.bool <- is.finite(measures$npv[[i]]) & is.finite(measures$pcmiss[[i]])
      expect_equal(measures$npv[[i]][finite.bool] + measures$pcmiss[[i]][finite.bool],
                   rep(1,length(measures$acc[[i]]))[finite.bool])
      
      expect_equal(measures$rpp[[i]] + measures$rnp[[i]], rep(1, length(measures$rpp[[i]])))
    }
    
  }
  
  ############################################################
  # test length of performance measures
  
  .check.performance.object <- function(perf) {
    ylen <- length(perf@y.values)
    xlen <- length(perf@x.values)
    alphalen <- length(perf@alpha.values)
    
    expect_equal( (xlen==0 || xlen==ylen) && (alphalen==0 || (alphalen==xlen && alphalen==ylen)), T )
    
    if (xlen==ylen) {
      for (i in 1:ylen) expect_equal( length(perf@x.values[[i]]), length(perf@y.values[[i]])  )
    }
    if (alphalen==ylen) {
      for (i in 1:ylen) expect_equal( length(perf@alpha.values[[i]]), length(perf@y.values[[i]])  )
    }
  }
  
  .check.prediction.object <- function( pred) {
    # 1. all entries in prediction object must have equals number of cross-validation runs
    lenvec <- c(length(pred@predictions), length(pred@labels), length(pred@cutoffs), length(pred@fp),
                length(pred@tp), length(pred@fn), length(pred@tn), length(pred@n.pos),
                length(pred@n.neg), length(pred@n.pos.pred), length(pred@n.neg.pred))
    expect_equal( length(unique(lenvec)), 1)
    
    # 2. inside: xval runs:
    for (i in 1:length(pred@predictions)) {
      expect_equal( length(pred@predictions[[i]]), length(pred@labels[[i]]))
      lenvec <- c(length(pred@cutoffs[[i]]), length(pred@fp[[i]]),
                  length(pred@tp[[i]]), length(pred@fn[[i]]),
                  length(pred@tn[[i]]), length(pred@n.pos.pred[[i]]), length(pred@n.neg.pred[[i]]))
      expect_equal( length(unique(lenvec)), 1)
      expect_equal( unique(lenvec), length(unique(pred@predictions[[i]]))+1 )
    }
    # 3. cutoffs sorted in descending order?
    for (i in 1:length(pred@predictions)) {
      expect_equal( sort(pred@cutoffs[[i]], decreasing=TRUE ), pred@cutoffs[[i]] ) 
    }
    
    # 4. check 2x2 table for consistency with marginal sums
    for (i in 1:length(pred@predictions)) {
      expect_equal( pred@tp[[i]] + pred@fp[[i]], pred@n.pos.pred[[i]] )
      expect_equal( pred@fn[[i]] + pred@tn[[i]], pred@n.neg.pred[[i]] )
      expect_equal( pred@tp[[i]] + pred@fn[[i]], rep( pred@n.pos[[i]], length(pred@tp[[i]]))  )
      expect_equal( pred@fp[[i]] + pred@tn[[i]], rep( pred@n.neg[[i]], length(pred@tp[[i]]))  )
      expect_equal(pred@n.pos.pred[[i]] + pred@n.neg.pred[[i]],
                   rep( pred@n.pos[[i]] + pred@n.neg[[i]], length(pred@n.pos.pred[[i]])) )
      expect_equal(pred@n.pos[[i]] + pred@n.neg[[i]], length(pred@labels[[i]]))
    }
    
  }
  
  #  
  .mock.prediction <- function( n.predictions, error.rate ) {
    if ( length(n.predictions) > 1 && length(error.rate)==1) {
      error.rate <- rep(error.rate, length(n.predictions) )
    }
    
    if (length(n.predictions)>1) {
      predictions <- list()
      labels <- list()
    } else {
      predictions <- c()
      labels <- c()
    }
    
    for (i in 1:length(n.predictions)) {
      current.predictions <- runif( n.predictions[i] )
      current.labels <- as.numeric( current.predictions >= 0.5)
      flip.indices <- sample( n.predictions[i], round( error.rate[i] * n.predictions[i] ))
      current.labels[ flip.indices ] <- !current.labels[ flip.indices ]
      # current.labels[ current.labels=="1" ] <- "+"
      # current.labels[ current.labels=="0" ] <- "-"
      
      if (length(n.predictions)>1) {
        predictions <- c( predictions, list( current.predictions ))
        labels <- c( labels, list( current.labels ))
      }
    }
    
    if (length( n.predictions)==1) {
      predictions <- list(current.predictions)
      labels <- list(current.labels)
    }
    
    ans <- list(predictions= predictions, labels= labels)
    
    # ensure, that random labels have exactly two levels
    if (any(   sapply(labels, function(run) {length(unique(run))})  !=  rep(2, length(labels)) )) {
      # print(paste("XXX", labels, str(n.predictions), str(error.rate)))
      return(.mock.prediction(n.predictions, error.rate))
    }
    else return( ans )
  }
  
  ##############################################################################
  # consistency
  for (i in 1:100) {
    n.folds <- sample(1:10,1)
    fold.sizes <- sample(10:100, n.folds, replace=T)
    error.rates <- runif( n.folds )
    pp <- .mock.prediction( fold.sizes, error.rates )
    pred <- prediction( pp$predictions, pp$labels )
    .check.prediction.object(pred)
    a <- .get.performance.measures( pred, c('acc','err','fpr','tpr','fnr','tnr','prec','pcfall','npv','pcmiss','rpp','rnp'))
    .check.consistency( a)
  }
  
  ##############################################################################
  # test errors
  crashCases <- list(       ## cases that are ok to crash:
    list(pred= c(0), lab= c(0)), #-> Number of classes is not equal to 2.
    list(pred= c(1), lab= c(1)), #-> Number of classes is not equal to 2.
    list(pred= c(0.1, 0.2, 0.5), lab= c(1,1,1)), #-> Number of classes is not equal to 2.
    list(pred= c(0.1, 0.2, 0.5), lab= c(0,0,0)), #-> Number of classes is not equal to 2.
    list(pred= c(0.1, 0.2, 0.5), lab= c("a", "a", "a")),  #-> Number of classes is not equal to 2.
    list(pred= c(0.1, 0.2, 0.5), lab= c(T, T, T)), #-> Number of classes is not equal to 2.
    list(pred= c(0.1, 0.2, 0.5), lab= c(F, F, F))  #-> Number of classes is not equal to 2.
  )
  for (case in crashCases) {
    # cat(case$pred, " ", case$lab, "\n")
    expect_error(pred <- prediction(case$pred, case$lab))
    #checkException(measures <- .get.performance.measures(pred))
  }
  
  ##############################################################################
  ## use consistency checks to validate results on pathological input cases
  performance.measures <- c('tpr','fpr','acc','err','rec','sens','fnr','tnr','spec',
                            'ppv','prec','npv','fall','miss','pcfall','pcmiss','rpp','rnp',
                            'auc','prbe','rch','mxe','rmse','phi','mat','mi','chisq',
                            'odds','lift','f','sar','ecost','cost')

  # mxe needs 0,1 labels (warning otherwise),
  # rmse needs numeric labels (warning otherwise), sar as well
  pred <- prediction( c(0.1, 0.2, 0.5), c("a", "a", "b"))
  .check.prediction.object(pred)
  measures.to.evaluate <- performance.measures[ performance.measures != 'mxe' &
                                                  performance.measures != 'rmse' &
                                                  performance.measures != 'sar']
  measures <- expect_warning(.get.performance.measures(pred, measures.to.evaluate),
                             "Chi-squared approximation may be incorrect")
  .check.consistency( measures)

  pred <- prediction( c(0.1, 0.2, 0.5), c(F, F, T))
  .check.prediction.object(pred)
  measures.to.evaluate <- performance.measures[ performance.measures != 'mxe' &
                                                  performance.measures != 'rmse' &
                                                  performance.measures != 'sar']

  measures <- expect_warning(.get.performance.measures(pred, measures.to.evaluate),
                             "Chi-squared approximation may be incorrect")
  .check.consistency( measures)

  pred <- prediction( c(0.1, 0.2, 0.5), c("1", "1", "0"))
  .check.prediction.object(pred)
  measures.to.evaluate <- performance.measures
  measures <- expect_warning(.get.performance.measures(pred, measures.to.evaluate),
                             "Chi-squared approximation may be incorrect")
  .check.consistency( measures)

  pred <- prediction( c(0.1, 0.2, 0.5), c(T, F, F))
  .check.prediction.object(pred)
  measures.to.evaluate <- performance.measures[ performance.measures != 'mxe' &
                                                  performance.measures != 'rmse' &
                                                  performance.measures != 'sar' ]
  measures <- expect_warning(.get.performance.measures(pred, measures.to.evaluate),
                             "Chi-squared approximation may be incorrect")
  .check.consistency( measures)

  # prbe cannot be computed, because only one prec/rec pair available.
  pred <- prediction( c(0,0,0), c(0,1,1))
  .check.prediction.object(pred)
  measures.to.evaluate <- performance.measures[ performance.measures != 'prbe' ]
  measures <- expect_warning(.get.performance.measures(pred, measures.to.evaluate),
                             "Chi-squared approximation may be incorrect")
  .check.consistency( measures)

  pred <- prediction( c(0,0,0), ordered(c(0,0,0), levels=c(0,1)))
  .check.prediction.object(pred)
  measures.to.evaluate <- performance.measures[ performance.measures != 'auc' &
                                                  performance.measures != 'prbe' &
                                                  performance.measures != 'rch' &
                                                  performance.measures != 'sar' &
                                                  performance.measures != 'ecost']
  measures <- expect_warning(.get.performance.measures(pred, measures.to.evaluate),
                             "Chi-squared approximation may be incorrect")
  .check.consistency( measures)

  pred <- prediction( c(-1,-0.2,-0.6), ordered(c(1,0,1), levels=c(0,1)))
  .check.prediction.object(pred)
  measures.to.evaluate <- performance.measures[ performance.measures != 'mxe' ]
  measures <- expect_warning(.get.performance.measures(pred, measures.to.evaluate),
                             "Chi-squared approximation may be incorrect")
  .check.consistency( measures)

  pred <- prediction( c(-1,-0.2,-0.6), c(-1,1,-1))
  .check.prediction.object(pred)
  measures.to.evaluate <- performance.measures[ performance.measures != 'mxe']
  measures <- expect_warning(.get.performance.measures(pred, measures.to.evaluate),
                             "Chi-squared approximation may be incorrect")
  .check.consistency( measures)

  pred <- prediction( c(-1,-0.2,-0.6), c(3,2,3))
  .check.prediction.object(pred)
  measures.to.evaluate <- performance.measures[ performance.measures != 'mxe']
  measures <- expect_warning(.get.performance.measures(pred, measures.to.evaluate),
                             "Chi-squared approximation may be incorrect")
  .check.consistency( measures)

  pred <- prediction( c(1), ordered(c("a"),levels=c('a','b')))
  .check.prediction.object(pred)
  measures.to.evaluate <- performance.measures[ performance.measures != 'auc' &
                                                  performance.measures != 'prbe' &
                                                  performance.measures != 'rch' &
                                                  performance.measures != 'mxe' &
                                                  performance.measures != 'rmse' &
                                                  performance.measures != 'sar' &
                                                  performance.measures != 'ecost']
  measures <- expect_warning(.get.performance.measures(pred, measures.to.evaluate),
                             "Chi-squared approximation may be incorrect")

  .check.consistency( measures)
  
  ##############################################################################
  # test measures for consistency on supplied data sets
  data(ROCR.simple)
  pred <- prediction(ROCR.simple$predictions, ROCR.simple$labels)
  .check.prediction.object(pred)
  measures.to.evaluate <- performance.measures
  measures <- expect_warning(.get.performance.measures(pred, measures.to.evaluate),
                             "Chi-squared approximation may be incorrect")
  .check.consistency( measures)
  
  data(ROCR.xval)
  pred <- prediction(ROCR.xval$predictions, ROCR.xval$labels)
  .check.prediction.object(pred)
  measures.to.evaluate <- performance.measures
  measures <- expect_warning(.get.performance.measures(pred, measures.to.evaluate),
                             "Chi-squared approximation may be incorrect")
  .check.consistency( measures)
  
  data(ROCR.hiv)
  pred <- prediction(ROCR.hiv$hiv.nn$predictions, ROCR.hiv$hiv.nn$labels)
  .check.prediction.object(pred)
  measures.to.evaluate <- performance.measures[performance.measures != 'mxe' &
                                                 performance.measures != 'cal']
  measures <- expect_warning(.get.performance.measures(pred, measures.to.evaluate),
                             "Chi-squared approximation may be incorrect")
  .check.consistency( measures)
  
  pred <- prediction(ROCR.hiv$hiv.svm$predictions, ROCR.hiv$hiv.svm$labels)
  .check.prediction.object(pred)
  measures.to.evaluate <- performance.measures[performance.measures != 'mxe' &
                                                 performance.measures != 'cal']
  measures <- expect_warning(.get.performance.measures(pred, measures.to.evaluate),
                             "Chi-squared approximation may be incorrect")
  .check.consistency( measures)
  
  skip_on_cran()
  skip_on_ci()
  ##############################################################################
  # Combining
  measures <- c('tpr','fpr','acc','err','rec','sens','fnr','tnr','spec',
                'ppv','prec','npv','fall','miss','pcfall','pcmiss','rpp','rnp',
                'phi','mat','mi','chisq','odds','lift')
  #                  'auc','prbe','rch','mxe','rmse','phi','mat','mi','chisq',
  #                  'odds','lift','f','sar','ecost','cost')
  
  for (measure1 in measures) {
    # print(measure1)
    for (measure2 in measures) {
      n.folds <- sample(1:2,1)
      fold.sizes <- sample(10:20, n.folds, replace=T)
      error.rates <- runif( n.folds )
      pp <- .mock.prediction( fold.sizes, error.rates )
      pred <- prediction( pp$predictions, pp$labels )
      .check.prediction.object(pred)
      perf1 <- suppressWarnings(performance( pred, measure1 ))
      perf2 <- suppressWarnings(performance( pred, measure2 ))
      perf3 <- suppressWarnings(performance( pred, measure2, measure1 ))
      .check.performance.object(perf1)
      .check.performance.object(perf2)
      .check.performance.object(perf3)
      
      for (i in 1:n.folds) {
        #check elements
        expect_equal(setequal( c( perf1@x.values[[i]], perf2@x.values[[i]]), perf3@alpha.values[[i]] ),T)
        expect_equal(setequal( perf1@y.values[[i]], perf3@x.values[[i]] ),T)
        expect_equal(setequal( perf2@y.values[[i]], perf3@y.values[[i]] ),T)
        
        #check order
        ind <- sapply( perf1@x.values[[i]], function(x) { min(which(x==perf3@alpha.values[[i]]))})
        expect_equal( unname(perf1@y.values[[i]]), perf3@x.values[[i]][ind] )
        expect_equal( unname(perf2@y.values[[i]]), perf3@y.values[[i]][ind] )
        
      }
    }
  }
  
  ##############################################################################
  # test datavase combine
  measures <- c('tpr','fpr','acc','err','rec','sens','fnr','tnr','spec',
                'ppv','prec','npv','fall','miss','pcfall','pcmiss','rpp','rnp',
                'phi','mat','mi','chisq','odds','lift')
  #'auc','prbe','rch','mxe','rmse','phi','mat','mi','chisq',
  #'odds','lift','f','sar','ecost','cost')
  # print("Database combine test deactivated.")
  data(ROCR.simple)
  data(ROCR.xval)
  data(ROCR.hiv)
  all.pred <- list(prediction(ROCR.simple$predictions, ROCR.simple$labels),
                   prediction(ROCR.xval$predictions, ROCR.xval$labels),
                   prediction(ROCR.hiv$hiv.nn$predictions, ROCR.hiv$hiv.nn$labels),
                   prediction(ROCR.hiv$hiv.svm$predictions, ROCR.hiv$hiv.svm$labels))
  lapply(all.pred, .check.prediction.object)
  for (pred in all.pred) {
    for (measure1 in measures) {
      # print(measure1)
      for (measure2 in measures) {
        perf1 <- suppressWarnings(performance( pred, measure1 ))
        perf2 <- suppressWarnings(performance( pred, measure2 ))
        perf3 <- suppressWarnings(performance( pred, measure2, measure1 ))
        .check.performance.object(perf1)
        .check.performance.object(perf2)
        .check.performance.object(perf3)

        for (i in 1:length(pred@labels)) {
          #check elements
          expect_equal(setequal( c( perf1@x.values[[i]], perf2@x.values[[i]]), perf3@alpha.values[[i]] ),T)
          expect_equal(setequal( perf1@y.values[[i]], perf3@x.values[[i]] ),T)
          expect_equal(setequal( perf2@y.values[[i]], perf3@y.values[[i]] ),T)

          # check order
          ind <- sapply( perf1@x.values[[i]], function(x) { min(which(x==perf3@alpha.values[[i]]))})
          expect_equal( unname(perf1@y.values[[i]]), perf3@x.values[[i]][ind] )
          expect_equal( unname(perf2@y.values[[i]]), perf3@y.values[[i]][ind] )

        }
      }
    }
  }
})

## ------------------------------------------------------------------------
## classical machine learning contingency table measures
## ------------------------------------------------------------------------

.performance.accuracy <-
  function(predictions, labels, cutoffs, fp, tp, fn, tn,
           n.pos, n.neg, n.pos.pred, n.neg.pred) {

    list( cutoffs, (tn+tp) / length(predictions) )
  }

.performance.error.rate <-
  function(predictions, labels, cutoffs, fp, tp, fn, tn,
           n.pos, n.neg, n.pos.pred, n.neg.pred) {

    list( cutoffs, (fn+fp) / length(predictions) )
  }

.performance.false.positive.rate <-
  function(predictions, labels, cutoffs, fp, tp, fn, tn,
           n.pos, n.neg, n.pos.pred, n.neg.pred) {

    list( cutoffs, fp / n.neg )
  }

.performance.true.positive.rate <-
  function(predictions, labels, cutoffs, fp, tp, fn, tn,
           n.pos, n.neg, n.pos.pred, n.neg.pred) {

    list( cutoffs, tp / n.pos )
  }

.performance.false.negative.rate <-
  function(predictions, labels, cutoffs, fp, tp, fn, tn,
           n.pos, n.neg, n.pos.pred, n.neg.pred) {

    list( cutoffs, fn / n.pos )
  }

.performance.true.negative.rate <-
  function(predictions, labels, cutoffs, fp, tp, fn, tn,
           n.pos, n.neg, n.pos.pred, n.neg.pred) {

    list( cutoffs, tn / n.neg )
  }

.performance.positive.predictive.value <-
  function(predictions, labels, cutoffs, fp, tp, fn, tn,
           n.pos, n.neg, n.pos.pred, n.neg.pred) {

    ppv <- tp / (fp + tp)
    list( cutoffs, ppv )
  }

.performance.negative.predictive.value <-
  function(predictions, labels, cutoffs, fp, tp, fn, tn,
           n.pos, n.neg, n.pos.pred, n.neg.pred) {

    npv <- tn / (tn + fn)
    list( cutoffs, npv )
  }

.performance.prediction.conditioned.fallout <-
  function(predictions, labels, cutoffs, fp, tp, fn, tn,
           n.pos, n.neg, n.pos.pred, n.neg.pred) {
    ppv <- .performance.positive.predictive.value(predictions, labels,
                                                  cutoffs, fp, tp, fn, tn,
                                                  n.pos, n.neg, n.pos.pred,
                                                  n.neg.pred)[[2]]
    list( cutoffs, 1 - ppv )
  }

.performance.prediction.conditioned.miss <-
  function(predictions, labels, cutoffs, fp, tp, fn, tn,
           n.pos, n.neg, n.pos.pred, n.neg.pred) {
    npv <- .performance.negative.predictive.value(predictions, labels,
                                                  cutoffs, fp, tp, fn, tn,
                                                  n.pos, n.neg, n.pos.pred,
                                                  n.neg.pred)[[2]]
    list( cutoffs, 1 - npv )
  }

## ------------------------------------------------------------------------
## ...not actually performance measures, but very useful as a second axis
## against which to plot a "real" performance measure
## (popular example: lift charts)
## ------------------------------------------------------------------------

.performance.rate.of.positive.predictions <-
  function(predictions, labels, cutoffs, fp, tp, fn, tn,
           n.pos, n.neg, n.pos.pred, n.neg.pred) {

    list( cutoffs, n.pos.pred / (n.pos + n.neg) )
  }

.performance.rate.of.negative.predictions <-
  function(predictions, labels, cutoffs, fp, tp, fn, tn,
           n.pos, n.neg, n.pos.pred, n.neg.pred) {

    list( cutoffs, n.neg.pred / (n.pos + n.neg) )
  }


## ------------------------------------------------------------------------
## Classical statistical contingency table measures
## ------------------------------------------------------------------------

.performance.phi <-
  function(predictions, labels, cutoffs, fp, tp, fn, tn,
           n.pos, n.neg, n.pos.pred, n.neg.pred) {

    list(cutoffs,
         (tn*tp - fn*fp) /
           (sqrt(n.pos) * sqrt(n.neg) * sqrt(n.pos.pred) * sqrt(n.neg.pred))
        )
  }

.performance.mutual.information <-
  function(predictions, labels, cutoffs, fp, tp, fn, tn,
           n.pos, n.neg, n.pos.pred, n.neg.pred) {

    n.samples <- n.pos + n.neg
    mi <- c()
    for (k in 1:length(cutoffs)) {
      kij <- rbind( c(tn[k],fn[k]), c(fp[k],tp[k]) )

      ki.j. <- rbind(c(n.neg * n.neg.pred[k], n.neg.pred[k] * n.pos),
                     c(n.neg * n.pos.pred[k], n.pos * n.pos.pred[k]))

      log.matrix <- log2( kij / ki.j.)
      log.matrix[kij/ki.j.==0] <- 0

      mi <- c(mi,  log2(n.samples) + sum( kij * log.matrix) / n.samples  )
    }

    list( cutoffs, mi )
  }

#' @importFrom stats chisq.test
.performance.chisq <-
  function(predictions, labels, cutoffs, fp, tp, fn, tn,
           n.pos, n.neg, n.pos.pred, n.neg.pred) {

    chisq <- c()
    for (i in 1:length(cutoffs)) {
      A <- rbind( c( tn[i], fn[i]), c(fp[i], tp[i]) )
      chisq <- c(chisq, stats::chisq.test(A, correct=FALSE)$statistic )
    }
    list( cutoffs, chisq )
  }

.performance.odds.ratio <-
  function(predictions, labels, cutoffs, fp, tp, fn, tn,
           n.pos, n.neg, n.pos.pred, n.neg.pred) {


    list( cutoffs, tp * tn / (fn * fp) )

  }

## ------------------------------------------------------------------------
## Other measures based on contingency tables
## ------------------------------------------------------------------------

.performance.lift <-
  function(predictions, labels, cutoffs, fp, tp, fn, tn,
           n.pos, n.neg, n.pos.pred, n.neg.pred) {

    n.samples <- n.pos + n.neg
    list( cutoffs, (tp / n.pos) / (n.pos.pred / n.samples) )
  }

.performance.f <-
  function(predictions, labels, cutoffs, fp, tp, fn, tn,
           n.pos, n.neg, n.pos.pred, n.neg.pred, alpha) {

    prec <- .performance.positive.predictive.value(predictions, labels,
                                                   cutoffs, fp, tp, fn, tn,
                                                   n.pos, n.neg, n.pos.pred,
                                                   n.neg.pred)[[2]]
    list( cutoffs,  1/ ( alpha*(1/prec) + (1-alpha)*(1/(tp/n.pos))  ) )
  }

#' @importFrom grDevices chull
.performance.rocconvexhull <-
  function(predictions, labels, cutoffs, fp, tp, fn, tn,
           n.pos, n.neg, n.pos.pred, n.neg.pred) {

    x <- fp / n.neg
    y <- tp / n.pos

    finite.bool <- is.finite(x) & is.finite(y)
    x <- x[ finite.bool ]
    y <- y[ finite.bool ]
    if (length(x) < 2) {
      stop("Not enough distinct predictions to compute ROC convex hull.")
    }

    ## keep only points on the convex hull
    ind <- grDevices::chull(x, y)
    x.ch <- x[ind]
    y.ch <- y[ind]

    ## keep only convex hull points above the diagonal, except (0,0)
    ## and (1,1)
    ind.upper.triangle <- x.ch < y.ch
    x.ch <- c(0, x.ch[ind.upper.triangle], 1)
    y.ch <- c(0, y.ch[ind.upper.triangle], 1)

    ## sort remaining points by ascending x value
    ind <- order(x.ch)
    x.ch <- x.ch[ind]
    y.ch <- y.ch[ind]

    list( x.ch, y.ch )
  }

## ----------------------------------------------------------------------------
## Cutoff-independent measures
## ----------------------------------------------------------------------------

#' @importFrom stats approxfun
.performance.auc <-
  function(predictions, labels, cutoffs, fp, tp, fn, tn,
           n.pos, n.neg, n.pos.pred, n.neg.pred, fpr.stop) {

    x <- fp / n.neg
    y <- tp / n.pos

    finite.bool <- is.finite(x) & is.finite(y)
    x <- x[ finite.bool ]
    y <- y[ finite.bool ]
    if (length(x) < 2) {
      stop(paste("Not enough distinct predictions to compute area",
                 "under the ROC curve."))
    }

    if (fpr.stop < 1) {
      ind <- max(which( x <= fpr.stop ))
      tpr.stop <- stats::approxfun( x[ind:(ind+1)], y[ind:(ind+1)] )(fpr.stop)
      x <- c(x[1:ind], fpr.stop)
      y <- c(y[1:ind], tpr.stop)
    }

    ans <- list()
    auc <- 0
    for (i in 2:length(x)) {
      auc <- auc + 0.5 * (x[i] - x[i-1]) * (y[i] + y[i-1])
    }
    ans <- list( c(), auc)
    names(ans) <- c("x.values","y.values")
    return(ans)
  }

# written by Thomas Unterthiner (unterthiner@bioinf.jku.at)
.performance.aucpr <-
  function(predictions, labels, cutoffs, fp, tp, fn, tn,
           n.pos, n.neg, n.pos.pred, n.neg.pred) {
      
      
      tmp <- aggregate(list(fp=fp), by=list(tp=tp), min)
      tp <- tmp$tp
      fp <- tmp$fp
      prec <- tp / (fp + tp)
      rec <- tp / n.pos
      if (fp[1] == 0 & tp[1] == 0) {
          prec[1] = 1
      }

      finite.bool <- is.finite(prec) & is.finite(rec)
      prec <- prec[ finite.bool ]
      rec <- rec[ finite.bool ]	
      if (length(rec) < 2) {
          stop(paste("Not enough distinct predictions to compute area",
                     "under the Precision/Recall curve."))
      }
            
      # if two points are too distant from each other, we need to 
      # correctly interpolate between them. This is done according to
      # Davis & Goadrich, 
      #"The Relationship Between Precision-Recall and ROC Curves", ICML'06
      for (i in seq_along(rec[-length(rec)])) {
          if (tp[i+1] - tp[i] > 2) {
              skew = (fp[i+1]-fp[i]) / (tp[i+1]-tp[i])
              x = seq(1, tp[i+1]-tp[i], by=1)
              rec <- append(rec, (x+tp[i])/n.pos, after=i)
              prec <- append(prec, (x+tp[i])/(tp[i]+fp[i]+x+ skew*x), after=i)
          }
      }

      auc <- 0
      for (i in seq.int(from = 2, to = length(rec))) {
          auc <- auc + 0.5 * (rec[i] - rec[i-1]) * (prec[i] + prec[i-1])
      }
    
      ans <- list( c(), auc)
      names(ans) <- c("x.values","y.values")
      return(ans)
 }

#' @importFrom stats uniroot approxfun
.performance.precision.recall.break.even.point <-
  function(predictions, labels, cutoffs, fp, tp, fn, tn,
           n.pos, n.neg, n.pos.pred, n.neg.pred) {

    pred <- prediction( predictions, labels)
    perf <- performance( pred, measure="prec", x.measure="rec")
    x <- rev(perf@x.values[[1]])
    y <- rev(perf@y.values[[1]])
    alpha <- rev(perf@alpha.values[[1]])

    finite.bool <- is.finite(alpha) & is.finite(x) & is.finite(y)
    x <- x[ finite.bool ]
    y <- y[ finite.bool ]
    alpha <- alpha[ finite.bool ]

    if (length(x) < 2) {
      stop(paste("Not enough distinct predictions to compute",
                 "precision/recall intersections."))
    }
    intersection.cutoff <- c()
    intersection.pr <- c()

    ## find all intersection points by looking at all intervals (i,i+1):
    ## if the difference function between x and y has different signs at the
    ## interval boundaries, then an intersection point is in the interval;
    ## compute as the root of the difference function
    if ( (x[1]-y[1]) == 0) {
      intersection.cutoff <- c( alpha[1] )
      intersection.pr <- c( x[1] )
    }

    for (i in (1:(length(alpha)-1))) {
      if ((x[i+1]-y[i+1]) == 0) {
        intersection.cutoff <- c( intersection.cutoff, alpha[i+1] )
        intersection.pr <- c( intersection.pr, x[i+1] )
      } else if ((x[i]-y[i])*(x[i+1]-y[i+1]) < 0 ) {
        ans <- stats::uniroot(stats::approxfun(c(alpha[i], alpha[i+1] ),
                                               c(x[i]-y[i], x[i+1]-y[i+1])),
                       c(alpha[i],alpha[i+1]))
        intersection.cutoff <- c(intersection.cutoff, ans$root)
        intersection.pr <- c(intersection.pr, ans$f.root)
      }
    }

    list( rev(intersection.cutoff), rev(intersection.pr) )
  }

#' @importFrom stats median
.performance.calibration.error <-
  function(predictions, labels, cutoffs, fp, tp, fn, tn,
           n.pos, n.neg, n.pos.pred, n.neg.pred, window.size) {

    if (window.size > length(predictions)) {
      stop("Window size exceeds number of predictions.")
    }
    if (min(predictions)<0 || max(predictions)>1) {
      stop("Calibration error needs predictions between 0 and 1")
    }

    pos.label <- levels(labels)[2]
    neg.label <- levels(labels)[1]

    ordering <- rev(order( predictions ))
    predictions <- predictions[ ordering ]
    labels <- labels[ ordering ]

    median.cutoffs <- c()
    calibration.errors <- c()

    for (left.index in 1 : (length(predictions) - window.size+1) ) {
      right.index <- left.index + window.size - 1
      pos.fraction <-
        sum(labels[left.index : right.index] == pos.label) / window.size
      mean.prediction <- mean( predictions[ left.index : right.index ] )

      calibration.errors <- c(calibration.errors,
                              abs(pos.fraction - mean.prediction))
      median.cutoffs <- c(median.cutoffs,
                          stats::median(predictions[left.index:right.index]))
    }
    list( median.cutoffs, calibration.errors )
  }


.performance.mean.cross.entropy <-
  function(predictions, labels, cutoffs, fp, tp, fn, tn,
           n.pos, n.neg, n.pos.pred, n.neg.pred) {

    if (! all(levels(labels)==c(0,1)) ||
        any(predictions<0) || any(predictions>1) ) {
      stop(paste("Class labels need to be 0 and 1 and predictions between",
                 "0 and 1 for mean cross entropy."))
    }

    pos.label <- levels(labels)[2]
    neg.label <- levels(labels)[1]

    list( c(), - 1/length(predictions) *
            (sum( log( predictions[which(labels==pos.label)] ))  +
               sum( log( 1 - predictions[which(labels==neg.label)] ))) )
  }


.performance.root.mean.squared.error <-
  function(predictions, labels, cutoffs, fp, tp, fn, tn,
           n.pos, n.neg, n.pos.pred, n.neg.pred) {
    ## convert labels from factor to numeric values
    labels <- as.numeric(levels(labels))[labels]
    if (any(is.na(labels))) {
      stop("For rmse predictions have to be numeric.")
    }
    list( c(),  sqrt( 1/length(predictions) *
                        sum( (predictions - labels)^2 ))  )
  }

## ----------------------------------------------------------------------------
## Derived measures:
## ----------------------------------------------------------------------------

.performance.sar <- function( predictions, labels, cutoffs, fp, tp, fn, tn,
                              n.pos, n.neg, n.pos.pred, n.neg.pred) {

  pred <- prediction( predictions, labels)
  perf.acc <- performance( pred, measure="acc")
  perf.rmse <- performance( pred, measure="rmse")
  perf.auc <- performance( pred, measure="auc")

  list(cutoffs,
       1/3 * (perf.acc@y.values[[1]] +
                (1 - perf.rmse@y.values[[1]]) +
                perf.auc@y.values[[1]]))
}

## ----------------------------------------------------------------------------
## Measures taking into account actual cost considerations
## ----------------------------------------------------------------------------

.performance.expected.cost <-
  function(predictions, labels, cutoffs, fp, tp, fn, tn,
           n.pos, n.neg, n.pos.pred, n.neg.pred) {

    ## kick out suboptimal values (i.e. fpr/tpr pair for which another one
    ## with same fpr and higher tpr exists,
    ## or one for which one with same tpr but lower fpr exists

    if (n.neg==0 || n.pos==0) {
      stop(paste("At least one positive and one negative sample are",
                 "needed to compute a cost curve."))
    }
    fpr <- fp / n.neg
    tpr <- tp / n.pos

    ## sort by fpr (ascending), in case of ties by descending tpr
    ind <- order(fpr,-tpr)

    fpr <- fpr[ind]
    tpr <- tpr[ind]
    ## for tied fprs, only the one with the highest tpr is kept
    ind <- !duplicated(fpr)
    fpr <- fpr[ind]
    tpr <- tpr[ind]

    ## for tied tprs, only keep the one with the lowest fpr
    ind <- order(-tpr,fpr)
    fpr <- fpr[ind]
    tpr <- tpr[ind]
    ind <- !duplicated(tpr)
    fpr <- fpr[ind]
    tpr <- tpr[ind]

    if (!any(0==fpr & 0==tpr)) {
      fpr <- c(0,fpr)
      tpr <- c(0,tpr)
    }
    if (!any(1==fpr & 1==tpr)) {
      fpr <- c(fpr,1)
      tpr <- c(tpr,1)
    }

    ## compute all functions
    f <- list()
    for (i in 1:length(fpr)) {
      f <- c(f, .construct.linefunct( 0, fpr[i], 1, 1-tpr[i] ))
    }

    ## compute all intersection points
    x.values <- c()
    y.values <- c()
    for (i in 1:(length(fpr)-1)) {
      for (j in (i+1):length(fpr)) {
        ans <- .intersection.point( f[[i]], f[[j]] )
        if (all(is.finite(ans))) {
          y.values.at.current.x <- c()
          for (k in 1:length(f)) {
            y.values.at.current.x <- c(y.values.at.current.x,
                                       f[[k]](ans[1]))
          }
          if (abs(ans[2] - min(y.values.at.current.x )) <
              sqrt(.Machine$double.eps)) {

            x.values <- c(x.values, ans[1])
            y.values <- c(y.values, ans[2])
          }
        }
      }
    }

    if (!any(0==x.values & 0==y.values)) {
      x.values <- c(0,x.values)
      y.values <- c(0,y.values)
    }
    if (!any(1==x.values & 0==y.values)) {
      x.values <- c(x.values,1)
      y.values <- c(y.values,0)
    }

    ind <- order( x.values)
    list( x.values[ind], y.values[ind] )
  }


.performance.cost <-
  function(predictions, labels, cutoffs, fp, tp, fn, tn,
           n.pos, n.neg, n.pos.pred, n.neg.pred, cost.fp, cost.fn) {

    n.samples <- n.pos + n.neg
    cost <- ((n.pos / n.samples) * (fn / n.pos) * cost.fn +
               (n.neg / n.samples) * (fp / n.neg) * cost.fp)
    list( cutoffs, cost )
  }

#' @name performance
#'
#' @title Function to create performance objects
#'
#' @description
#' All kinds of predictor evaluations are performed using this function.
#'
#' @details
#' Here is the list of available performance measures. Let Y and
#' \eqn{\hat{Y}}{Yhat} be random variables representing the class and the prediction for
#' a randomly drawn sample, respectively. We denote by
#' \eqn{\oplus}{+} and \eqn{\ominus}{-} the positive and
#' negative class, respectively. Further, we use the following
#' abbreviations for empirical quantities: P (# positive
#' samples), N (# negative samples), TP (# true positives), TN (# true
#' negatives), FP (# false positives), FN (# false negatives).
#' \describe{
#'  \item{\code{acc}:}{Accuracy. \eqn{P(\hat{Y}=Y)}{P(Yhat = Y)}. Estimated
#'    as: \eqn{\frac{TP+TN}{P+N}}{(TP+TN)/(P+N)}.}
#'  \item{\code{err}:}{Error rate. \eqn{P(\hat{Y}\ne Y)}{P(Yhat !=
#'                                                           Y)}. Estimated as: \eqn{\frac{FP+FN}{P+N}}{(FP+FN)/(P+N)}.}
#'  \item{\code{fpr}:}{False positive rate. \eqn{P(\hat{Y}=\oplus | Y =
#'                                                    \ominus)}{P(Yhat = + | Y = -)}. Estimated as:
#'      \eqn{\frac{FP}{N}}{FP/N}.}
#'  \item{\code{fall}:}{Fallout. Same as \code{fpr}.}
#'  \item{\code{tpr}:}{True positive
#'    rate. \eqn{P(\hat{Y}=\oplus|Y=\oplus)}{P(Yhat = + | Y = +)}. Estimated
#'    as: \eqn{\frac{TP}{P}}{TP/P}.}
#'  \item{\code{rec}:}{Recall. Same as \code{tpr}.}
#'  \item{\code{sens}:}{Sensitivity. Same as \code{tpr}.}
#'  \item{\code{fnr}:}{False negative
#'    rate. \eqn{P(\hat{Y}=\ominus|Y=\oplus)}{P(Yhat = - | Y =
#'                                                +)}. Estimated as: \eqn{\frac{FN}{P}}{FN/P}.}
#'  \item{\code{miss}:}{Miss. Same as \code{fnr}.}
#'  \item{\code{tnr}:}{True negative rate. \eqn{P(\hat{Y} =
#'                                                   \ominus|Y=\ominus)}{P(Yhat = - | Y = -)}.}
#'  \item{\code{spec}:}{Specificity. Same as \code{tnr}.}
#'  \item{\code{ppv}:}{Positive predictive
#'    value. \eqn{P(Y=\oplus|\hat{Y}=\oplus)}{P(Y = + | Yhat =
#'                                                +)}. Estimated as: \eqn{\frac{TP}{TP+FP}}{TP/(TP+FP)}.}
#'  \item{\code{prec}:}{Precision. Same as \code{ppv}.}
#'  \item{\code{npv}:}{Negative predictive
#'    value. \eqn{P(Y=\ominus|\hat{Y}=\ominus)}{P(Y = - | Yhat =
#'                                                  -)}. Estimated as: \eqn{\frac{TN}{TN+FN}}{TN/(TN+FN)}.}
#'  \item{\code{pcfall}:}{Prediction-conditioned
#'    fallout. \eqn{P(Y=\ominus|\hat{Y}=\oplus)}{P(Y = - | Yhat =
#'                                                   +)}. Estimated as: \eqn{\frac{FP}{TP+FP}}{FP/(TP+FP)}.}
#'  \item{\code{pcmiss}:}{Prediction-conditioned
#'    miss. \eqn{P(Y=\oplus|\hat{Y}=\ominus)}{P(Y = + | Yhat =
#'                                                -)}. Estimated as: \eqn{\frac{FN}{TN+FN}}{FN/(TN+FN)}.}
#'  \item{\code{rpp}:}{Rate of positive predictions. \eqn{P( \hat{Y} =
#'                                                            \oplus)}{P(Yhat = +)}. Estimated as: (TP+FP)/(TP+FP+TN+FN).}
#'  \item{\code{rnp}:}{Rate of negative predictions. \eqn{P( \hat{Y} =
#'                                                            \ominus)}{P(Yhat = -)}. Estimated as: (TN+FN)/(TP+FP+TN+FN).}
#'  \item{\code{phi}:}{Phi correlation coefficient. \eqn{\frac{TP \cdot
#'    TN - FP \cdot FN}{\sqrt{ (TP+FN) \cdot (TN+FP) \cdot (TP+FP)
#'      \cdot (TN+FN)}}}{(TP*TN -
#'                          FP*FN)/(sqrt((TP+FN)*(TN+FP)*(TP+FP)*(TN+FN)))}. Yields a
#'    number between -1 and 1, with 1 indicating a perfect
#'    prediction, 0 indicating a random prediction. Values below 0
#'    indicate a worse than random prediction.}
#'  \item{\code{mat}:}{Matthews correlation coefficient. Same as \code{phi}.}
#'  \item{\code{mi}:}{Mutual information. \eqn{I(\hat{Y},Y) := H(Y) -
#'      H(Y|\hat{Y})}{I(Yhat, Y) := H(Y) - H(Y | Yhat)}, where H is the
#'    (conditional) entropy. Entropies are estimated naively (no bias
#'                                                            correction).}
#'  \item{\code{chisq}:}{Chi square test statistic. \code{?chisq.test}
#'    for details. Note that R might raise a warning if the sample size
#'    is too small.}
#'  \item{\code{odds}:}{Odds ratio. \eqn{\frac{TP \cdot TN}{FN \cdot
#'    FP}}{(TP*TN)/(FN*FP)}. Note that odds ratio produces
#'    Inf or NA values for all cutoffs corresponding to FN=0 or
#'    FP=0. This can substantially decrease the plotted cutoff region.}
#'  \item{\code{lift}:}{Lift
#'    value. \eqn{\frac{P(\hat{Y}=\oplus|Y=\oplus)}{P(\hat{Y}=\oplus)}}{P(Yhat = + |
#'                                                                          Y = +)/P(Yhat = +)}.}
#'  \item{\code{f}:}{Precision-recall F measure (van Rijsbergen, 1979). Weighted
#'    harmonic mean of precision (P) and recall (R). \eqn{F =
#'      \frac{1}{\alpha \frac{1}{P} + (1-\alpha)\frac{1}{R}}}{F = 1/
#'        (alpha*1/P + (1-alpha)*1/R)}. If
#'    \eqn{\alpha=\frac{1}{2}}{alpha=1/2}, the mean is balanced. A
#'    frequent equivalent formulation is
#'    \eqn{F = \frac{(\beta^2+1) \cdot P \cdot R}{R + \beta^2 \cdot
#'      P}}{F = (beta^2+1) * P * R / (R + beta^2 * P)}. In this formulation, the
#'      mean is balanced if \eqn{\beta=1}{beta=1}. Currently, ROCR only accepts
#'      the alpha version as input (e.g. \eqn{\alpha=0.5}{alpha=0.5}). If no 
#'      value for alpha is given, the mean will be balanced by default.}
#'  \item{\code{rch}:}{ROC convex hull. A ROC (=\code{tpr} vs \code{fpr}) curve 
#'    with concavities (which represent suboptimal choices of cutoff) removed 
#'    (Fawcett 2001). Since the result is already a parametric performance 
#'    curve, it cannot be used in combination with other measures.}
#'  \item{\code{auc}:}{Area under the ROC curve. This is equal to the value of the
#'    Wilcoxon-Mann-Whitney test statistic and also the probability that the
#'    classifier will score are randomly drawn positive sample higher than a
#'    randomly drawn negative sample. Since the output of
#'    \code{auc} is cutoff-independent, this
#'    measure cannot be combined with other measures into a parametric
#'    curve. The partial area under the ROC curve up to a given false
#'    positive rate can be calculated by passing the optional parameter
#'    \code{fpr.stop=0.5} (or any other value between 0 and 1) to 
#'    \code{performance}.}
#'  \item{\code{aucpr}:}{Area under the Precision/Recall curve. Since the output
#'    of \code{aucpr} is cutoff-independent, this measure cannot be combined 
#'    with other measures into a parametric curve.}
#'  \item{\code{prbe}:}{Precision-recall break-even point. The cutoff(s) where
#'    precision and recall are equal. At this point, positive and negative
#'    predictions are made at the same rate as their prevalence in the
#'    data. Since the output of
#'    \code{prbe} is just a cutoff-independent scalar, this
#'    measure cannot be combined with other measures into a parametric curve.}
#'  \item{\code{cal}:}{Calibration error. The calibration error is the
#'    absolute difference between predicted confidence and actual reliability. This
#'    error is estimated at all cutoffs by sliding a window across the
#'    range of possible cutoffs. The default window size of 100 can be
#'    adjusted by passing the optional parameter \code{window.size=200}
#'    to \code{performance}. E.g., if for several
#'    positive samples the output of the classifier is around 0.75, you might
#'    expect from a well-calibrated classifier that the fraction of them
#'    which is correctly predicted as positive is also around 0.75. In a
#'    well-calibrated classifier, the probabilistic confidence estimates
#'    are realistic. Only for use with
#'    probabilistic output (i.e. scores between 0 and 1).}
#'  \item{\code{mxe}:}{Mean cross-entropy. Only for use with
#'    probabilistic output. \eqn{MXE :=-\frac{1}{P+N}( \sum_{y_i=\oplus}
#'                                                    ln(\hat{y}_i) + \sum_{y_i=\ominus} ln(1-\hat{y}_i))}{MXE := - 1/(P+N) \sum_{y_i=+}
#'                                                      ln(yhat_i) + \sum_{y_i=-} ln(1-yhat_i)}. Since the output of
#'    \code{mxe} is just a cutoff-independent scalar, this
#'    measure cannot be combined with other measures into a parametric curve.}
#'  \item{\code{rmse}:}{Root-mean-squared error. Only for use with
#'    numerical class labels. \eqn{RMSE:=\sqrt{\frac{1}{P+N}\sum_i (y_i
#'                                                                  - \hat{y}_i)^2}}{RMSE := sqrt(1/(P+N) \sum_i (y_i -
#'                                                                                                                  yhat_i)^2)}. Since the output of
#'    \code{rmse} is just a cutoff-independent scalar, this
#'    measure cannot be combined with other measures into a parametric curve.}
#'  \item{\code{sar}:}{Score combinining performance measures of different
#'    characteristics, in the attempt of creating a more "robust"
#'    measure (cf. Caruana R., ROCAI2004):
#'      SAR = 1/3 * ( Accuracy + Area under the ROC curve + Root
#'                    mean-squared error ).}
#'  \item{\code{ecost}:}{Expected cost. For details on cost curves,
#'    cf. Drummond&Holte 2000,2004. \code{ecost} has an obligatory x
#'    axis, the so-called 'probability-cost function'; thus it cannot be
#'    combined with other measures. While using \code{ecost} one is
#'    interested in the lower envelope of a set of lines, it might be
#'    instructive to plot the whole set of lines in addition to the lower
#'    envelope. An example is given in \code{demo(ROCR)}.}
#'  \item{\code{cost}:}{Cost of a classifier when
#'    class-conditional misclassification costs are explicitly given.
#'    Accepts the optional parameters \code{cost.fp} and
#'    \code{cost.fn}, by which the costs for false positives and
#'    negatives can be adjusted, respectively. By default, both are set
#'    to 1.}
#' }
#'
#' @note
#' Here is how to call \code{performance()} to create some standard
#' evaluation plots:
#' \describe{
#'   \item{ROC curves:}{measure="tpr", x.measure="fpr".}
#'   \item{Precision/recall graphs:}{measure="prec", x.measure="rec".}
#'   \item{Sensitivity/specificity plots:}{measure="sens", x.measure="spec".}
#'   \item{Lift charts:}{measure="lift", x.measure="rpp".}
#' }
#'
#' @param prediction.obj An object of class \code{prediction}.
#' @param measure Performance measure to use for the evaluation. A complete list
#'   of the performance measures that are available for \code{measure} and
#'   \code{x.measure} is given in the 'Details' section.
#' @param x.measure A second performance measure. If different from the default,
#'   a two-dimensional curve, with \code{x.measure} taken to be the unit in
#'   direction of the x axis, and \code{measure} to be the unit in direction of
#'   the y axis, is created. This curve is parametrized with the cutoff.
#' @param ... Optional arguments (specific to individual performance measures).
#'
#' @return An S4 object of class \code{performance}.
#'
#' @references
#' A detailed list of references can be found on the ROCR homepage at
#' \url{https://ipa-tys.github.io/ROCR/}.
#'
#' @author
#' Tobias Sing \email{tobias.sing@gmail.com}, Oliver Sander
#' \email{osander@gmail.com}
#'
#' @seealso
#' \code{\link{prediction}},
#' \code{\link{prediction-class}},
#' \code{\link{performance-class}},
#' \code{\link{plot.performance}}
#'
#' @export
#'
#' @examples
#' # computing a simple ROC curve (x-axis: fpr, y-axis: tpr)
#' library(ROCR)
#' data(ROCR.simple)
#' pred <- prediction( ROCR.simple$predictions, ROCR.simple$labels)
#' pred
#' perf <- performance(pred,"tpr","fpr")
#' perf
#' plot(perf)
#'
#' # precision/recall curve (x-axis: recall, y-axis: precision)
#' perf <- performance(pred, "prec", "rec")
#' perf
#' plot(perf)
#'
#' # sensitivity/specificity curve (x-axis: specificity,
#' # y-axis: sensitivity)
#' perf <- performance(pred, "sens", "spec")
#' perf
#' plot(perf)
performance <- function(prediction.obj,
                        measure,
                        x.measure="cutoff",
                        ...) {

  ## define the needed environments
  envir.list <- .define.environments()
  long.unit.names <- envir.list$long.unit.names
  function.names <- envir.list$function.names
  obligatory.x.axis <- envir.list$obligatory.x.axis
  optional.arguments <- envir.list$optional.arguments
  default.values <- envir.list$default.values

  ## abort in case of misuse
  if (!is(prediction.obj, "prediction") ||
      !exists(measure, where=long.unit.names, inherits=FALSE) ||
      !exists(x.measure, where=long.unit.names, inherits=FALSE)) {
    stop(paste("Wrong argument types: First argument must be of type",
               "'prediction'; second and optional third argument must",
               "be available performance measures!"))
  }

  ## abort, if attempt is made to use a measure that has an obligatory
  ## x.axis as the x.measure (cannot be combined)
  if (exists( x.measure, where=obligatory.x.axis, inherits=FALSE )) {
    message <- paste("The performance measure",
                     x.measure,
                     "can only be used as 'measure', because it has",
                     "the following obligatory 'x.measure':\n",
                     get( x.measure, envir=obligatory.x.axis))
    stop(message)
  }

  ## if measure is a performance measure with obligatory x.axis, then
  ## enforce this axis:
  if (exists( measure, where=obligatory.x.axis, inherits=FALSE )) {
    x.measure <- get( measure, envir=obligatory.x.axis )
  }

  if (x.measure == "cutoff" ||
      exists( measure, where=obligatory.x.axis, inherits=FALSE )) {

    ## fetch from '...' any optional arguments for the performance
    ## measure at hand that are given, otherwise fill up the default values
    optional.args <- list(...)
    argnames <- c()
    if ( exists( measure, where=optional.arguments, inherits=FALSE )) {
      argnames <- get( measure, envir=optional.arguments )
      default.arglist <- list()
      for (i in 1:length(argnames)) {
        default.arglist <- c(default.arglist,
                             get(paste(measure,":",argnames[i],sep=""),
                                 envir=default.values, inherits=FALSE))
      }
      names(default.arglist) <- argnames

      for (i in 1:length(argnames)) {
        templist <- list(optional.args,
                         default.arglist[[i]])
        names(templist) <- c('arglist', argnames[i])

        optional.args <- do.call('.farg', templist)
      }
    }
    optional.args <- .select.args( optional.args, argnames )

    ## determine function name
    function.name <- get( measure, envir=function.names )

    ## for each x-validation run, compute the requested performance measure
    x.values <- list()
    y.values <- list()
    for (i in 1:length( prediction.obj@predictions )) {
      argumentlist <- .sarg(optional.args,
                            predictions= prediction.obj@predictions[[i]],
                            labels= prediction.obj@labels[[i]],
                            cutoffs= prediction.obj@cutoffs[[i]],
                            fp= prediction.obj@fp[[i]],
                            tp= prediction.obj@tp[[i]],
                            fn= prediction.obj@fn[[i]],
                            tn= prediction.obj@tn[[i]],
                            n.pos= prediction.obj@n.pos[[i]],
                            n.neg= prediction.obj@n.neg[[i]],
                            n.pos.pred= prediction.obj@n.pos.pred[[i]],
                            n.neg.pred= prediction.obj@n.neg.pred[[i]])

      ans <- do.call( function.name, argumentlist )

      if (!is.null(ans[[1]])) x.values <- c( x.values, list( ans[[1]] ))
      y.values <- c( y.values, list( ans[[2]] ))
    }

    if (! (length(x.values)==0 || length(x.values)==length(y.values)) ) {
      stop("Consistency error.")
    }

    ## create a new performance object
    return( new("performance",
                x.name       = get( x.measure, envir=long.unit.names ),
                y.name       = get( measure, envir=long.unit.names ),
                alpha.name   = "none",
                x.values     = x.values,
                y.values     = y.values,
                alpha.values = list() ))
  } else {
    perf.obj.1 <- performance( prediction.obj, measure=x.measure, ... )
    perf.obj.2 <- performance( prediction.obj, measure=measure, ... )
    return( .combine.performance.objects( perf.obj.1, perf.obj.2 ) )
  }
}

#' @importFrom stats approxfun
.combine.performance.objects <- function( p.obj.1, p.obj.2 ) {
  ## some checks for misusage (in any way, this function is
  ## only for internal use)
  if ( p.obj.1@x.name != p.obj.2@x.name ) {
    stop("Error: Objects need to have identical x axis.")
  }
  if ( p.obj.1@alpha.name != "none" || p.obj.2@alpha.name != "none") {
    stop("Error: At least one of the two objects has already been merged.")
  }
  if (length(p.obj.1@x.values) != length(p.obj.2@x.values)) {
    stop(paste("Only performance objects with identical number of",
               "cross-validation runs can be combined."))
  }

  x.values <- list()
  x.name <- p.obj.1@y.name
  y.values <- list()
  y.name <- p.obj.2@y.name
  alpha.values <- list()
  alpha.name <- p.obj.1@x.name

  for (i in 1:length( p.obj.1@x.values )) {
    x.values.1 <- p.obj.1@x.values[[i]]
    y.values.1 <- p.obj.1@y.values[[i]]
    x.values.2 <- p.obj.2@x.values[[i]]
    y.values.2 <- p.obj.2@y.values[[i]]

    ## cutoffs of combined object = merged cutoffs of simple objects
    cutoffs <- sort( unique( c(x.values.1, x.values.2)), decreasing=TRUE )

    ## calculate y.values at cutoffs using step function
    y.values.int.1 <- stats::approxfun(x.values.1, y.values.1,
                                       method="constant",f=1,rule=2)(cutoffs)
    y.values.int.2 <- stats::approxfun(x.values.2, y.values.2,
                                       method="constant",f=1,rule=2)(cutoffs)

    ## 'approxfun' ignores NA and NaN
    objs <- list( y.values.int.1, y.values.int.2)
    objs.x <- list( x.values.1, x.values.2 )
    na.cutoffs.1.bool <- is.na( y.values.1) & !is.nan( y.values.1 )
    nan.cutoffs.1.bool <- is.nan( y.values.1)
    na.cutoffs.2.bool <- is.na( y.values.2) & !is.nan( y.values.2 )
    nan.cutoffs.2.bool <- is.nan( y.values.2)
    bools <- list(na.cutoffs.1.bool, nan.cutoffs.1.bool,
                  na.cutoffs.2.bool, nan.cutoffs.2.bool)
    values <- c(NA,NaN,NA,NaN)

    for (j in 1:4) {
      for (k in which(bools[[j]])) {
        interval.max <- objs.x[[ ceiling(j/2) ]][k]
        interval.min <- -Inf
        if (k < length(objs.x[[ ceiling(j/2) ]])) {
          interval.min <- objs.x[[ ceiling(j/2) ]][k+1]
        }
        objs[[ ceiling(j/2) ]][cutoffs <= interval.max &
                                 cutoffs > interval.min ] <- values[j]
      }
    }

    alpha.values <- c(alpha.values, list(cutoffs))
    x.values <- c(x.values, list(objs[[1]]))
    y.values <- c(y.values, list(objs[[2]]))
  }

  return( new("performance",
              x.name=x.name, y.name=y.name,
              alpha.name=alpha.name, x.values=x.values,
              y.values=y.values, alpha.values=alpha.values))
}

.define.environments <- function() {
  ## There are five environments: long.unit.names, function.names,
  ## obligatory.x.axis, optional.arguments, default.values

  ## Define long names corresponding to the measure abbreviations.
  long.unit.names <- new.env()
  assign("none","None", envir=long.unit.names)
  assign("cutoff", "Cutoff", envir=long.unit.names)
  assign("acc", "Accuracy", envir=long.unit.names)
  assign("err", "Error Rate", envir=long.unit.names)
  assign("fpr", "False positive rate", envir=long.unit.names)
  assign("tpr", "True positive rate", envir=long.unit.names)
  assign("rec", "Recall", envir=long.unit.names)
  assign("sens", "Sensitivity", envir=long.unit.names)
  assign("fnr", "False negative rate", envir=long.unit.names)
  assign("tnr", "True negative rate", envir=long.unit.names)
  assign("spec", "Specificity", envir=long.unit.names)
  assign("ppv", "Positive predictive value", envir=long.unit.names)
  assign("prec", "Precision", envir=long.unit.names)
  assign("npv", "Negative predictive value", envir=long.unit.names)
  assign("fall", "Fallout", envir=long.unit.names)
  assign("miss", "Miss", envir=long.unit.names)
  assign("pcfall", "Prediction-conditioned fallout", envir=long.unit.names)
  assign("pcmiss", "Prediction-conditioned miss", envir=long.unit.names)
  assign("rpp", "Rate of positive predictions", envir=long.unit.names)
  assign("rnp", "Rate of negative predictions", envir=long.unit.names)
  assign("auc","Area under the ROC curve", envir=long.unit.names)
  assign("aucpr","Area under the Precision/Recall curve", envir=long.unit.names)
  assign("cal", "Calibration error", envir=long.unit.names)
  assign("mwp", "Median window position", envir=long.unit.names)
  assign("prbe","Precision/recall break-even point", envir=long.unit.names)
  assign("rch", "ROC convex hull", envir=long.unit.names)
  assign("mxe", "Mean cross-entropy", envir=long.unit.names)
  assign("rmse","Root-mean-square error", envir=long.unit.names)
  assign("phi", "Phi correlation coefficient", envir=long.unit.names)
  assign("mat","Matthews correlation coefficient", envir=long.unit.names)
  assign("mi", "Mutual information", envir=long.unit.names)
  assign("chisq", "Chi-square test statistic", envir=long.unit.names)
  assign("odds","Odds ratio", envir=long.unit.names)
  assign("lift", "Lift value", envir=long.unit.names)
  assign("f","Precision-Recall F measure", envir=long.unit.names)
  assign("sar", "SAR", envir=long.unit.names)
  assign("ecost", "Expected cost", envir=long.unit.names)
  assign("cost", "Explicit cost", envir=long.unit.names)

  ## Define function names corresponding to the measure abbreviations.
  function.names <- new.env()
  assign("acc", ".performance.accuracy", envir=function.names)
  assign("err", ".performance.error.rate", envir=function.names)
  assign("fpr", ".performance.false.positive.rate", envir=function.names)
  assign("tpr", ".performance.true.positive.rate", envir=function.names)
  assign("rec", ".performance.true.positive.rate", envir=function.names)
  assign("sens", ".performance.true.positive.rate", envir=function.names)
  assign("fnr", ".performance.false.negative.rate", envir=function.names)
  assign("tnr", ".performance.true.negative.rate", envir=function.names)
  assign("spec", ".performance.true.negative.rate", envir=function.names)
  assign("ppv", ".performance.positive.predictive.value",
         envir=function.names)
  assign("prec", ".performance.positive.predictive.value",
         envir=function.names)
  assign("npv", ".performance.negative.predictive.value",
         envir=function.names)
  assign("fall", ".performance.false.positive.rate", envir=function.names)
  assign("miss", ".performance.false.negative.rate", envir=function.names)
  assign("pcfall", ".performance.prediction.conditioned.fallout",
         envir=function.names)
  assign("pcmiss", ".performance.prediction.conditioned.miss",
         envir=function.names)
  assign("rpp", ".performance.rate.of.positive.predictions",
         envir=function.names)
  assign("rnp", ".performance.rate.of.negative.predictions",
         envir=function.names)
  assign("auc", ".performance.auc", envir=function.names)
  assign("aucpr", ".performance.aucpr", envir=function.names)
  assign("cal", ".performance.calibration.error", envir=function.names)
  assign("prbe", ".performance.precision.recall.break.even.point",
         envir=function.names)
  assign("rch", ".performance.rocconvexhull", envir=function.names)
  assign("mxe", ".performance.mean.cross.entropy", envir=function.names)
  assign("rmse", ".performance.root.mean.squared.error",
         envir=function.names)
  assign("phi", ".performance.phi", envir=function.names)
  assign("mat", ".performance.phi", envir=function.names)
  assign("mi", ".performance.mutual.information", envir=function.names)
  assign("chisq", ".performance.chisq", envir=function.names)
  assign("odds", ".performance.odds.ratio", envir=function.names)
  assign("lift", ".performance.lift", envir=function.names)
  assign("f", ".performance.f", envir=function.names)
  assign("sar", ".performance.sar", envir=function.names)
  assign("ecost", ".performance.expected.cost", envir=function.names)
  assign("cost", ".performance.cost", envir=function.names)

  ## If a measure comes along with an obligatory x axis (including "none"),
  ## list it here.
  obligatory.x.axis <- new.env()
  assign("mxe", "none", envir=obligatory.x.axis)
  assign("rmse", "none", envir=obligatory.x.axis)
  assign("prbe", "none", envir=obligatory.x.axis)
  assign("auc", "none", envir=obligatory.x.axis)
  assign("aucpr", "none", envir=obligatory.x.axis)
  assign("rch","none", envir=obligatory.x.axis)
  ## ecost requires probability cost function as x axis, which is handled
  ## implicitly, not as an explicit performance measure.
  assign("ecost","none", envir=obligatory.x.axis)

  ## If a measure has optional arguments, list the names of the
  ## arguments here.
  optional.arguments <- new.env()
  assign("cal", "window.size", envir=optional.arguments)
  assign("f", "alpha", envir=optional.arguments)
  assign("cost", c("cost.fp", "cost.fn"), envir=optional.arguments)
  assign("auc", "fpr.stop", envir=optional.arguments)

  ## If a measure has additional arguments, list the default values
  ## for them here. Naming convention: e.g. "cal" has an optional
  ## argument "window.size" the key to use here is "cal:window.size"
  ## (colon as separator)
  default.values <- new.env()
  assign("cal:window.size", 100, envir=default.values)
  assign("f:alpha", 0.5, envir=default.values)
  assign("cost:cost.fp", 1, envir=default.values)
  assign("cost:cost.fn", 1, envir=default.values)
  assign("auc:fpr.stop", 1, envir=default.values)

  list(long.unit.names=long.unit.names, function.names=function.names,
       obligatory.x.axis=obligatory.x.axis,
       optional.arguments=optional.arguments,
       default.values=default.values)
}

context("plot")
test_that("plot:",{
    some.predictions <- c(0.02495517, 0.92535646,
                          0.86251887, 0.80946685,
                          0.70922858, 0.69762824,
                          0.50604485, 0.25446810,
                          0.10837728, 0.07250349)
    some.labels <- c(0,1,1,0,1,1,0,1,0,0)
    
    .get.performance <- function(pred) {
        
        tpr <- performance(pred, "tpr")
        fpr <- performance(pred, "fpr")
        acc <- performance(pred, "acc")
        err <- performance(pred, "err")
        
        rec <- performance(pred, "rec")
        sens<- performance(pred, "sens")
        fnr <- performance(pred, "fnr")
        tnr <- performance(pred, "tnr")
        spec<- performance(pred, "spec")
        ppv <- performance(pred, "ppv")
        prec<- performance(pred, "prec")
        npv <- performance(pred, "npv")
        
        fall<- performance(pred, "fall")
        miss<- performance(pred, "miss")
        pcfall <- performance(pred, "pcfall")
        pcmiss <- performance(pred, "pcmiss")
        rpp <- performance(pred, "rpp")
        rnp <- performance(pred, "rnp")
        
        auc <- performance(pred, "auc")
        prbe<- performance(pred, "prbe")
        rch <- performance(pred, "rch")
        
        mxe <- performance(pred, "mxe")
        rmse<- performance(pred, "rmse")
        
        phi <- performance(pred, "phi")
        mat <- performance(pred, "mat")
        mi  <- performance(pred, "mi")
        chisq<- performance(pred, "chisq")
        odds<- performance(pred, "odds")
        lift<- performance(pred, "lift")
        f   <- performance(pred, "f")
        sar <- performance(pred,"sar")
        ecost  <- performance(pred, "ecost")
        cost  <- performance(pred, "cost")
        return(list(tpr=tpr, fpr=fpr, acc=acc, err=err,
                    rec=rec, sens=sens, fnr=fnr, tnr=tnr,
                    spec=spec, ppv=ppv, prec=prec, npv=npv, 
                    fall=fall, miss=miss, pcfall=pcfall, pcmiss=pcmiss, rpp=rpp, rnp=rnp,
                    auc=auc, prbe=prbe, rch=rch, mxe=mxe, 
                    rmse=rmse, phi=phi, mat=mat, mi=mi, chisq=chisq, odds=odds,
                    lift=lift, f=f, sar=sar, ecost=ecost, cost=cost))
        
    }
    pred <- prediction(some.predictions, some.labels)
    measures <- expect_warning(.get.performance(pred),
                               "Chi-squared approximation may be incorrect")
    actual1 <- measures[[1]]
    expect_error(plot(measures[[1]], colorize = TRUE),
                 "Threshold coloring or labeling cannot be performed")
    for(i in seq_along(measures)){
        if(names(measures[i]) %in% c("auc","mxe","rmse")){
            expect_error(plot(measures[[i]]))
        }
    }
    
    data(ROCR.hiv)
    pp <- ROCR.hiv$hiv.svm$predictions
    ll <- ROCR.hiv$hiv.svm$labels
    pred <- prediction(pp, ll)
    expect_error(ROCR:::.combine.performance.objects(actual1,performance(pred, "fpr")),
                 "Only performance objects with identical number of cross-validation")
    # plot failures
    perf <- performance(pred, "tpr", "fpr")
    perf@x.values <- list(c(1))
    expect_error(plot(perf),
                 "Performance object cannot be plotted")
    perf <- performance(pred, "tpr", "fpr")
    perf@y.values <- list(c(1))
    expect_error(plot(perf),
                 "Performance object cannot be plotted")
    perf <- performance(pred, "tpr", "fpr")
    perf@alpha.values <- list()
    expect_null({
        plot <- plot(perf) # no error
    })
    expect_error(plot(perf,colorize = TRUE),
                 "Threshold coloring or labeling cannot be performed")
    expect_error(plot(perf,print.cutoffs.at = 0.5),
                 "Threshold coloring or labeling cannot be performed")
    perf <- performance(pred, "tpr", "fpr")
    expect_null({
        plot <- plot(perf,avg = "horizontal") # no error
    })
    expect_error(plot(perf,avg = "horizontal", colorize=TRUE),
                 "Threshold coloring or labeling is only")
    expect_error(plot(perf,avg = "horizontal", print.cutoffs.at=0.5),
                 "Threshold coloring or labeling is only")
    #
    perf <- performance(pred, "tpr", "fpr")
    expect_null({
        plot(perf, avg= "threshold", colorize=TRUE, lwd= 3,
             main= "With ROCR you can produce standard plots like ROC curves ...")
        plot(perf, lty=3, col="grey78", add=TRUE)
    })
    expect_null({
        plot.performance(perf, avg= "threshold", colorize=TRUE, lwd= 3,
             main= "With ROCR you can produce standard plots like ROC curves ...")
        plot.performance(perf, lty=3, col="grey78", add=TRUE)
    })
    perf <- performance(pred, "prec", "rec")
    expect_null({
        plot(perf, avg= "threshold", colorize=TRUE, lwd= 3,
             main= "... Precision/Recall graphs ...")
        plot(perf, lty=3, col="grey78", add=TRUE)
    })
    perf <- performance(pred, "sens", "spec")
    expect_null({
        plot(perf, avg= "threshold", colorize=TRUE, lwd= 3,
             main="... Sensitivity/Specificity plots ...")
        plot(perf, lty=3, col="grey78", add=TRUE)
    })
    perf <- performance(pred, "lift", "rpp")
    expect_null({
        plot(perf, avg= "threshold", colorize=TRUE, lwd= 3,
             main= "... and Lift charts.")
        plot(perf, lty=3, col="grey78", add=TRUE)
    })
    
    perf <- performance(pred, "tpr", "fpr")
    expect_null({
        plot(perf, avg= "threshold", colorize=TRUE, lwd= 3,
             main= "With ROCR you can produce standard plots like ROC curves ...",
             downsampling = 0.5)
    })
    expect_null({
        plot(perf, avg= "threshold", colorize=TRUE, lwd= 3,
             main= "With ROCR you can produce standard plots like ROC curves ...",
             downsampling = 0.9)
    })
    expect_null({
        plot(perf, avg= "threshold", colorize=TRUE, lwd= 3,
             main= "With ROCR you can produce standard plots like ROC curves ...",
             downsampling = 1)
    })
    
    data(ROCR.xval)
    pp <- ROCR.xval$predictions
    ll <- ROCR.xval$labels
    pred <- prediction(pp,ll)
    perf <- performance(pred,'tpr','fpr')
    
    expect_null({
        plot(perf, colorize=TRUE, lwd=2,
             main='ROC curves from 10-fold cross-validation')
    })
    expect_null({
        plot(perf, avg='vertical', spread.estimate='stderror',lwd=3,
             main='Vertical averaging + 1 standard error',col='blue')
    })
    expect_null({
        plot(perf, avg='horizontal', spread.estimate='stderror',lwd=3,
             main='Horizontal averaging + boxplots',col='blue')
    })
    expect_null({
        plot(perf, avg='horizontal', spread.estimate='boxplot',lwd=3,
             main='Horizontal averaging + boxplots',col='blue')
    })
    expect_null({
        plot(perf, avg='vertical', spread.estimate='boxplot',lwd=3,
             main='Horizontal averaging + boxplots',col='blue')
    })
    expect_null({
        plot(perf, avg='threshold', spread.estimate='stddev',lwd=2,
             main='Threshold averaging + 1 standard deviation',colorize=TRUE)
    })
    expect_null({
        plot(perf, avg='threshold', spread.estimate='boxplot',lwd=2,
             main='Threshold averaging + 1 standard deviation',colorize=TRUE)
    })
    expect_null({
        plot(perf, avg='threshold', spread.estimate='boxplot',lwd=2,
             main='Threshold averaging + 1 standard deviation',colorize=TRUE,
             colorkey.pos="top")
    })
    expect_null({
        plot(perf,
             print.cutoffs.at=seq(0,1,by=0.2),
             text.cex=0.8,
             text.y=lapply(as.list(seq(0,0.5,by=0.05)),
                           function(x) { rep(x,length(perf@x.values[[1]])) } ),
             col= as.list(terrain.colors(10)),
             text.col= as.list(terrain.colors(10)), 
             points.col= as.list(terrain.colors(10)), 
             main= "Cutoff stability")
    })
    
   
    
    ############################################################################
    # removed because vdiffr is not available on mac
    ############################################################################
    
    # vdiffr
    # skip_on_ci()
    # skip_on_os("mac")
    # skip_if_not_installed("vdiffr")
    # for(i in seq_along(measures)){
    #     if(!(names(measures[i]) %in% c("auc","mxe","rmse"))){
    #         vdiffr::expect_doppelganger(names(measures[i]), plot(measures[[i]]))
    #     } else {
    #         expect_error(plot(measures[[i]]))
    #     }
    # }
    # 
    # data(ROCR.hiv)
    # pp <- ROCR.hiv$hiv.svm$predictions
    # ll <- ROCR.hiv$hiv.svm$labels
    # pred <- prediction(pp, ll)
    # expect_error(ROCR:::.combine.performance.objects(actual1,performance(pred, "fpr")),
    #              "Only performance objects with identical number of cross-validation")
    # perf <- performance(pred, "tpr", "fpr")
    # vdiffr::expect_doppelganger("ROC-curve",{
    #     plot(perf, avg= "threshold", colorize=TRUE, lwd= 3,
    #          main= "With ROCR you can produce standard plots like ROC curves ...")
    #     plot(perf, lty=3, col="grey78", add=TRUE)
    # })
    # perf <- performance(pred, "prec", "rec")
    # vdiffr::expect_doppelganger("Precision-Recall-graph",{
    #     plot(perf, avg= "threshold", colorize=TRUE, lwd= 3,
    #          main= "... Precision/Recall graphs ...")
    #     plot(perf, lty=3, col="grey78", add=TRUE)
    # })
    # perf <- performance(pred, "sens", "spec")
    # vdiffr::expect_doppelganger("Sensitivity-Specificity-plots",{
    #     plot(perf, avg= "threshold", colorize=TRUE, lwd= 3,
    #          main="... Sensitivity/Specificity plots ...")
    #     plot(perf, lty=3, col="grey78", add=TRUE)
    # })
    # perf <- performance(pred, "lift", "rpp")
    # vdiffr::expect_doppelganger("lift-chart",{
    #     plot(perf, avg= "threshold", colorize=TRUE, lwd= 3,
    #          main= "... and Lift charts.")
    #     plot(perf, lty=3, col="grey78", add=TRUE)
    # })
    # 
    # perf <- performance(pred, "tpr", "fpr")
    # vdiffr::expect_doppelganger("ROC-curve-downsampling1",{
    #     plot(perf, avg= "threshold", colorize=TRUE, lwd= 3,
    #          main= "With ROCR you can produce standard plots like ROC curves ...",
    #          downsampling = 0.5)
    # })
    # vdiffr::expect_doppelganger("ROC-curve-downsampling2",{
    #     plot(perf, avg= "threshold", colorize=TRUE, lwd= 3,
    #          main= "With ROCR you can produce standard plots like ROC curves ...",
    #          downsampling = 0.9)
    # })
    # vdiffr::expect_doppelganger("ROC-curve-downsampling3",{
    #     plot(perf, avg= "threshold", colorize=TRUE, lwd= 3,
    #          main= "With ROCR you can produce standard plots like ROC curves ...",
    #          downsampling = 1)
    # })
    # expect_error(plot(perf, avg= "threshold", colorize=TRUE, lwd= 3,
    #                   main= "With ROCR you can produce standard plots like ROC curves ...",
    #                   downsampling = 1.1),
    #              "'from' must be a finite number")
    # dev.off()
    # 
    # data(ROCR.xval)
    # pp <- ROCR.xval$predictions
    # ll <- ROCR.xval$labels
    # pred <- prediction(pp,ll)
    # perf <- performance(pred,'tpr','fpr')
    # 
    # vdiffr::expect_doppelganger("ROC-cross-valid",{
    #     plot(perf, colorize=TRUE, lwd=2,
    #          main='ROC curves from 10-fold cross-validation')
    # })
    # vdiffr::expect_doppelganger("ROC-vertical-avg",{
    #     plot(perf, avg='vertical', spread.estimate='stderror',lwd=3,
    #          main='Vertical averaging + 1 standard error',col='blue')
    # })
    # vdiffr::expect_doppelganger("ROC-horizontal-avg",{
    #     plot(perf, avg='horizontal', spread.estimate='boxplot',lwd=3,
    #          main='Horizontal averaging + boxplots',col='blue')
    # })
    # vdiffr::expect_doppelganger("ROC-vertical-avg-box",{
    #     plot(perf, avg='vertical', spread.estimate='boxplot',lwd=3,
    #          main='Horizontal averaging + boxplots',col='blue')
    # })
    # vdiffr::expect_doppelganger("ROC-threshold-avg",{
    #     plot(perf, avg='threshold', spread.estimate='stddev',
    #                         lwd=2,
    #          main='Threshold averaging + 1 standard deviation',colorize=TRUE)
    # })
    
})

context("simple")
test_that("simple:",{
  some.predictions <- c(0.02495517, 0.92535646,
                        0.86251887, 0.80946685,
                        0.70922858, 0.69762824,
                        0.50604485, 0.25446810,
                        0.10837728, 0.07250349)
  some.labels <- c(0,1,1,0,1,1,0,1,0,0)
  
  tp.reference <- c(0, 1, 2, 2, 3, 4, 4, 5, 5, 5, 5)
  fp.reference <- c(0, 0, 0, 1, 1, 1, 2, 2, 3, 4, 5)
  
  pp.reference <- c(0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10)
  np.reference <- c(10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0)
  p.reference <- rep(5, 11)
  n.reference <- rep(5, 11)
  
  tn.reference <- n.reference-fp.reference
  fn.reference <- p.reference-tp.reference
  
  # manually calculated reference measures
  rpp.reference <- c(0,0.1,0.2,0.3,0.4,0.5,0.6,0.7,0.8,0.9,1.0)
  rnp.reference <- c(1.0, 0.9, 0.8, 0.7, 0.6, 0.5, 0.4, 0.3, 0.2, 0.1, 0.0)
  
  tpr.reference <- c(0.0, 0.2, 0.4, 0.4, 0.6, 0.8, 0.8, 1.0, 1.0, 1.0, 1.0)
  fpr.reference <- c(0.0, 0.0, 0.0, 0.2, 0.2, 0.2, 0.4, 0.4, 0.6, 0.8, 1.0)
  acc.reference <- c(0.5, 0.6, 0.7, 0.6, 0.7, 0.8, 0.7, 0.8, 0.7, 0.6, 0.5)
  err.reference <- c(0.5, 0.4, 0.3, 0.4, 0.3, 0.2, 0.3, 0.2, 0.3, 0.4, 0.5)
  
  rec.reference <- tpr.reference
  sens.reference<- tpr.reference
  fnr.reference <- c(1.0, 0.8, 0.6, 0.6, 0.4, 0.2, 0.2, 0.0, 0.0, 0.0, 0.0) 
  tnr.reference <- c(1.0, 1.0, 1.0, 0.8, 0.8, 0.8, 0.6, 0.6, 0.4, 0.2, 0.0)
  spec.reference<- tnr.reference
  
  ppv.reference <- c(0/0, 1/1, 2/2, 2/3, 3/4, 4/5, 4/6, 5/7, 5/8, 5/9, 5/10)
  npv.reference <- c(5/10, 5/9, 5/8, 4/7, 4/6, 4/5, 3/4, 3/3, 2/2, 1/1, 0/0)
  prec.reference<- ppv.reference
  
  fall.reference <- fpr.reference
  miss.reference <- fnr.reference
  pcfall.reference <- c(0/0, 0/1, 0/2, 1/3, 1/4, 1/5, 2/6, 2/7, 3/8, 4/9, 5/10)
  pcmiss.reference <- c(5/10, 4/9, 3/8, 3/7, 2/6, 1/5, 1/4, 0/3, 0/2, 0/1, 0/0)
  
  auc.reference <- 0.84
  aucpr.reference <- 0.8814286
  
  cal.reference <- c()
  ind <- rev(order(some.predictions))
  sorted.predictions <- some.predictions[ind]
  sorted.labels <- some.labels[ind]
  for (i in 1:8) {
    mean.pred <- mean( sorted.predictions[i:(i+2)] )
    frac.pos <- sum( sorted.labels[i:(i+2)] ) / 3
    cal.reference <- c(cal.reference, abs( mean.pred - frac.pos ))
  }
  
  prbe.reference<- 0.8
  prbe.reference.x <- 0.69762824
  rch.reference.x <- fpr.reference[c(1,3,6,8,11)]
  rch.reference.y <- tpr.reference[c(1,3,6,8,11)]
  mxe.reference <- -(1/length(some.predictions)) * sum(some.labels*log(some.predictions) +
                                                         (1-some.labels)*log(1-some.predictions))
  rmse.reference <- sqrt((1/length(some.predictions)) * sum((some.predictions-some.labels)^2))
  
  phi.reference <- (tp.reference*tn.reference-fp.reference*fn.reference) /
    sqrt(p.reference*n.reference*pp.reference*np.reference)
  mat.reference <- phi.reference
  
  my.log2 <- function( x ) {
    ans <- log2(x)
    ans[ ans==-Inf ] <- 0
    ans
  }
  
  mi.reference <- (tn.reference * my.log2( tn.reference / (n.reference*np.reference)) +
                     fn.reference*my.log2(fn.reference/(np.reference*p.reference)) +
                     fp.reference*my.log2(fp.reference/(n.reference*pp.reference)) +
                     tp.reference*my.log2(tp.reference/(p.reference*pp.reference))) / length(some.labels) + log2(length(some.labels))
  
  chisq.reference <-
    (((pp.reference*p.reference/length(some.predictions)) -
        tp.reference)^2 / (pp.reference*p.reference/length(some.predictions))
     + ((pp.reference*n.reference/length(some.predictions)) -
          fp.reference)^2 / (pp.reference*n.reference/length(some.predictions))
     + ((np.reference*p.reference/length(some.predictions)) -
          fn.reference)^2 / (np.reference*p.reference/length(some.predictions))
     + ((np.reference*n.reference/length(some.predictions)) -
          tn.reference)^2 / (np.reference*n.reference/length(some.predictions)))
  
  odds.reference <- (tp.reference*tn.reference) / (fn.reference*fp.reference)
  
  lift.reference <- (tp.reference/p.reference) / (pp.reference/(p.reference+n.reference))
  
  f.reference <- 1 / (0.5 * ((1/prec.reference) + (1/rec.reference)))
  
  sar.reference <- 1/3 * (acc.reference + auc.reference + (1-rmse.reference))
  cost.reference <- (fpr.reference * n.reference/length(some.labels) * 1 +
                       fnr.reference * p.reference/length(some.labels) * 1)
  
  .get.performance.measures <- function(pred) {
    .get.performance.measure.result <- function(pred, measure){
      perf <- performance(pred, measure)
      show(perf)
      perf@y.values[[1]]
    }
    tpr <- .get.performance.measure.result(pred, "tpr")
    fpr <- .get.performance.measure.result(pred, "fpr")
    acc <- .get.performance.measure.result(pred, "acc")
    err <- .get.performance.measure.result(pred, "err")
    
    rec <- .get.performance.measure.result(pred, "rec")
    sens<- .get.performance.measure.result(pred, "sens")
    fnr <- .get.performance.measure.result(pred, "fnr")
    tnr <- .get.performance.measure.result(pred, "tnr")
    spec<- .get.performance.measure.result(pred, "spec")
    ppv <- .get.performance.measure.result(pred, "ppv")
    prec<- .get.performance.measure.result(pred, "prec")
    npv <- .get.performance.measure.result(pred, "npv")
    
    fall<- .get.performance.measure.result(pred, "fall")
    miss<- .get.performance.measure.result(pred, "miss")
    pcfall <- .get.performance.measure.result(pred, "pcfall")
    pcmiss <- .get.performance.measure.result(pred, "pcmiss")
    rpp <- .get.performance.measure.result(pred, "rpp")
    rnp <- .get.performance.measure.result(pred, "rnp")
    
    auc <- performance(pred, "auc")@y.values[[1]]
    aucpr <- performance(pred, "aucpr")@y.values[[1]]
    prbe<- performance(pred, "prbe")@y.values[[1]]
    rch <- performance(pred, "rch")@y.values[[1]]

    mxe <- .get.performance.measure.result(pred, "mxe")
    rmse<- .get.performance.measure.result(pred, "rmse")
    
    phi <- .get.performance.measure.result(pred, "phi")
    mat <- .get.performance.measure.result(pred, "mat")
    mi  <- .get.performance.measure.result(pred, "mi")
    chisq<- .get.performance.measure.result(pred, "chisq")
    odds<- .get.performance.measure.result(pred, "odds")
    lift<- .get.performance.measure.result(pred, "lift")
    f   <- .get.performance.measure.result(pred, "f")
    sar <- .get.performance.measure.result(pred,"sar")
    ecost  <- .get.performance.measure.result(pred, "ecost")
    cost  <- .get.performance.measure.result(pred, "cost")
    return(list(tpr=tpr, fpr=fpr, acc=acc, err=err,
                rec=rec, sens=sens, fnr=fnr, tnr=tnr,
                spec=spec, ppv=ppv, prec=prec, npv=npv, 
                fall=fall, miss=miss, pcfall=pcfall, pcmiss=pcmiss, rpp=rpp, rnp=rnp,
                auc=auc, aucpr=aucpr, prbe=prbe, rch=rch, mxe=mxe, 
                rmse=rmse, phi=phi, mat=mat, mi=mi, chisq=chisq, odds=odds,
                lift=lift, f=f, sar=sar, ecost=ecost, cost=cost))
    
  }
  
  ##############################################################################
  # test PerformanceMeasuresReference
  expect_error(prediction(some.predictions[-1], some.labels),
               "Number of predictions in each run must be equal")
  expect_error(prediction(c(NA,some.predictions[-1]), some.labels),
               "'predictions' contains NA.")
  expect_error(prediction(as.list(matrix(some.predictions)), some.labels),
               "Number of cross-validation runs must be equal")
  expect_error(prediction(some.predictions, factor(some.labels,ordered = TRUE),
                          label.ordering = c(1,0)),
               "'labels' is already ordered. No additional 'label.ordering'")
  expect_error()
  
  pred <- prediction(some.predictions, some.labels)
  expect_output(show(pred))
  actual <- prediction(some.predictions, factor(some.labels),
                       label.ordering = c(0,1))
  expect_equal(pred, actual)
  expect_error(performance("tpr",pred),
               "Wrong argument types")
  expect_error(performance(pred,"tpr","mxe"),
               "The performance measure mxe can only be used as 'measure'")
  actual1 <- performance(pred, "tpr")
  actual2 <- performance(pred, "fpr")
  actual <- ROCR:::.combine.performance.objects(actual1,actual2)
  expect_s4_class(actual,"performance")
  actual3 <- performance(pred, "mxe")
  expect_error(ROCR:::.combine.performance.objects(actual1,actual3),
               "Objects need to have identical x axis")
  expect_error(ROCR:::.combine.performance.objects(actual,actual),
               "At least one of the two objects has already been merged")
  
  measures <- expect_output(
    expect_warning(.get.performance.measures(pred),
                   "Chi-squared approximation may be incorrect"))
  attach(measures)
  
  expect_equal(tpr, tpr.reference)
  expect_equal(fpr, fpr.reference)
  expect_equal(acc, acc.reference)    
  expect_equal(err, err.reference)
  expect_equal(rec, rec.reference)
  expect_equal(sens, sens.reference)
  expect_equal(fnr, fnr.reference)
  expect_equal(tnr, tnr.reference)
  expect_equal(spec, spec.reference)
  
  expect_equal(ppv, ppv.reference)
  expect_equal(prec,prec.reference)
  expect_equal(npv, npv.reference)
  
  expect_equal(fall, fall.reference)
  expect_equal(miss,miss.reference)
  expect_equal(pcfall, pcfall.reference)
  expect_equal(pcmiss,pcmiss.reference)
  expect_equal(rpp, rpp.reference)
  expect_equal(rnp,rnp.reference)
  
  expect_equal(auc, auc.reference)
  expect_equal(aucpr, aucpr.reference, tolerance = .0000001)
  expect_equal(prbe, prbe.reference)
  
  expect_equal(mxe, mxe.reference)
  
  expect_equal(rmse, rmse.reference)
  expect_equal(phi, phi.reference)
  expect_equal(mat, mat.reference)
  
  expect_equal(mi, mi.reference)
  
  expect_equal(unname(chisq), chisq.reference)
  expect_equal(odds, odds.reference)
  expect_equal(lift, lift.reference)
  
  expect_equal(f, f.reference)
  expect_equal(sar,sar.reference)
  
  expect_equal(cost, cost.reference)
  
  ##############################################################################
  # ecost
  ecost.x.reference <- c(0,1/3,0.5,1)
  ecost.y.reference <- c(0,0.2,0.2,0)
  
  pred <- prediction(some.predictions, some.labels)
  perf <- performance(pred, "ecost")
  ecost.x <- perf@x.values[[1]]
  ecost.y <- perf@y.values[[1]]
  
  expect_equal( ecost.x, ecost.x.reference )
  
  expect_equal( ecost.y, ecost.y.reference )
  
  ##############################################################################
  # test cal
  pred <- prediction(some.predictions, some.labels)
  cal <- performance(pred, "cal", window.size=floor(length(pred@predictions[[1]])/3))@y.values[[1]]
  cal.x <- performance(pred, "cal", window.size=floor(length(pred@predictions[[1]])/3))@x.values[[1]]
  cal.x.reference <- rev(sort( some.predictions ))[2:(length(some.predictions)-1)]
  
  expect_equal( cal, cal.reference)
  expect_equal( cal.x, cal.x.reference)
  
  ##############################################################################
  # test cost
  pred <- prediction(some.predictions, some.labels)
  for (cost.fp in rnorm(50)) {
    cost.fn <- rnorm(1)
    
    perf <- performance(pred, "cost", cost.fp=cost.fp, cost.fn=cost.fn)
    cost <- perf@y.values[[1]]
    my.cost.reference <- (fpr.reference * n.reference/length(some.labels) * cost.fp +
                            fnr.reference * p.reference/length(some.labels) * cost.fn)
    
    expect_equal( cost, my.cost.reference)
  }
  
  ##############################################################################
  # test Rch
  pred <- prediction(some.predictions, some.labels)
  perf <- performance( pred, "rch")
  rch.x <- perf@x.values[[1]]
  rch.y <- perf@y.values[[1]]
  
  expect_equal( rch.x, rch.reference.x )
  expect_equal( rch.y, rch.reference.y )
  
  ##############################################################################
  # test RMSE
  pred <- prediction(c(0, 0, 1, 1),
                     ordered(c(0, 0, 1, 1)))
  rmse <- performance(pred, "rmse")@y.values[[1]]
  expect_equal(rmse, 0)
  
  pred <- prediction(c(0.0, 0.0, 1.0, 1.0),
                     ordered(c(1, 1, 0, 0), levels=c(1,0)))
  rmse <- performance(pred, "rmse")@y.values[[1]]
  
  expect_equal(rmse, 1)
  
  pred <- prediction(c(0.0, 0.0, 1.0, 1.0),
                     ordered(c(2, 2, 3, 3)))
  rmse <- performance(pred, "rmse")@y.values[[1]]
  
  expect_equal( rmse, 2)
  
  pred <- prediction(c(-0.5, 0.2, 2.5, 0.3),
                     ordered(c(-1, -1, 1, 1)))
  rmse <- performance(pred, "rmse")@y.values[[1]]
  
  expect_equal( rmse, sqrt(1/4*(0.5^2 + 1.2^2 + 1.5^2 + 0.7^2)))
  
  ##############################################################################
  # test PRBE
  pred <- prediction(some.predictions, some.labels)
  prbe.y <- performance(pred, "prbe")@y.values[[1]]
  prbe.x <- performance(pred, "prbe")@x.values[[1]]
  expect_equal(prbe.y, prbe.reference)
  expect_equal(prbe.x, prbe.reference.x)
  
  ##############################################################################
  # test prediction interface
  pred <- prediction(seq(0, 1, length=10),
                     c(rep(0,5), rep(1,5)))
  expect_equal(performance(pred, "auc")@y.values[[1]],
              1)    
  pred <- prediction(seq(1, 0, length=10),
                     c(rep(0,5), rep(1,5)))
  expect_equal(performance(pred, "auc")@y.values[[1]],
              0)    
  pred <- prediction(seq(0, 1, length=10),
                     factor(c(rep(0,5), rep(1,5))))
  expect_equal(performance(pred, "auc")@y.values[[1]],
              1)    
  pred <- prediction(seq(0, 1, length=10),
                     ordered(c(rep(0,5), rep(1,5))))
  expect_equal(performance(pred, "auc")@y.values[[1]],
              1)
  pred <- prediction(seq(0, 1, length=10),
                     ordered(c(rep(0,5), rep(1,5)), levels=c(1,0)))
  expect_equal(performance(pred, "auc")@y.values[[1]],
              0)    
  pred <- prediction(seq(0, 1, length=10),
                     ordered(c(rep("A",5), rep("B",5))))
  expect_equal(performance(pred, "auc")@y.values[[1]],
              1)
  
  expect_error(pred <- prediction(seq(0, 1, length=10),
                                    c(rep(0,5), rep(1,5)), label.ordering=c(1,2)))
  expect_error(pred <- prediction(list(c(0.1,0.3,0.7,1),
                                         c(0,0.2,0.8,1)),
                                    list(factor(c(0,0,1,1)),
                                         factor(c(1,1,2,2))))) 
  expect_error(pred <- prediction(list(c(0.2,0.3,0.7,1),
                                         c(0,0.2,0.8,1)),
                                    list(factor(c(0,0,1,1)),
                                         ordered(c(0,0,1,1)))))
  pred <- prediction(list(c(0,0.3,0.7,1),
                          c(0,0.2,0.8,1)),
                     list(factor(c(0,0,1,1)),
                          factor(c(0,0,1,1))))
  expect_equal(performance(pred, "auc")@y.values,
              list(1, 1))
  
  pred1 <- prediction(data.frame(c(0,0.3,0.7,1),
                                 c(0,0.2,0.8,1)),
                      data.frame(factor(c(0,0,1,1)),
                                 factor(c(0,0,1,1))))
  expect_equal( pred, pred1)
  
  pred2 <- prediction(cbind(c(0,0.3,0.7,1),
                            c(0,0.2,0.8,1)),
                      cbind(c(0,0,1,1),
                            c(0,0,1,1)))
  expect_equal(pred, pred2)
})

context("aux")
test_that("aux:",{
    # Farg
    ll <- list(arg1=c(1,2,3), arg2=c(4,5,6))
    expect_equal(.farg(ll, arg3=c(7,8,9)),
                 list(arg1=c(1,2,3), arg2=c(4,5,6), arg3=c(7,8,9)))
    expect_equal(.farg(ll, arg1=c(1,4,3)),
                 list(arg1=c(1,2,3), arg2=c(4,5,6)))
    # Garg
    ll <- list(arg1=list(1,2,3), arg2=list(4,5,6))
    expect_equal(.garg(ll, 'arg1'), 1)
    expect_equal(.garg(ll, 'arg1',2), 2)
    expect_equal(.garg(ll, 'arg2',3), 6)
    expect_equal(.garg(ll, 'arg3'), ll$arg3)
    # Slice
    ll <- list(arg1=list(c(1,2,3), c(2,3,4), c(3,4,5)),
               arg2=list('a', 'b', 'c'))
    expect_equal(.slice.run(ll, 1), list(arg1=c(1,2,3), arg2='a'))
    expect_equal(.slice.run(ll, 2), list(arg1=c(2,3,4), arg2='b'))
    expect_equal(.slice.run(ll, 3), list(arg1=c(3,4,5), arg2='c'))
    
    ll <- list(arg1=list(c(1,2,3), c(2,3,4), c(3,4,5)),
               arg2=c('a', 'b', 'c'))
    expect_equal(.slice.run(ll, 1), list(arg1=c(1,2,3), arg2=c('a', 'b', 'c')))
    expect_equal(.slice.run(ll, 2), list(arg1=c(2,3,4), arg2=c('a', 'b', 'c')))
    expect_equal(.slice.run(ll, 3), list(arg1=c(3,4,5), arg2=c('a', 'b', 'c')))
    # .select.args
    actual <- ROCR:::.select.args(ll, "arg1")
    expect_equal(actual,ll["arg1"])
    actual <- ROCR:::.select.args(ll, "arg1", complement = TRUE)
    expect_equal(actual,ll["arg2"])
    # .construct.linefunct
    actual <- ROCR:::.construct.linefunct(1,2,3,4)
    expect_type(actual, "closure")
    expect_error(ROCR:::.construct.linefunct(1,2,1,4),
                 "Cannot construct a function from data.")
    # .intersection.point
    f <-  ROCR:::.construct.linefunct(1,2,3,4)
    g <-  ROCR:::.construct.linefunct(2,3,4,5)
    actual <- ROCR:::.intersection.point(f,g)
    expect_equal(actual, c(Inf,Inf))
    g <-  ROCR:::.construct.linefunct(2,3,1,5)
    actual <-  ROCR:::.intersection.point(f,g)
    expect_equal(actual, c(2,3))
})

---
title: "ROCR: visualizing classifier performance in R"
output: rmarkdown::html_vignette
author: Tobias Sing, Oliver Sander, Niko Beerenwinkel, Thomas Lengauer
abstract:
  ROCR is a package for evaluating and visualizing the performance of scoring 
  classifiers in the statistical language R. It features over 25 performance 
  measures that can be freely combined to create two-dimensional performance 
  curves. Standard methods for investigating trade-offs between specific 
  performance measures are available within a uniform framework, including 
  receiver operating characteristic (ROC) graphs, precision/recall plots, lift 
  charts and cost curves. ROCR integrates tightly with R's powerful graphics 
  capabilities, thus allowing for highly adjustable plots. Being equipped with 
  only three commands and reasonable default values for optional parameters, 
  ROCR combines flexibility with ease of usage.
vignette: >
  %\VignetteIndexEntry{ROCR}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
bibliography: references.bibtex
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

# Introduction

```{r setup}
library(ROCR)
```

Pattern classification has become a central tool in bioinformatics, offering
rapid insights into large data sets [[@Baldi2001]](#References). While one area
of our work involves predicting phenotypic properties of HIV-1 from genotypic
information
[[@Beerenwinkel2002;@Beerenwinkel2003;@Sing04learningmixtures]](#References),
scoring or ranking predictors are also vital in a wide range of other biological
problems. Examples include microarray analysis (e.g. prediction of tissue
condition based on gene expression), protein structural and functional
characterization (remote homology detection, prediction of post-translational
modifications and molecular function annotation based on sequence or structural
motifs), genome annotation (gene finding and splice site identification),
protein–ligand interactions (virtual screening and molecular docking) and
structure–activity relationships (predicting bioavailability or toxicity of drug
compounds). In many of these cases, considerable class skew, class-specific
misclassification costs, and extensive noise due to variability in experimental
assays complicate predictive modelling. Thus, careful predictor validation is
compulsory.

```{r, echo = FALSE, results = 'asis'}
table <- data.frame(group = c("Contingency ratios",
                              "Discrete covariation measures",
                              "Information retrieval measures",
                              "Performance in ROC space",
                              "Absolute scoring performance",
                              "Cost measures"),
                    measure = c("error rate, accuracy, sensitivity, specificity, true/false positive rate, fallout, miss, precision, recall, negative predictive value, prediction-conditioned fallout/miss.",
                                "Phi/Matthews correlation coefficient, mutual information, Chi-squared test statistic, odds ratio",
                                "F-measure, lift, precision-recall break-even point",
                                "ROC convex hull, area under the ROC curve",
                                "calibration error, mean cross-entropy, root mean-squared error",
                                "expected cost, explicit cost"))
knitr::kable(table,
             caption = "***Table 1:**Performance measures in the ROCR package*",
             col.names = c("",""),
             align = "l")
```

The real-valued output of scoring classifiers is turned into a binary class
decision by choosing a cutoff. As no cutoff is optimal according to all possible
performance criteria, cutoff choice involves a trade-off among different
measures. Typically, a trade-off between a pair of criteria (e.g. sensitivity
versus specificity) is visualized as a cutoff-parametrized curve in the plane
spanned by the two measures. Popular examples of such trade-off visualizations
include receiver operating characteristic (ROC) graphs, sensitivity/specificity
curves, lift charts and precision/recall plots. [@Fawcett2004](#References)
provides a general introduction into evaluating scoring classifiers with a focus
on ROC graphs.

Although functions for drawing ROC graphs are provided by the Bioconductor
project (https://www.bioconductor.org) or by the machine learning package Weka
(https://www.cs.waikato.ac.nz/ml), for example, no comprehensive evaluation suite
is available to date. ROCR is a flexible evaluation package for R
(https://www.r-project.org), a statistical language that is widely used in
biomedical data analysis. Our tool allows for creating cutoff-parametrized
performance curves by freely combining two out of more than 25 performance
measures (Table 1). Curves from different cross-validation or bootstrapping runs
can be averaged by various methods. Standard deviations, standard errors and box
plots are available to summarize the variability across the runs. The
parametrization can be visualized by printing cutoff values at the corresponding
curve positions, or by coloring the curve according to the cutoff. All
components of a performance plot are adjustable using a flexible mechanism for
dispatching optional arguments. Despite this flexibility, ROCR is easy to use,
with only three commands and reasonable default values for all optional
parameters.

In the example below, we will briefly introduce ROCR's three
commands—prediction, performance and plot—applied to a 10-fold cross-validation
set of predictions and corresponding class labels from a study on predicting HIV
coreceptor usage from the sequence of the viral envelope protein. After loading
the dataset, a prediction object is created from the raw predictions and class
labels.

```{r}
data(ROCR.hiv)
predictions <- ROCR.hiv$hiv.svm$predictions
labels <- ROCR.hiv$hiv.svm$labels
pred <- prediction(predictions, labels)
pred
```

Performance measures or combinations thereof are computed by invoking the
performance method on this prediction object. The resulting performance object
can be visualized using the method plot. For example, an ROC curve that trades
off the rate of true positives against the rate of false positives is obtained
as follows:

```{r, fig.asp=1, fig.width=5, fig.align='center'}
perf <- performance(pred, "tpr", "fpr")
perf
plot(perf,
     avg="threshold",
     spread.estimate="boxplot")
```

The optional parameter avg selects a particular form of performance curve
averaging across the validation runs; the visualization of curve variability is
determined with the parameter spread.estimate.

```{r, echo=FALSE, results='asis', fig.asp=0.35, fig.width=7, fig.align='center',fig.cap="***Fig 1:** Visualizations of classifier performance (HIV coreceptor usage data): (a) receiver operating characteristic (ROC) curve; (b) peak accuracy across a range of cutoffs; (c) absolute difference between empirical and predicted rate of positives for windowed cutoff ranges, in order to evaluate how well the scores are calibrated as probability estimates. Owing to the probabilistic interpretation, cutoffs need to be in the interval [0,1], in contrast to other performance plots. (d) Score density estimates for the negative (solid) and positive (dotted) class.*"}
data(ROCR.hiv)
pp.unnorm <- ROCR.hiv$hiv.svm$predictions
ll <- ROCR.hiv$hiv.svm$labels

# normalize predictions to 0..1
v <- unlist(pp.unnorm)
pp <- lapply(pp.unnorm, function(run) {approxfun(c(min(v), max(v)), c(0,1))(run)})

par(mfrow=c(1,4))
pred<- prediction(pp, ll)
perf <- performance(pred, "tpr", "fpr")

plot(perf, avg= "threshold", colorize=TRUE, lwd= 3,
     coloraxis.at=seq(0,1,by=0.2),)
plot(perf, col="gray78", add=TRUE)
plot(perf, avg= "threshold", colorize=TRUE, colorkey=FALSE,lwd= 3,,add=TRUE)
mtext(paste0("(a)"), side = 3, adj = 0.01,line = 1)

perf <- performance(pred, "acc")
plot(perf, avg= "vertical", spread.estimate="boxplot", lwd=3,col='blue',
     show.spread.at= seq(0.1, 0.9, by=0.1),)
mtext(paste0("(b)"), side = 3, adj = 0.01,line = 1)


plot(performance(pred, "cal", window.size= 10),
     avg="vertical",)
mtext(paste0("(c)"), side = 3, adj = 0.01,line = 1)

plot(0,0,type="n", xlim= c(0,1), ylim=c(0,7),
     xlab="Cutoff", ylab="Density",)
mtext(paste0("(d)"), side = 3, adj = 0.01,line = 1)
for (runi in 1:length(pred@predictions)) {
  lines(density(pred@predictions[[runi]][pred@labels[[runi]]=="-1"]), col= "red")
  lines(density(pred@predictions[[runi]][pred@labels[[runi]]=="1"]), col="green")
}
```

Issuing `demo(ROCR)` starts a demonstration of further graphical capabilities of
ROCR. The command `help(package=ROCR)` points to the available help pages. In
particular, a complete list of available performance measures can be obtained
via help(performance). A reference manual can be downloaded from the ROCR
website.

In conclusion, ROCR is a comprehensive tool for evaluating scoring classifiers
and producing publication-quality figures. It allows for studying the
intricacies inherent to many biological datasets and their implications on
classifier performance.

## Additional examples

Below you can find many additional examples of ROCR's features of performance 
measurement and the possibilites in plotting. However, this only a first 
taste. For more examples, please run `demo(ROCR)` and make sure the plotting
deminsions are big enough.

### ROC curves, Precision/Recall graphs and more ...

```{r, fig.asp=1, fig.width=5, fig.align='center'}
perf <- performance(pred, "tpr", "fpr")
plot(perf,
     avg= "threshold",
     colorize=TRUE,
     lwd= 3,
     main= "With ROCR you can produce standard plots\nlike ROC curves ...")
plot(perf,
     lty=3,
     col="grey78",
     add=TRUE)
```


```{r, fig.asp=1, fig.width=5, fig.align='center'}
perf <- performance(pred, "prec", "rec")
plot(perf,
     avg= "threshold",
     colorize=TRUE,
     lwd= 3,
     main= "... Precision/Recall graphs ...")
plot(perf,
     lty=3,
     col="grey78",
     add=TRUE)
```


```{r, fig.asp=1, fig.width=5, fig.align='center'}
perf <- performance(pred, "sens", "spec")
plot(perf,
     avg= "threshold",
     colorize=TRUE,
     lwd= 3,
     main="... Sensitivity/Specificity plots ...")
plot(perf,
     lty=3,
     col="grey78",
     add=TRUE)
```


```{r, fig.asp=1, fig.width=5, fig.align='center'}
perf <- performance(pred, "lift", "rpp")
plot(perf,
     avg= "threshold",
     colorize=TRUE,
     lwd= 3,
     main= "... and Lift charts.")
plot(perf,
     lty=3,
     col="grey78",
     add=TRUE)
```

### Averaging over multiple predictions

Multiple batches of predictions can be analyzed at the same time.

```{r}
data(ROCR.xval)
predictions <- ROCR.xval$predictions
labels <- ROCR.xval$labels
length(predictions)
```

```{r}
pred <- prediction(predictions, labels)
perf <- performance(pred,'tpr','fpr')
```

This can be used for plotting averages using the `avg` argument.

```{r, fig.asp=1, fig.width=5, fig.align='center'}
plot(perf,
     colorize=TRUE,
     lwd=2,
     main='ROC curves from 10-fold cross-validation')
```

```{r, fig.asp=1, fig.width=5, fig.align='center'}
plot(perf,
     avg='vertical',
     spread.estimate='stderror',
     lwd=3,main='Vertical averaging + 1 standard error',
     col='blue')
```

```{r, fig.asp=1, fig.width=5, fig.align='center'}
plot(perf,
     avg='horizontal',
     spread.estimate='boxplot',
     lwd=3,
     main='Horizontal averaging + boxplots',
     col='blue')
```

```{r, fig.asp=1, fig.width=5, fig.align='center'}
plot(perf,
     avg='threshold',
     spread.estimate='stddev',
     lwd=2,
     main='Threshold averaging + 1 standard deviation',
     colorize=TRUE)
```

### Cutoff stacking

```{r, fig.asp=1, fig.width=6, fig.align='center'}
plot(perf,
     print.cutoffs.at=seq(0,1,by=0.2),
     text.cex=0.8,
     text.y=lapply(as.list(seq(0,0.5,by=0.05)), function(x) { rep(x,length(perf@x.values[[1]])) } ),
     col= as.list(terrain.colors(10)),
     text.col= as.list(terrain.colors(10)), 
     points.col= as.list(terrain.colors(10)), 
     main= "Cutoff stability")
```

### Combination of performance measures

Performance measures can be combined freely.

```{r}
perf <- performance(pred,"pcmiss","lift")
```

```{r, fig.asp=1, fig.width=5, fig.align='center'}
plot(perf,
     colorize=TRUE,
     print.cutoffs.at=seq(0,1,by=0.1),
     text.adj=c(1.2,1.2),
     avg="threshold",
     lwd=3,
     main= "You can freely combine performance measures ...")
```


# Acknowledgement

Work at MPI supported by EU NoE BioSapiens (LSHG-CT-2003-503265).

<a name="References"></a>

# References

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz.R
\docType{class}
\name{prediction-class}
\alias{prediction-class}
\title{Class \code{prediction}}
\description{
Object to encapsulate numerical predictions together with the
corresponding true class labels, optionally collecting predictions and
labels for several cross-validation or bootstrapping runs.
}
\section{Slots}{

\describe{
\item{\code{predictions}}{A list, in which each element is a vector of predictions
(the list has length > 1 for x-validation data.}

\item{\code{labels}}{Analogously, a list in which each element is a vector of true
class labels.}

\item{\code{cutoffs}}{A list in which each element is a vector of all necessary
cutoffs. Each cutoff vector consists of the predicted scores (duplicates
removed), in descending order.}

\item{\code{fp}}{A list in which each element is a vector of the number (not the
rate!) of false positives induced by the cutoffs given in the corresponding
'cutoffs' list entry.}

\item{\code{tp}}{As fp, but for true positives.}

\item{\code{tn}}{As fp, but for true negatives.}

\item{\code{fn}}{As fp, but for false negatives.}

\item{\code{n.pos}}{A list in which each element contains the number of positive
samples in the given x-validation run.}

\item{\code{n.neg}}{As n.pos, but for negative samples.}

\item{\code{n.pos.pred}}{A list in which each element is a vector of the number of
samples predicted as positive at the cutoffs given in the corresponding
'cutoffs' entry.}

\item{\code{n.neg.pred}}{As n.pos.pred, but for negatively predicted samples.}
}}

\note{
Every \code{prediction} object contains information about the 2x2
contingency table consisting of tp,tn,fp, and fn, along with the
marginal sums n.pos,n.neg,n.pos.pred,n.neg.pred, because these form
the basis for many derived performance measures.
}
\section{Objects from the Class}{

Objects can be created by using the \code{prediction} function.
}

\references{
A detailed list of references can be found on the ROCR homepage at
\url{https://ipa-tys.github.io/ROCR/}.
}
\seealso{
\code{\link{prediction}},
\code{\link{performance}},
\code{\link{performance-class}},
\code{\link{plot.performance}}
}
\author{
Tobias Sing \email{tobias.sing@gmail.com}, Oliver Sander
\email{osander@gmail.com}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prediction.R
\name{prediction}
\alias{prediction}
\title{Function to create prediction objects}
\usage{
prediction(predictions, labels, label.ordering = NULL)
}
\arguments{
\item{predictions}{A vector, matrix, list, or data frame containing the
predictions.}

\item{labels}{A vector, matrix, list, or data frame containing the true class
labels. Must have the same dimensions as \code{predictions}.}

\item{label.ordering}{The default ordering (cf.details)  of the classes can
be changed by supplying a vector containing the negative and the positive
class label.}
}
\value{
An S4 object of class \code{prediction}.
}
\description{
Every classifier evaluation using ROCR starts with creating a
\code{prediction} object. This function is used to transform the input data
(which can be in vector, matrix, data frame, or list form) into a
standardized format.
}
\details{
\code{predictions} and \code{labels} can simply be vectors of the same
length. However, in the case of cross-validation data, different
cross-validation runs can be provided as the *columns* of a matrix or
data frame, or as the entries of a list. In the case of a matrix or
data frame, all cross-validation runs must have the same length, whereas
in the case of a list, the lengths can vary across the cross-validation
runs. Internally, as described in section 'Value', all of these input
formats are converted to list representation.

Since scoring classifiers give relative tendencies towards a negative
(low scores) or positive (high scores) class, it has to be declared
which class label denotes the negative, and which the positive class.
Ideally, labels should be supplied as ordered factor(s), the lower
level corresponding to the negative class, the upper level to the
positive class. If the labels are factors (unordered), numeric,
logical or characters, ordering of the labels is inferred from
R's built-in \code{<} relation (e.g. 0 < 1, -1 < 1, 'a' < 'b',
FALSE < TRUE). Use \code{label.ordering} to override this default
ordering. Please note that the ordering can be locale-dependent
e.g. for character labels '-1' and '1'.

Currently, ROCR supports only binary classification (extensions toward
multiclass classification are scheduled for the next release,
however). If there are more than two distinct label symbols, execution
stops with an error message. If all predictions use the same two
symbols that are used for the labels, categorical predictions are
assumed. If there are more than two predicted values, but all numeric,
continuous predictions are assumed (i.e. a scoring
classifier). Otherwise, if more than two symbols occur in the
predictions, and not all of them are numeric, execution stops with an
error message.
}
\examples{
# create a simple prediction object
library(ROCR)
data(ROCR.simple)
pred <- prediction(ROCR.simple$predictions,ROCR.simple$labels)
pred
}
\references{
A detailed list of references can be found on the ROCR homepage at
\url{https://ipa-tys.github.io/ROCR/}.
}
\seealso{
\code{\link{prediction-class}},
\code{\link{performance}},
\code{\link{performance-class}},
\code{\link{plot.performance}}
}
\author{
Tobias Sing \email{tobias.sing@gmail.com}, Oliver Sander
\email{osander@gmail.com}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz.R
\docType{class}
\name{performance-class}
\alias{performance-class}
\title{Class \code{performance}}
\description{
Object to capture the result of a performance evaluation, optionally
collecting evaluations from several cross-validation or bootstrapping runs.
}
\details{
A \code{performance} object can capture information from four
different evaluation scenarios:
  \itemize{
    \item The behaviour of a cutoff-dependent performance measure across
    the range of all cutoffs (e.g. \code{performance( predObj, 'acc' )} ). Here,
    \code{x.values} contains the cutoffs, \code{y.values} the
    corresponding values of the performance measure, and
    \code{alpha.values} is empty.\cr
    \item The trade-off between two performance measures across the
    range of all cutoffs (e.g. \code{performance( predObj,
                                                  'tpr', 'fpr' )} ). In this case, the cutoffs are stored in
    \code{alpha.values}, while \code{x.values} and \code{y.values}
    contain the corresponding values of the two performance measures.\cr
    \item A performance measure that comes along with an obligatory
    second axis (e.g. \code{performance( predObj, 'ecost' )} ). Here, the measure values are
    stored in \code{y.values}, while the corresponding values of the
    obligatory axis are stored in \code{x.values}, and \code{alpha.values}
    is empty.\cr
    \item A performance measure whose value is just a scalar
    (e.g. \code{performance( predObj, 'auc' )} ). The value is then stored in
    \code{y.values}, while \code{x.values} and \code{alpha.values} are
    empty.
  }
}
\section{Slots}{

\describe{
\item{\code{x.name}}{Performance measure used for the x axis.}

\item{\code{y.name}}{Performance measure used for the y axis.}

\item{\code{alpha.name}}{Name of the unit that is used to create the parametrized
curve. Currently, curves can only be parametrized by cutoff, so
\code{alpha.name} is either \code{none} or \code{cutoff}.}

\item{\code{x.values}}{A list in which each entry contains the x values of the curve
of this particular cross-validation run. \code{x.values[[i]]},
\code{y.values[[i]]}, and \code{alpha.values[[i]]} correspond to each
other.}

\item{\code{y.values}}{A list in which each entry contains the y values of the curve
of this particular cross-validation run.}

\item{\code{alpha.values}}{A list in which each entry contains the cutoff values of
the curve of this particular cross-validation run.}
}}

\section{Objects from the Class}{

Objects can be created by using the \code{performance} function.
}

\references{
A detailed list of references can be found on the ROCR homepage at
\url{https://ipa-tys.github.io/ROCR/}.
}
\seealso{
\code{\link{prediction}}
\code{\link{performance}},
\code{\link{prediction-class}},
\code{\link{plot.performance}}
}
\author{
Tobias Sing \email{tobias.sing@gmail.com}, Oliver Sander
\email{osander@gmail.com}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz.R
\docType{data}
\name{ROCR.xval}
\alias{ROCR.xval}
\title{Data set: Artificial cross-validation data for use with ROCR}
\format{
A two element list. The first element, \code{ROCR.xval$predictions}, is
itself a 10 element list. Each of these 10 elements is a vector of numerical
predictions for each cross-validation run. Likewise, the second list entry,
\code{ROCR.xval$labels} is a 10 element list in which each element is a
vector of true class labels corresponding to the predictions.
}
\usage{
data(ROCR.xval)
}
\description{
A mock data set containing 10 sets of predictions and corresponding labels as
would be obtained from 10-fold cross-validation.
}
\examples{
# plot ROC curves for several cross-validation runs (dotted
# in grey), overlaid by the vertical average curve and boxplots
# showing the vertical spread around the average.
library(ROCR)
data(ROCR.xval)
pred <- prediction(ROCR.xval$predictions, ROCR.xval$labels)
pred
perf <- performance(pred,"tpr","fpr")
perf
plot(perf,col="grey82",lty=3)
plot(perf,lwd=3,avg="vertical",spread.estimate="boxplot",add=TRUE)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/performance.R
\name{performance}
\alias{performance}
\title{Function to create performance objects}
\usage{
performance(prediction.obj, measure, x.measure = "cutoff", ...)
}
\arguments{
\item{prediction.obj}{An object of class \code{prediction}.}

\item{measure}{Performance measure to use for the evaluation. A complete list
of the performance measures that are available for \code{measure} and
\code{x.measure} is given in the 'Details' section.}

\item{x.measure}{A second performance measure. If different from the default,
a two-dimensional curve, with \code{x.measure} taken to be the unit in
direction of the x axis, and \code{measure} to be the unit in direction of
the y axis, is created. This curve is parametrized with the cutoff.}

\item{...}{Optional arguments (specific to individual performance measures).}
}
\value{
An S4 object of class \code{performance}.
}
\description{
All kinds of predictor evaluations are performed using this function.
}
\details{
Here is the list of available performance measures. Let Y and
\eqn{\hat{Y}}{Yhat} be random variables representing the class and the prediction for
a randomly drawn sample, respectively. We denote by
\eqn{\oplus}{+} and \eqn{\ominus}{-} the positive and
negative class, respectively. Further, we use the following
abbreviations for empirical quantities: P (# positive
samples), N (# negative samples), TP (# true positives), TN (# true
negatives), FP (# false positives), FN (# false negatives).
\describe{
 \item{\code{acc}:}{Accuracy. \eqn{P(\hat{Y}=Y)}{P(Yhat = Y)}. Estimated
   as: \eqn{\frac{TP+TN}{P+N}}{(TP+TN)/(P+N)}.}
 \item{\code{err}:}{Error rate. \eqn{P(\hat{Y}\ne Y)}{P(Yhat !=
                                                          Y)}. Estimated as: \eqn{\frac{FP+FN}{P+N}}{(FP+FN)/(P+N)}.}
 \item{\code{fpr}:}{False positive rate. \eqn{P(\hat{Y}=\oplus | Y =
                                                   \ominus)}{P(Yhat = + | Y = -)}. Estimated as:
     \eqn{\frac{FP}{N}}{FP/N}.}
 \item{\code{fall}:}{Fallout. Same as \code{fpr}.}
 \item{\code{tpr}:}{True positive
   rate. \eqn{P(\hat{Y}=\oplus|Y=\oplus)}{P(Yhat = + | Y = +)}. Estimated
   as: \eqn{\frac{TP}{P}}{TP/P}.}
 \item{\code{rec}:}{Recall. Same as \code{tpr}.}
 \item{\code{sens}:}{Sensitivity. Same as \code{tpr}.}
 \item{\code{fnr}:}{False negative
   rate. \eqn{P(\hat{Y}=\ominus|Y=\oplus)}{P(Yhat = - | Y =
                                               +)}. Estimated as: \eqn{\frac{FN}{P}}{FN/P}.}
 \item{\code{miss}:}{Miss. Same as \code{fnr}.}
 \item{\code{tnr}:}{True negative rate. \eqn{P(\hat{Y} =
                                                  \ominus|Y=\ominus)}{P(Yhat = - | Y = -)}.}
 \item{\code{spec}:}{Specificity. Same as \code{tnr}.}
 \item{\code{ppv}:}{Positive predictive
   value. \eqn{P(Y=\oplus|\hat{Y}=\oplus)}{P(Y = + | Yhat =
                                               +)}. Estimated as: \eqn{\frac{TP}{TP+FP}}{TP/(TP+FP)}.}
 \item{\code{prec}:}{Precision. Same as \code{ppv}.}
 \item{\code{npv}:}{Negative predictive
   value. \eqn{P(Y=\ominus|\hat{Y}=\ominus)}{P(Y = - | Yhat =
                                                 -)}. Estimated as: \eqn{\frac{TN}{TN+FN}}{TN/(TN+FN)}.}
 \item{\code{pcfall}:}{Prediction-conditioned
   fallout. \eqn{P(Y=\ominus|\hat{Y}=\oplus)}{P(Y = - | Yhat =
                                                  +)}. Estimated as: \eqn{\frac{FP}{TP+FP}}{FP/(TP+FP)}.}
 \item{\code{pcmiss}:}{Prediction-conditioned
   miss. \eqn{P(Y=\oplus|\hat{Y}=\ominus)}{P(Y = + | Yhat =
                                               -)}. Estimated as: \eqn{\frac{FN}{TN+FN}}{FN/(TN+FN)}.}
 \item{\code{rpp}:}{Rate of positive predictions. \eqn{P( \hat{Y} =
                                                           \oplus)}{P(Yhat = +)}. Estimated as: (TP+FP)/(TP+FP+TN+FN).}
 \item{\code{rnp}:}{Rate of negative predictions. \eqn{P( \hat{Y} =
                                                           \ominus)}{P(Yhat = -)}. Estimated as: (TN+FN)/(TP+FP+TN+FN).}
 \item{\code{phi}:}{Phi correlation coefficient. \eqn{\frac{TP \cdot
   TN - FP \cdot FN}{\sqrt{ (TP+FN) \cdot (TN+FP) \cdot (TP+FP)
     \cdot (TN+FN)}}}{(TP*TN -
                         FP*FN)/(sqrt((TP+FN)*(TN+FP)*(TP+FP)*(TN+FN)))}. Yields a
   number between -1 and 1, with 1 indicating a perfect
   prediction, 0 indicating a random prediction. Values below 0
   indicate a worse than random prediction.}
 \item{\code{mat}:}{Matthews correlation coefficient. Same as \code{phi}.}
 \item{\code{mi}:}{Mutual information. \eqn{I(\hat{Y},Y) := H(Y) -
     H(Y|\hat{Y})}{I(Yhat, Y) := H(Y) - H(Y | Yhat)}, where H is the
   (conditional) entropy. Entropies are estimated naively (no bias
                                                           correction).}
 \item{\code{chisq}:}{Chi square test statistic. \code{?chisq.test}
   for details. Note that R might raise a warning if the sample size
   is too small.}
 \item{\code{odds}:}{Odds ratio. \eqn{\frac{TP \cdot TN}{FN \cdot
   FP}}{(TP*TN)/(FN*FP)}. Note that odds ratio produces
   Inf or NA values for all cutoffs corresponding to FN=0 or
   FP=0. This can substantially decrease the plotted cutoff region.}
 \item{\code{lift}:}{Lift
   value. \eqn{\frac{P(\hat{Y}=\oplus|Y=\oplus)}{P(\hat{Y}=\oplus)}}{P(Yhat = + |
                                                                         Y = +)/P(Yhat = +)}.}
 \item{\code{f}:}{Precision-recall F measure (van Rijsbergen, 1979). Weighted
   harmonic mean of precision (P) and recall (R). \eqn{F =
     \frac{1}{\alpha \frac{1}{P} + (1-\alpha)\frac{1}{R}}}{F = 1/
       (alpha*1/P + (1-alpha)*1/R)}. If
   \eqn{\alpha=\frac{1}{2}}{alpha=1/2}, the mean is balanced. A
   frequent equivalent formulation is
   \eqn{F = \frac{(\beta^2+1) \cdot P \cdot R}{R + \beta^2 \cdot
     P}}{F = (beta^2+1) * P * R / (R + beta^2 * P)}. In this formulation, the
     mean is balanced if \eqn{\beta=1}{beta=1}. Currently, ROCR only accepts
     the alpha version as input (e.g. \eqn{\alpha=0.5}{alpha=0.5}). If no 
     value for alpha is given, the mean will be balanced by default.}
 \item{\code{rch}:}{ROC convex hull. A ROC (=\code{tpr} vs \code{fpr}) curve 
   with concavities (which represent suboptimal choices of cutoff) removed 
   (Fawcett 2001). Since the result is already a parametric performance 
   curve, it cannot be used in combination with other measures.}
 \item{\code{auc}:}{Area under the ROC curve. This is equal to the value of the
   Wilcoxon-Mann-Whitney test statistic and also the probability that the
   classifier will score are randomly drawn positive sample higher than a
   randomly drawn negative sample. Since the output of
   \code{auc} is cutoff-independent, this
   measure cannot be combined with other measures into a parametric
   curve. The partial area under the ROC curve up to a given false
   positive rate can be calculated by passing the optional parameter
   \code{fpr.stop=0.5} (or any other value between 0 and 1) to 
   \code{performance}.}
 \item{\code{aucpr}:}{Area under the Precision/Recall curve. Since the output
   of \code{aucpr} is cutoff-independent, this measure cannot be combined 
   with other measures into a parametric curve.}
 \item{\code{prbe}:}{Precision-recall break-even point. The cutoff(s) where
   precision and recall are equal. At this point, positive and negative
   predictions are made at the same rate as their prevalence in the
   data. Since the output of
   \code{prbe} is just a cutoff-independent scalar, this
   measure cannot be combined with other measures into a parametric curve.}
 \item{\code{cal}:}{Calibration error. The calibration error is the
   absolute difference between predicted confidence and actual reliability. This
   error is estimated at all cutoffs by sliding a window across the
   range of possible cutoffs. The default window size of 100 can be
   adjusted by passing the optional parameter \code{window.size=200}
   to \code{performance}. E.g., if for several
   positive samples the output of the classifier is around 0.75, you might
   expect from a well-calibrated classifier that the fraction of them
   which is correctly predicted as positive is also around 0.75. In a
   well-calibrated classifier, the probabilistic confidence estimates
   are realistic. Only for use with
   probabilistic output (i.e. scores between 0 and 1).}
 \item{\code{mxe}:}{Mean cross-entropy. Only for use with
   probabilistic output. \eqn{MXE :=-\frac{1}{P+N}( \sum_{y_i=\oplus}
                                                   ln(\hat{y}_i) + \sum_{y_i=\ominus} ln(1-\hat{y}_i))}{MXE := - 1/(P+N) \sum_{y_i=+}
                                                     ln(yhat_i) + \sum_{y_i=-} ln(1-yhat_i)}. Since the output of
   \code{mxe} is just a cutoff-independent scalar, this
   measure cannot be combined with other measures into a parametric curve.}
 \item{\code{rmse}:}{Root-mean-squared error. Only for use with
   numerical class labels. \eqn{RMSE:=\sqrt{\frac{1}{P+N}\sum_i (y_i
                                                                 - \hat{y}_i)^2}}{RMSE := sqrt(1/(P+N) \sum_i (y_i -
                                                                                                                 yhat_i)^2)}. Since the output of
   \code{rmse} is just a cutoff-independent scalar, this
   measure cannot be combined with other measures into a parametric curve.}
 \item{\code{sar}:}{Score combinining performance measures of different
   characteristics, in the attempt of creating a more "robust"
   measure (cf. Caruana R., ROCAI2004):
     SAR = 1/3 * ( Accuracy + Area under the ROC curve + Root
                   mean-squared error ).}
 \item{\code{ecost}:}{Expected cost. For details on cost curves,
   cf. Drummond&Holte 2000,2004. \code{ecost} has an obligatory x
   axis, the so-called 'probability-cost function'; thus it cannot be
   combined with other measures. While using \code{ecost} one is
   interested in the lower envelope of a set of lines, it might be
   instructive to plot the whole set of lines in addition to the lower
   envelope. An example is given in \code{demo(ROCR)}.}
 \item{\code{cost}:}{Cost of a classifier when
   class-conditional misclassification costs are explicitly given.
   Accepts the optional parameters \code{cost.fp} and
   \code{cost.fn}, by which the costs for false positives and
   negatives can be adjusted, respectively. By default, both are set
   to 1.}
}
}
\note{
Here is how to call \code{performance()} to create some standard
evaluation plots:
\describe{
  \item{ROC curves:}{measure="tpr", x.measure="fpr".}
  \item{Precision/recall graphs:}{measure="prec", x.measure="rec".}
  \item{Sensitivity/specificity plots:}{measure="sens", x.measure="spec".}
  \item{Lift charts:}{measure="lift", x.measure="rpp".}
}
}
\examples{
# computing a simple ROC curve (x-axis: fpr, y-axis: tpr)
library(ROCR)
data(ROCR.simple)
pred <- prediction( ROCR.simple$predictions, ROCR.simple$labels)
pred
perf <- performance(pred,"tpr","fpr")
perf
plot(perf)

# precision/recall curve (x-axis: recall, y-axis: precision)
perf <- performance(pred, "prec", "rec")
perf
plot(perf)

# sensitivity/specificity curve (x-axis: specificity,
# y-axis: sensitivity)
perf <- performance(pred, "sens", "spec")
perf
plot(perf)
}
\references{
A detailed list of references can be found on the ROCR homepage at
\url{https://ipa-tys.github.io/ROCR/}.
}
\seealso{
\code{\link{prediction}},
\code{\link{prediction-class}},
\code{\link{performance-class}},
\code{\link{plot.performance}}
}
\author{
Tobias Sing \email{tobias.sing@gmail.com}, Oliver Sander
\email{osander@gmail.com}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz.R
\name{plot-methods}
\alias{plot-methods}
\alias{plot,performance,missing-method}
\alias{plot.performance}
\title{Plot method for performance objects}
\usage{
\S4method{plot}{performance,missing}(
  x,
  y,
  ...,
  avg = "none",
  spread.estimate = "none",
  spread.scale = 1,
  show.spread.at = c(),
  colorize = FALSE,
  colorize.palette = rev(rainbow(256, start = 0, end = 4/6)),
  colorkey = colorize,
  colorkey.relwidth = 0.25,
  colorkey.pos = "right",
  print.cutoffs.at = c(),
  cutoff.label.function = function(x) {
     round(x, 2)
 },
  downsampling = 0,
  add = FALSE
)

\method{plot}{performance}(...)
}
\arguments{
\item{x}{an object of class \code{performance}}

\item{y}{not used}

\item{...}{Optional graphical parameters to adjust different components of
the performance plot. Parameters are directed to their target component by
prefixing them with the name of the component (\code{component.parameter},
e.g. \code{text.cex}). The following components are available:
\code{xaxis}, \code{yaxis}, \code{coloraxis}, \code{box} (around the
plotting region), \code{points}, \code{text}, \code{plotCI} (error bars),
\code{boxplot}. The names of these components are influenced by the R
functions that are used to create them. Thus, \code{par(component)} can be
used to see which parameters are available for a given component (with the
expection of the three axes; use \code{par(axis)} here). To adjust the
canvas or the performance curve(s), the standard \code{plot} parameters can
be used without any prefix.}

\item{avg}{If the performance object describes several curves (from
cross-validation runs or bootstrap evaluations of one particular method),
the curves from each of the runs can be averaged. Allowed values are
\code{none} (plot all curves separately), \code{horizontal} (horizontal
averaging), \code{vertical} (vertical averaging), and \code{threshold}
(threshold (=cutoff) averaging). Note that while threshold averaging is
always feasible, vertical and horizontal averaging are not well-defined if
the graph cannot be represented as a function x->y and y->x, respectively.}

\item{spread.estimate}{When curve averaging is enabled, the variation around
the average curve can be visualized as standard error bars
(\code{stderror}), standard deviation bars (\code{stddev}), or by using box
plots (\code{boxplot}). Note that the function \code{plotCI}, which is used
internally by ROCR to draw error bars, might raise a warning if the spread
of the curves at certain positions is 0.}

\item{spread.scale}{For \code{stderror} or \code{stddev}, this is a scalar
factor to be multiplied with the length of the standard error/deviation
bar. For example, under normal assumptions, \code{spread.scale=2} can be
used to get approximate 95\% confidence intervals.}

\item{show.spread.at}{For vertical averaging, this vector determines the x
positions for which the spread estimates should be visualized. In contrast,
for horizontal and threshold averaging, the y positions and cutoffs are
determined, respectively. By default, spread estimates are shown at 11
equally spaced positions.}

\item{colorize}{This logical determines whether the curve(s) should be
colorized according to cutoff.}

\item{colorize.palette}{If curve colorizing is enabled, this determines the
color palette onto which the cutoff range is mapped.}

\item{colorkey}{If true, a color key is drawn into the 4\% border
region (default of \code{par(xaxs)} and \code{par(yaxs)}) of the
plot. The color key visualizes the mapping from cutoffs to colors.}

\item{colorkey.relwidth}{Scalar between 0 and 1 that determines the
fraction of the 4\% border region that is occupied by the colorkey.}

\item{colorkey.pos}{Determines if the colorkey is drawn vertically at
the \code{right} side, or horizontally at the \code{top} of the
plot.}

\item{print.cutoffs.at}{This vector specifies the cutoffs which should
be printed as text along the curve at the corresponding curve positions.}

\item{cutoff.label.function}{By default, cutoff annotations along the curve
or at the color key are rounded to two decimal places before printing.
Using a custom \code{cutoff.label.function}, any other transformation can
be performed on the cutoffs instead (e.g. rounding with different precision
or taking the logarithm).}

\item{downsampling}{ROCR can efficiently compute most performance measures
even for data sets with millions of elements. However, plotting of large
data sets can be slow and lead to PS/PDF documents of considerable size. In
that case, performance curves that are indistinguishable from the original
can be obtained by using only a fraction of the computed performance
values. Values for downsampling between 0 and 1 indicate the fraction of
the original data set size to which the performance object should be
downsampled, integers above 1 are interpreted as the actual number of
performance values to which the curve(s) should be downsampled.}

\item{add}{If \code{TRUE}, the curve(s) is/are added to an already existing
plot; otherwise a new plot is drawn.}
}
\description{
This is the method to plot all objects of class performance.
}
\examples{
# plotting a ROC curve:
library(ROCR)
data(ROCR.simple)
pred <- prediction( ROCR.simple$predictions, ROCR.simple$labels )
pred
perf <- performance( pred, "tpr", "fpr" )
perf
plot( perf )

# To entertain your children, make your plots nicer
# using ROCR's flexible parameter passing mechanisms
# (much cheaper than a finger painting set)
par(bg="lightblue", mai=c(1.2,1.5,1,1))
plot(perf, main="ROCR fingerpainting toolkit", colorize=TRUE,
     xlab="Mary's axis", ylab="", box.lty=7, box.lwd=5,
     box.col="gold", lwd=17, colorkey.relwidth=0.5, xaxis.cex.axis=2,
     xaxis.col='blue', xaxis.col.axis="blue", yaxis.col='green', yaxis.cex.axis=2,
     yaxis.at=c(0,0.5,0.8,0.85,0.9,1), yaxis.las=1, xaxis.lwd=2, yaxis.lwd=3,
     yaxis.col.axis="orange", cex.lab=2, cex.main=2)
}
\references{
A detailed list of references can be found on the ROCR homepage at
\url{https://ipa-tys.github.io/ROCR/}.
}
\seealso{
\code{\link{prediction}},
\code{\link{performance}},
\code{\link{prediction-class}},
\code{\link{performance-class}}
}
\author{
Tobias Sing \email{tobias.sing@gmail.com}, Oliver Sander
\email{osander@gmail.com}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz.R
\docType{data}
\name{ROCR.hiv}
\alias{ROCR.hiv}
\title{Data set: Support vector machines and neural networks applied to the
  prediction of HIV-1 coreceptor usage}
\format{
A list consisting of the SVM (\code{ROCR.hiv$hiv.svm}) and NN
(\code{ROCR.hiv$hiv.nn}) classification data. Each of those is in turn a list
consisting of the two elements \code{$predictions} and \code{$labels} (10
element list representing cross-validation data).
}
\usage{
data(ROCR.hiv)
}
\description{
Linear support vector machines (libsvm) and neural networks (R package
nnet) were applied to predict usage of the coreceptors CCR5 and CXCR4
based on sequence data of the third variable loop of the HIV envelope
protein.
}
\examples{
library(ROCR)
data(ROCR.hiv)
attach(ROCR.hiv)
pred.svm <- prediction(hiv.svm$predictions, hiv.svm$labels)
pred.svm
perf.svm <- performance(pred.svm, 'tpr', 'fpr')
perf.svm
pred.nn <- prediction(hiv.nn$predictions, hiv.svm$labels)
pred.nn
perf.nn <- performance(pred.nn, 'tpr', 'fpr')
perf.nn
plot(perf.svm, lty=3, col="red",main="SVMs and NNs for prediction of
HIV-1 coreceptor usage")
plot(perf.nn, lty=3, col="blue",add=TRUE)
plot(perf.svm, avg="vertical", lwd=3, col="red",
     spread.estimate="stderror",plotCI.lwd=2,add=TRUE)
plot(perf.nn, avg="vertical", lwd=3, col="blue",
     spread.estimate="stderror",plotCI.lwd=2,add=TRUE)
legend(0.6,0.6,c('SVM','NN'),col=c('red','blue'),lwd=3)
}
\references{
Sing, T. & Beerenwinkel, N. & Lengauer, T.  "Learning mixtures
of localized rules by maximizing the area under the ROC curve".  1st
  International Workshop on ROC Analysis in AI, 89-96, 2004.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz.R
\docType{data}
\name{ROCR.simple}
\alias{ROCR.simple}
\title{Data set: Simple artificial prediction data for use with ROCR}
\format{
A two element list. The first element, \code{ROCR.simple$predictions}, is a
vector of numerical predictions. The second element,
\code{ROCR.simple$labels}, is a vector of corresponding class labels.
}
\usage{
data(ROCR.simple)
}
\description{
A mock data set containing a simple set of predictions and corresponding
class labels.
}
\examples{
# plot a ROC curve for a single prediction run
# and color the curve according to cutoff.
library(ROCR)
data(ROCR.simple)
pred <- prediction(ROCR.simple$predictions, ROCR.simple$labels)
pred
perf <- performance(pred,"tpr","fpr")
perf
plot(perf,colorize=TRUE)
}
\keyword{datasets}
